// ----------------------------------------------------------------------------
// CommandServer.java
// A server handling GUI actions.
// $Id: CommandServer.java,v 1.4 2011/05/18 14:05:56 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import java.net.*;
import java.util.*;
import java.nio.charset.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.SWT.*;

/**
 * 
 * A server handling GUI actions.
 *
 */
public final class CommandServer extends Thread
{
  // the top window of the GUI
  private TopWindow topWindow;
  
  // the socket to listen on
  private ServerSocket serverSocket; 
  
  // the character set to be used for encoding files
  private Charset charSet;
  
  /***************************************************************************
   * Creates a new command server (which is already started).
   * @param topWindow the top window of the application.
   *************************************************************************/
  public CommandServer(TopWindow topWindow)
  {
    this.topWindow = topWindow;
    charSet = Charset.forName("UTF-8");
    try
    {
      serverSocket = new ServerSocket(0);
    }
    catch(IOException e)
    {
      Main.internalError("error in creating socket: " + e);
    }
    start();
  }
  
  /***************************************************************************
   * Get port number on which server is listening.
   * @return the port number.
   **************************************************************************/
  public int getPortNumber()
  {
    return serverSocket.getLocalPort();
  }
  
  /****************************************************************************
   * Server listens on socket and executes received commands.
   ***************************************************************************/
  public synchronized void run()
  {
    while (true)
    {
      try
      {
        Socket socket = serverSocket.accept();
        BufferedReader reader = 
          new BufferedReader(new InputStreamReader(socket.getInputStream()));
        PrintWriter writer =
          new PrintWriter(new OutputStreamWriter(socket.getOutputStream(), charSet));
        int counter = 0;
        int contentLength = -1;
        String request = null;
        while (true)
        {
          String line = reader.readLine();
          if (line == null)
          {
            System.out.println("Empty string read from socket.");
            Main.exit(-1);
          }
          if (counter == 0) 
            request = line; 
          else if (line.startsWith("Content-Length: "))
          {
            String rest = line.substring(16);
            contentLength = Integer.parseInt(rest);
          }
          counter = counter+1;
          if (line.equals("")) break;
        }
        boolean okay = request.startsWith("POST /");
        if (!okay) 
          Main.internalError("unknown request: " + request);
        int space = request.indexOf(' ', 6);
        if (space == -1)
          Main.internalError("? expected in request: " + request);
        String command = request.substring(6, space);
        Map<String,String> arguments = processArguments(reader, contentLength);
        String result = execute(command, arguments);
        writer.println("HTTP/1.0 200 OK");
        writer.println("Content-Type: application/xhtml+xml");
        writer.println();
        writer.println(result);
        writer.close();
      }
      catch(IOException e)
      {
        System.out.println("Error in reading from socket: " + e);
        Main.exit(-1);
      }
    }
  }
  
  /***************************************************************************
   * Process arguments to mapping of names to values.
   * @param reader the reader to read from.
   * @param length the number of characters to read.
   * @return a mapping of names to values.
   **************************************************************************/
  private Map<String,String> processArguments(BufferedReader reader, int length)
  {
    char[] buffer = new char[length];
    try 
    {
      reader.read(buffer);
      String arguments = new String(buffer);
      Map<String,String> result = new MapClass<String,String>();
      int pos = 0;
      while (pos < length)
      {
        int pos1 = arguments.indexOf('=', pos+1);
        if (pos1 == -1) Main.internalError("no equality sign");
        int pos2 = arguments.indexOf('&', pos1+1);
        if (pos2 == -1) pos2 = length;
        String name = arguments.substring(pos, pos1);
        String value = arguments.substring(pos1+1, pos2);
        String value0 = URLDecoder.decode(value, "UTF-8");
        result.put(name, value0);
        pos = pos2+1;
      }
      return result;
    }
    catch(UnsupportedEncodingException e)
    {
      Main.internalError("error in decoding");
    }
    catch(IOException e)
    {
      Main.internalError("error in reading arguments");
    }
    return null;
  }
  
  /***************************************************************************
   * Process command with denoted arguments
   * @param command the command.
   * @param arguments a mapping of names to values.
   * @return the content of the result document to be displayed.
   **************************************************************************/
  private String execute(String command, Map<String,String> arguments)
  {
    if (command.equals(Presenter.nameFormStatAction))
    {
      return PropagatorCommand.execute(topWindow, arguments);
    }
    Main.internalError("unknown command: " + command);
    return null;
  }
}