// ----------------------------------------------------------------------------
// Directory.java
// The handling of directories.
// $Id: Directory.java,v 1.9 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import fmrisc.ProgramExplorer.*;

/**
 * 
 * The handling of directories.
 *
 */
public final class Directory
{ 
  // name of task directory
  public static final String taskdirname = ".PETASKS";
 
  // name of tag used to determine whether directory was previously created
  private final static String TAGNAME = ".PEDIR";
  
  // name of context directory (tag check omitted since created by ProofNavigator)
  public final static String CONTEXT = "ProofNavigator";
  
  // appendix of tag file that indicates directory may be reused 
  public final static String FREED = "FREED";
  
  /****************************************************************************
   * Set up task task directory in denoted parent directory.
   * @param parent the parent directory
   * @param error the stream to write error messages to.
   * @return the task directory.
   ***************************************************************************/
  public static File setup(File parent, ErrorStream error)
  {
    File dir = new File(parent, taskdirname);
    if (dir.exists()) // Directory.remove(dir);
    {
      if (!isCreatedDirectory(dir)) 
      {
        error.newError("directory " + dir + 
            " exists but was not previously created as a task directory");
        return null;
      }
      return dir;
    }
    return Directory.create(parent, taskdirname, true);
  }
  
  /****************************************************************************
   * Returns true if dir denotes a previously created directory.
   * @param dir a file.
   * @return true if dir denotes a directory that contains file named TAGNAME.
   ***************************************************************************/
  public static boolean isCreatedDirectory(File dir)
  {
    File tagFile = new File(dir, TAGNAME);
    return tagFile.exists();
  }
  
  /****************************************************************************
   * Create writable subdirectory in denoted parent directory.
   * Aborts if an error occurs.
   * @param parent the writable parent directory.
   * @param name the name of the subdirectory to be created.
   *        (if !fixed, the actual directory name may be different)
   * @param fixed false iff the actual directory name may be different.
   * @return the directory
   ***************************************************************************/
  public static File create(File parent, String name, boolean fixed)
  {
    // sanity checks
    if (!parent.isDirectory()) 
    {
      Main.internalError("path " + parent + " does not denote a directory");
      return null;
    }
    if (!parent.canWrite()) 
    {
      Main.internalError("directory " + parent + " is not writable");
      return null;
    }
    
    // determine directory name
    if (fixed)
    {
      if (name.indexOf(File.separatorChar) != -1)
      {
        Main.internalError("directory name " + name + " must not contain " +
            File.separatorChar);
        return null;
      }
      if (name.indexOf(' ') != -1)
      {
        Main.internalError("directory name " + name + 
            " must not contain white space");
        return null;
      }
    }
    else
    {
      name = name.replace(File.separatorChar, '_');
      name = name.replace(' ', '_');
    }
    
    // file names are time stamped
    long timeStamp = Main.getTimeStamp();
    
    // try to reuse appropriate directory from previous invocation
    File[] files = parent.listFiles();
    if (files == null)
    {
      Main.internalError("cannot list files from directory " + parent);
      return null;
    }
    for (File file0 : files)
    {
      if (!file0.isDirectory()) continue;
      String name0 = file0.getName();
      if (!name0.startsWith(name + ".")) continue;
      if (!isCreatedDirectory(file0)) continue;
      if (name0.startsWith(name + "." + timeStamp + ".")) 
      {
        File freed = new File(file0, FREED);
        if (!freed.exists()) continue;
        boolean okay = freed.delete();
        if (!okay) Main.internalError("cannot delete file " + freed.getAbsolutePath());
        return file0;
      }
      int counter = 0;
      while (true)
      {
        File file = new File(parent, name + "." + timeStamp + "." + counter);
        boolean okay = file0.renameTo(file);
        if (okay) 
        {
          File freed = new File(file, FREED);
          if (freed.exists()) 
          {
            okay = freed.delete();
            if (!okay) Main.internalError("cannot delete file " + freed.getAbsolutePath());
          }
          return file;
        }
        counter = counter+1;
      }
    }
      
    // create new directory
    int counter = 0;
    while (true)
    {
      File dir = new File(parent, name + "." + timeStamp + "." + counter);
      boolean okay = dir.mkdir();
      if (okay) 
      { 
        okay = dir.setWritable(true, true);
        if (!okay) 
        {
          Main.internalError("cannot set directory " + dir + " to writable");
          return null;
        }
        File file = new File(dir, TAGNAME);
        try
        {
          file.createNewFile();
        }
        catch(IOException e)
        {
          Main.internalError("cannot create file " + file);
          return null;
        }
        return dir;
      }
      if (fixed)
      {
        Main.internalError("cannot create directory " + dir);
        return null;
      }
      counter = counter+1;
    }
  }
  
  /****************************************************************************
   * Remove a previously created directory (including all its contents).
   * Aborts if an error occurs.
   * @param dir the directory to be removed.
   ***************************************************************************/
  public static void remove(File dir)
  {
    // check whether directory is created except for context directories
    if (!dir.getName().equals(CONTEXT) && !isCreatedDirectory(dir)) 
    {
      Main.internalError("path " + dir + 
          " does not denote a previously constructed directory");
    }
    File[] files = dir.listFiles();   
    for (File file : files)
    {
      if (file.isDirectory()) 
      {
        remove(file);
        continue;
      }
      boolean okay = file.delete();
      if (!okay)
      {
        Main.internalError("cannot remove file " + file);
      }
    }
    boolean okay = dir.delete();
    if (!okay) 
    {
      Main.internalError("cannot remove directory " + dir);
    }
  }
  
  /****************************************************************************
   * Attempt to invalidate a previously created directory for immediate reuse.
   * Aborts if an error occurs.
   * @param dir the directory to be invalidated and reused.
   * @param clear true if content of directory is to be removed.
   ***************************************************************************/
  public static void free(File dir, boolean clear)
  {
    // check whether directory is created except for context directories
    if (!dir.getName().equals(CONTEXT) && !isCreatedDirectory(dir)) 
    {
      Main.internalError("path " + dir + 
          " does not denote a previously constructed directory");
    }
    
    // clear directory EXCEPT FOR TAG
    if (clear)
    {
      File[] files = dir.listFiles();   
      for (File file : files)
      {
        if (file.isDirectory()) 
        {
          remove(file);
          continue;
        }
        if (file.getName().equals(TAGNAME)) continue;
        boolean okay = file.delete();
        if (!okay)
        {
          Main.internalError("cannot remove file " + file);
        }
      }
    }
    
    // attempt to create FREED file
    File file = new File(dir, FREED);
    try 
    {
      file.createNewFile();
    }
    catch(IOException e) 
    { 
      if (!file.exists())
        Main.internalError("exception when creating FREED file");
    }
  }
}
