// ----------------------------------------------------------------------------
// VirtualDirectoryServer.java
// A server handling HTTP requests to documents a virtual directory.
// $Id: VirtualDirectoryServer.java,v 1.1 2010/09/24 16:32:38 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.InOut;

import java.io.*;
import java.net.*;
import java.util.*;
import java.text.*;
import java.nio.charset.*;

import fmrisc.ProgramExplorer.*;

/**
 * 
 * A server handling HTTP requests to documents a virtual directory.
 *
 */
public final class VirtualDirectoryServer extends Thread
{
  // the socket to listen on
  private ServerSocket serverSocket; 

  // the directories to be served
  private Map<String, VirtualDirectory> directories;
  
  // auxiliary object for date formatting
  private SimpleDateFormat dateFormat;
  
  // the character set to be used for encoding files
  private Charset charSet;
  
  /***************************************************************************
   * Creates a new command server (which is already started).
   * @param directories the directories to be served.
   *************************************************************************/
  public VirtualDirectoryServer(Map<String, VirtualDirectory> directories)
  {
    this.directories = directories;
    dateFormat = new SimpleDateFormat("E, d MMM yyyy HH:mm:ss 'GMT'", Locale.US);
    dateFormat.setTimeZone(TimeZone.getTimeZone("GMT"));
    charSet = Charset.forName("UTF-8");
    try
    {
      serverSocket = new ServerSocket(0);
    }
    catch(IOException e)
    {
      Main.internalError("error in creating socket: " + e);
    }
    start();
  }
  
  /***************************************************************************
   * Get port number on which server is listening.
   * @return the port number.
   **************************************************************************/
  public int getPortNumber()
  {
    return serverSocket.getLocalPort();
  }
  
  /****************************************************************************
   * Server listens on socket and executes received commands.
   ***************************************************************************/
  public synchronized void run()
  {
    while (true)
    {
      try
      {
        Socket socket = serverSocket.accept();
        BufferedReader reader = 
          new BufferedReader(new InputStreamReader(socket.getInputStream()));
        PrintWriter writer =
          new PrintWriter(new OutputStreamWriter(socket.getOutputStream(), charSet));
        int counter = 0;
        String request = null;
        while (true)
        {
          String line = reader.readLine();
          if (line == null)
          {
            System.out.println("Empty string read from socket.");
            Main.exit(-1);
          }
          if (counter == 0) request = line; 
          counter = counter+1;
          if (line.equals("")) break;
        }
        boolean okay = request.startsWith("GET /");
        if (!okay) 
          Main.internalError("unknown request: " + request);
        int slash = request.indexOf('/', 5);
        if (slash == -1)
          Main.internalError("/ expected in request: " + request);
        int end = request.indexOf(' ', slash+1);
        if (end == -1)
          Main.internalError("space expected in request: " + request);
        request = request.substring(0, end);
        request = URLDecoder.decode(request, "UTF-8");
        String directory = request.substring(5, slash);
        String file = request.substring(slash+1, end);
        serve(directory, file, writer);
        writer.close();
      }
      catch(IOException e)
      {
        System.out.println("Error in reading from socket: " + e);
        Main.exit(-1);
      }
    }
  }
  
  /***************************************************************************
   * Serve requested file.
   * @param directory the name of the virtual directory.
   * @param file the name of the file.
   * @param writer the writer to which the file is to be served. 
   **************************************************************************/
  private void serve(String directory, String file, PrintWriter writer)
  {
    // determine string
    VirtualDirectory dir = directories.get(directory);
    if (dir == null) Main.internalError("unknown directory: " + directory);
    String result = dir.read(file);
    if (result == null) Main.internalError("unknown file: " + file);
    
    // write header
    writer.println("HTTP/1.0 200 OK");
    writer.println("Date: " + dateFormat.format(new Date()));
    writer.println("Content-Type: application/xhtml+xml");
    writer.println();
    writer.println(result);
  }
}

