// ----------------------------------------------------------------------------
// ExpressionLogic.java
// The judgements "E ~ T" and "E ~ F".
// $Id: ExpressionLogic.java,v 1.41 2013/01/23 15:10:28 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import fmrisc.ProgramExplorer.Main;
import fmrisc.ProgramExplorer.Util;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Semantics.*;

import java.util.*;

/*
 * 
 * The judgements "E ~ T" and "E ~ F".
 *
 */
public final class ExpressionLogic
{  
  // the base theory
  private final TheorySymbol baseTheory;
  
  // the mathematical functions corresponding to program constants/operations
  private final ValueSymbol minIntSymbol;
  private final ValueSymbol maxIntSymbol;
  private final ValueSymbol intPlusSymbol;
  private final ValueSymbol intNegSymbol;
  private final ValueSymbol intMinusSymbol;
  private final ValueSymbol intTimesSymbol;
  private final ValueSymbol intDivSymbol;
  private final ValueSymbol intDivSymbol0;
  private final ValueSymbol intRemSymbol;
  // private final ValueSymbol nullIntArraySymbol;
  // private final ValueSymbol nullCharArraySymbol;
  // private final ValueSymbol nullBooleanArraySymbol;
  // private final ValueSymbol nullStringArraySymbol;
  // private final ValueSymbol nullStringSymbol;
  
  // the basic types
  private Type intType;
  private Type booleanType;
  
  // true, if base functions without preconditions are to be applied
  public static boolean baseFunctions = false;
  
  // the logical constants TRUE and FALSE
  private static True tvalue = True.construct();
  private static False fvalue = False.construct();
  private static BitTrue bittvalue = BitTrue.construct();
  private static BitFalse bitfvalue = BitFalse.construct();
  static
  {
    Type btype = BooleanType.construct();
    tvalue.setType(btype);
    fvalue.setType(btype);
    Type bittype = BitType.construct();
    bittvalue.setType(bittype);
    bitfvalue.setType(bittype);
  }
  
  /***************************************************************************
   * Create judgement processor.
   * @param baseTheory the base theory
   **************************************************************************/
  public ExpressionLogic(TheorySymbol baseTheory)
  {
    this.baseTheory = baseTheory;
    SymbolTable<ValueSymbol> vtable = baseTheory.getValues();
    minIntSymbol = vtable.get(Identifier.construct(TypeTranslator.minIntName));
    maxIntSymbol = vtable.get(Identifier.construct(TypeTranslator.maxIntName));
    intPlusSymbol = vtable.get(Identifier.construct(TypeTranslator.intPlusName));
    intNegSymbol = vtable.get(Identifier.construct(TypeTranslator.intNegName));
    intMinusSymbol = vtable.get(Identifier.construct(TypeTranslator.intMinusName));
    intTimesSymbol = vtable.get(Identifier.construct(TypeTranslator.intTimesName));
    intDivSymbol = vtable.get(Identifier.construct(TypeTranslator.intDivName));
    intDivSymbol0 = vtable.get(Identifier.construct(TypeTranslator.intDivName0));
    intRemSymbol = vtable.get(Identifier.construct(TypeTranslator.intRemName));
    // nullIntArraySymbol = vtable.get(Identifier.construct(TypeTranslator.nullIntArrayName));
    // nullCharArraySymbol = vtable.get(Identifier.construct(TypeTranslator.nullCharArrayName));
    // nullBooleanArraySymbol = vtable.get(Identifier.construct(TypeTranslator.nullBooleanArrayName));
    // nullStringArraySymbol = vtable.get(Identifier.construct(TypeTranslator.nullStringArrayName));
    // nullStringSymbol = vtable.get(Identifier.construct(TypeTranslator.nullStringName));
    intType = LogicTypeTable.intType;
    booleanType = LogicTypeTable.booleanType;
  }
  
  // --------------------------------------------------------------------------
  //
  // Public access functions
  // 
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Translate a program expression to an equivalent logic term.
   * @param exp the program expression.
   * @param pre the vector to which to add preconditions (may be null)
   * @return the equivalent logic term.
   **************************************************************************/
  public Term 
  deriveTerm(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp,
      Vector<Formula> pre)
  {
    Term result = new Translator(pre).deriveTerm(exp);
    return result;
  }
  
  /***************************************************************************
   * Translate a program expression to an equivalent logic formula.
   * @param exp the program expression.
   * @param pre the vector to which to add preconditions (may be null)
   * @return the equivalent logic formula.
   **************************************************************************/
  public Formula
  deriveFormula(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp,
    Vector<Formula> pre)
  {
    Formula result = new Translator(pre).deriveFormula(exp);
    return result;
  }
  
  /****************************************************************************
   * Convert program selector to one or two logic selectors.
   * @param base the logic form of the base expression.
   * @param sel the program selector.
   * @param pre the vector to which to add preconditions (may be null)
   * @return the corresponding logic selectors (null, if none, i.e. 
   *         selector can be dropped)
   ***************************************************************************/
  public Selector[] deriveSelector(Term base,
    fmrisc.ProgramExplorer.Syntax.Program.Selector sel,
    Vector<Formula> pre)
  {
    Translator translator = new Translator(pre);
    translator.baseTerm = base;
    Selector[] result = translator.deriveSelector(sel);
    return result;
  }
  
  // --------------------------------------------------------------------------
  //
  // Auxiliary functions
  // 
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Return reference to entity in base theory.
   * @param ident the name of the entity.
   * @return the reference to the entity using the long name.
   ***************************************************************************/
  private Reference baseReference(Identifier ident)
  {
    return theoryReference(baseTheory, ident);
  }
  
  /****************************************************************************
   * Return reference to entity in string theory.
   * @param ident the name of the entity.
   * @return the reference to the entity using the long name.
   ***************************************************************************/
  private Reference stringReference(Identifier ident)
  {
    TheorySymbol stringTheory = 
        baseTheory.getPackage().getTypeTranslator().getStringTheory();
    return theoryReference(stringTheory, ident);
  }
  
  /****************************************************************************
   * Return reference to entity in given theory.
   * @param ident the name of the entity.
   * @return the reference to the entity using the long name.
   ***************************************************************************/
  private Reference theoryReference(TheorySymbol tsymbol, Identifier ident)
  {
    Reference result = 
      Reference.construct(Name.construct(tsymbol.getLongName(), ident));
    Symbol symbol = ident.getSymbol();
    if (symbol instanceof ValueSymbol)
    {
      Type type = ((ValueSymbol)symbol).getType();
      result.setType(type);
    }
    return result;
  }
  
  /****************************************************************************
   * Construct function application.
   * @param fun the function.
   * @param args the arguments.
   * @return the application.
   ***************************************************************************/
  private FunctionApplication functionApplication(Reference fun, Term[] args)
  {
    FunctionApplication result = FunctionApplication.construct(fun, args);
    ValueSymbol symbol = (ValueSymbol)fun.getName().getSymbol();
    FunctionType type = (FunctionType)symbol.getType();
    fun.setType(type);
    result.setType(type.getRange());
    return result;
  }
    
  /***************************************************************************
   * Convert boolean to constant term.
   * @param value a boolean value.
   * @return the corresponding constant term.
   **************************************************************************/
  public static Term booleanTerm(boolean value)
  {
    if (Main.firstOrder())
      return bitTerm(value);
    else
      return booleanFormula(value);
  }
  
  /***************************************************************************
   * Convert boolean to truth formula.
   * @param value a boolean value.
   * @return the corresponding truth formula.
   **************************************************************************/
  public static BooleanLiteral booleanFormula(boolean value)
  {
    return value ? tvalue : fvalue;
  }
  
  /***************************************************************************
   * Convert boolean to bit term.
   * @param value a boolean value.
   * @return the corresponding bit term.
   **************************************************************************/
  private static BitLiteral bitTerm(boolean value)
  {
    return value ? bittvalue : bitfvalue;
  }
  
  /***************************************************************************
   * Construct integer literal.
   * @param digits the digits of the literal.
   * @return the literal.
   **************************************************************************/
  public static IntLiteral intLiteral(String digits)
  {
    IntLiteral result = IntLiteral.construct(digits);
    result.setType(LogicTypeTable.intType);
    return result;
  }
  
  /****************************************************************************
   * Get type of value to which selector is applied.
   * @param sel the selector.
   * @return the base type.
   ***************************************************************************/
  private Type getBaseType(fmrisc.ProgramExplorer.Syntax.Program.Selector sel)
  {
    Type btype = sel.getLogicBaseType();
    if (btype instanceof Reference)
    {
      Reference btype0 = (Reference)btype;
      TypeSymbol symbol = (TypeSymbol)btype0.getName().getSymbol();
      btype = symbol.getType();
    }
    return btype;
  }
  
  /***************************************************************************
   * Derive identifier selector that selects from record type by ident.
   * @param base the base record type.
   * @param ident the identifier to be used for selection.
   * @return the logical selector.
   **************************************************************************/
  private static Selector identSelector(RecordType base, Identifier ident)
  {
    TypedIdentifier[] tids = base.getBase();
    int n = tids.length;
    String string = ident.getString();
    for (int i=0; i<n; i++)
    {
      TypedIdentifier tid = tids[i];
      Identifier id = tid.getName();
      if (string == id.getString())
      {
        if (id.getSymbol() == null)
          Main.internalError("type " + base);
        return SelectorIdentifier.construct(id);
      }
    }
    Main.internalError("error in processing selector");
    return null;
  }
  
  /*
   * 
   * Translate expression and store generated preconditions
   *
   */
  private final class Translator
  {
    // the constructed preconditions
    private Vector<Formula> preconditions;
    
    // the formulas that must be true such that the expression is evaluated
    private Vector<Formula> context; 
    
    /**************************************************************************
     * Generate translator that adds preconditions to denoted vector.
     * @param preconditions the preconditions to be used in translation
     *        (may be also null)
     *************************************************************************/
    public Translator(Vector<Formula> preconditions)
    {
      this.preconditions = preconditions;
      this.context = new Vector<Formula>();
    }

    /*************************************************************************
     * Add precondition to precondition vector (if not null).
     * @param pre the precondition.
     ************************************************************************/
    private void addPrecondition(Formula pre)
    {
      if (preconditions == null) return;
      Formula c = null;
      for (Formula f : context)
      {
        if (c == null)
          c = f;
        else
          c = AndFormula.construct(c, f);
      }
      if (c == null)
        preconditions.add(pre);
      else
        preconditions.add(ImpliesFormula.construct(c, pre));
    }
    
    /***************************************************************************
     * Translate a program expression to an equivalent logic term.
     * @param exp the program expression.
     * @return the equivalent logic term.
     **************************************************************************/
    public Term 
    deriveTerm(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp)
    {
      Expression exp0 = derive(exp);
      if (exp0 instanceof False) return booleanTerm(false);
      if (exp0 instanceof True) return booleanTerm(true);
      if (exp0 instanceof Term) return (Term)exp0;
      Term t = booleanTerm(true);
      Term f = booleanTerm(false);
      Term result = IfThenElseTerm.construct((Formula)exp0, t, f);
      result.setType(LogicTypeTable.booleanType);
      return result;
    }
    
    /***************************************************************************
     * Translate a program expression to an equivalent logic formula.
     * @param exp the program expression.
     * @return the equivalent logic formula.
     **************************************************************************/
    public Formula
    deriveFormula(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp)
    {
      Expression exp0 = derive(exp);
      if (!(exp0 instanceof Formula) ||
          (Main.firstOrder() && exp0 instanceof ProgramVariable))
      {
        return EqualsFormula.construct((Term)exp0, booleanTerm(true));
      }
      return (Formula)exp0;
    }
    
    /****************************************************************************
     * Convert program selector to one or two logic selectors.
     * @param sel the program selector.
     * @return the corresponding logic selectors (null, if none, i.e. 
     *         selector can be dropped)
     ***************************************************************************/
    private Selector[] deriveSelector(
      fmrisc.ProgramExplorer.Syntax.Program.Selector sel)
    {
      Selector[] sels = derive(sel);
      return sels;
    }
    
    /***************************************************************************
     * Translate a program expression to an equivalent logic expression.
     * @param exp the program expression.
     * @return the equivalent logic expression.
     **************************************************************************/
    public Expression derive(fmrisc.ProgramExplorer.Syntax.Program.ValueExpression exp)
    {
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.BinaryValueExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.BinaryValueExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.CharLiteral)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.CharLiteral)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.IntLiteral)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.IntLiteral)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NewArrayExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.NewArrayExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NullLiteral)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.NullLiteral)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.SelectorExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.SelectorExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.StringLiteral)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.StringLiteral)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.UnaryValueExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.UnaryValueExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableValue)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.VariableValue)exp);
      Main.internalError("unknown value expression: " + exp);
      return null; 
    }

    private Expression derive(fmrisc.ProgramExplorer.Syntax.Program.BinaryValueExpression exp)
    {
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.AndExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.AndExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.DividesExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.DividesExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.EqualsExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.EqualsExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.GreaterEqualExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.GreaterEqualExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.GreaterExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.GreaterExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.LessEqualExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.LessEqualExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.LessExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.LessExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.MinusExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.MinusExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NotEqualsExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.NotEqualsExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.OrExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.OrExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.PercentExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.PercentExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.PlusExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.PlusExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.TimesExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.TimesExpression)exp);
      Main.internalError("unknown binary value expression" + exp);
      return null;
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.AndExpression exp)
    {
      Formula f1 = deriveFormula(exp.first);
      context.add(f1);
      Formula f2 = deriveFormula(exp.second);
      context.removeElementAt(context.size()-1);
      return AndFormula.construct(f1, f2);
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.DividesExpression exp)
    {
      Term first = deriveTerm(exp.first);
      Term second = deriveTerm(exp.second);
      if (baseFunctions)
        return functionApplication(
            baseReference(intDivSymbol.getName()),
            new Term[] { first, second } );
      Term result = 
        functionApplication(
            baseReference(intDivSymbol0.getName()),
            new Term[] { first, second });
      result.setType(intType);
      Term zero = IntLiteral.construct("0");
      zero.setType(intType);
      Formula pre =
        AndFormula.construct(
            NotEqualsFormula.construct(second, zero),
            AndFormula.construct(
                LessEqualFormula.construct(
                    baseReference(minIntSymbol.getName()), result),
                LessEqualFormula.construct(
                    result, baseReference(maxIntSymbol.getName()))));
      addPrecondition(pre);
      return result;
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.EqualsExpression exp)
    {
      if (exp.second instanceof fmrisc.ProgramExplorer.Syntax.Program.NullLiteral)
        return nullFormula(deriveTerm(exp.first));
      return EqualsFormula.construct(deriveTerm(exp.first), deriveTerm(exp.second));
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.GreaterEqualExpression exp)
    {
      return GreaterEqualFormula.construct(deriveTerm(exp.first), deriveTerm(exp.second));
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.GreaterExpression exp)
    {
      return GreaterFormula.construct(deriveTerm(exp.first), deriveTerm(exp.second));
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.LessEqualExpression exp)
    {
      return LessEqualFormula.construct(deriveTerm(exp.first), deriveTerm(exp.second));
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.LessExpression exp)
    {
      return LessFormula.construct(deriveTerm(exp.first), deriveTerm(exp.second));
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.MinusExpression exp)
    {
      Term first = deriveTerm(exp.first);
      Term second = deriveTerm(exp.second);
      if (baseFunctions)
        return functionApplication(
            baseReference(intMinusSymbol.getName()),
            new Term[] { first, second } );
      Term result = MinusTerm.construct(first, second);
      result.setType(intType);
      Formula pre =
        AndFormula.construct(
            LessEqualFormula.construct(
                baseReference(minIntSymbol.getName()), result),
            LessEqualFormula.construct(
                result, baseReference(maxIntSymbol.getName())));
      addPrecondition(pre);
      return result;
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.NotEqualsExpression exp)
    {
      if (exp.second instanceof fmrisc.ProgramExplorer.Syntax.Program.NullLiteral)
        return notNullFormula(deriveTerm(exp.first));
      return NotEqualsFormula.construct(deriveTerm(exp.first), deriveTerm(exp.second));
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.OrExpression exp)
    {
      Formula f1 = deriveFormula(exp.first);
      context.add(NotFormula.construct(f1));
      Formula f2 = deriveFormula(exp.second);
      context.removeElementAt(context.size()-1);
      return OrFormula.construct(f1, f2);
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.PercentExpression exp)
    {
      Term first = deriveTerm(exp.first);
      Term second = deriveTerm(exp.second);
      if (baseFunctions)
        return functionApplication(
            baseReference(intRemSymbol.getName()),
            new Term[] { first, second } );
      Term result = 
        functionApplication(
            baseReference(intDivSymbol0.getName()),
            new Term[] { first, second });
      result.setType(intType);
      result = TimesTerm.construct(result, second);
      result.setType(intType);
      result = MinusTerm.construct(first, result);
      result.setType(intType);
      return result;
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.PlusExpression exp)
    {
      Term first = deriveTerm(exp.first);
      Term second = deriveTerm(exp.second);
      if (baseFunctions)
        return functionApplication(
            baseReference(intPlusSymbol.getName()),
            new Term[] { first, second } );
      Term result = PlusTerm.construct(first, second);
      result.setType(intType);
      Formula pre =
        AndFormula.construct(
            LessEqualFormula.construct(
                baseReference(minIntSymbol.getName()), result),
            LessEqualFormula.construct(
                result, baseReference(maxIntSymbol.getName())));
      addPrecondition(pre);
      return result;
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.TimesExpression exp)
    {
      Term first = deriveTerm(exp.first);
      Term second = deriveTerm(exp.second);
      if (baseFunctions)
        return functionApplication(
            baseReference(intTimesSymbol.getName()),
            new Term[] { first, second } );
      Term result = TimesTerm.construct(first, second);
      result.setType(intType);
      Formula pre =
        AndFormula.construct(
            LessEqualFormula.construct(
                baseReference(minIntSymbol.getName()), result),
            LessEqualFormula.construct(
                result, baseReference(maxIntSymbol.getName())));
      addPrecondition(pre);
      return result;
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.BooleanLiteral exp)
    {
      return booleanFormula(exp.value);
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.CharLiteral exp)
    {
      int code = (int)exp.value;
      return intLiteral(String.valueOf(code));
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.IntLiteral exp)
    {
      return intLiteral(exp.digits);
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.NewArrayExpression exp)
    {
      Term length = deriveTerm(exp.length);
      Formula pre = GreaterEqualFormula.construct(length, intLiteral("0"));
      addPrecondition(pre);
      fmrisc.ProgramExplorer.Syntax.Program.TypeExpression type = exp.type;
      ValueSymbol newArraySymbol = null;
      if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType)
      {
        fmrisc.ProgramExplorer.Syntax.Program.NamedType type0 =
          (fmrisc.ProgramExplorer.Syntax.Program.NamedType)type;
        ClassSymbol csymbol = type0.getSymbol();
        TheorySymbol tsymbol = csymbol.getClassTheory();
        SymbolTable<ValueSymbol> vtable = tsymbol.getValues();
        newArraySymbol = 
          vtable.get(Identifier.construct(TypeTranslator.newArrayName));
        return functionApplication(
            Reference.construct(Name.construct(tsymbol.getLongName(),
                newArraySymbol.getName())),
                new Term[] { length });
      }
      String name = null;
      if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.IntType)
        name = TypeTranslator.newIntArrayName;
      else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanType)
        name = TypeTranslator.newBooleanArrayName;
      else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.CharType)
        name = TypeTranslator.newCharArrayName;
      else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
        name = TypeTranslator.newStringArrayName;
      else
        Main.internalError("unknown base type of new array encountered");
      if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
      {
        TheorySymbol stringTheory = 
            baseTheory.getClassSymbol().getPackage().getTypeTranslator().getBaseTheory();
        SymbolTable<ValueSymbol> vtable = stringTheory.getValues();
        newArraySymbol = vtable.get(Identifier.construct(name));
        return functionApplication(
            stringReference(newArraySymbol.getName()),
            new Term[] { length });
      }
      else
      {
        SymbolTable<ValueSymbol> vtable = baseTheory.getValues();
        newArraySymbol = vtable.get(Identifier.construct(name));
        return functionApplication(
            baseReference(newArraySymbol.getName()),
            new Term[] { length });
      }
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.NullLiteral exp)
    {
      fmrisc.ProgramExplorer.Syntax.Program.TypeExpression type = exp.getType();
      boolean isArrayType = type instanceof fmrisc.ProgramExplorer.Syntax.Program.ArrayType;
      if (isArrayType)
      {
        fmrisc.ProgramExplorer.Syntax.Program.ArrayType type0 =
          (fmrisc.ProgramExplorer.Syntax.Program.ArrayType)type;
        type = type0.base;
        if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.ArrayType)
          Main.internalError("nested array type encountered");
        if (!(type instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType))
        {
          String name = null;
          if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.IntType)
            name = TypeTranslator.nullIntArrayName;
          else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanType)
            name = TypeTranslator.nullBooleanArrayName;
          else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.CharType)
            name = TypeTranslator.nullCharArrayName;
          else if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
            name = TypeTranslator.nullStringArrayName;
          else
            Main.internalError("unknown base type of array encountered");
          if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
          {
            TheorySymbol stringTheory = 
                baseTheory.getClassSymbol().getPackage().getTypeTranslator().getBaseTheory();
            SymbolTable<ValueSymbol> vtable = stringTheory.getValues();
            ValueSymbol nullArraySymbol = vtable.get(Identifier.construct(name));
            return stringReference(nullArraySymbol.getName());
          }
          else
          {
            SymbolTable<ValueSymbol> vtable = baseTheory.getValues();
            ValueSymbol nullArraySymbol = vtable.get(Identifier.construct(name));
            return baseReference(nullArraySymbol.getName());
          }
        }
      }
      if (type instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType)
      {
        fmrisc.ProgramExplorer.Syntax.Program.NamedType type0 = 
          (fmrisc.ProgramExplorer.Syntax.Program.NamedType)type;
        ClassSymbol csymbol = type0.getSymbol();
        TheorySymbol tsymbol = csymbol.getClassTheory();
        SymbolTable<ValueSymbol> vtable = tsymbol.getValues();
        Reference ref;
        if (isArrayType)
        {
          ValueSymbol nullArraySymbol = 
            vtable.get(Identifier.construct(TypeTranslator.nullArrayName));
          ref = Reference.construct(
              Name.construct(tsymbol.getLongName(), nullArraySymbol.getName()));
        }
        else
        {
          ValueSymbol nullSymbol = 
            vtable.get(Identifier.construct(TypeTranslator.nullName));
          ref = Reference.construct(
              Name.construct(tsymbol.getLongName(), nullSymbol.getName()));
        }
        ref.setType(booleanType);
        return ref;
      }
      Main.internalError("unexpected null type encountered: " + type);
      return null;
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.SelectorExpression exp)
    {
      int n = exp.sels.length;

      // determine true base expression and position of first true selector
      Term term = null;
      fmrisc.ProgramExplorer.Syntax.Program.TypeExpression type0 = null;
      int pos = 0;
      if (!(exp.base instanceof fmrisc.ProgramExplorer.Syntax.Program.VariableValue))
      {
        term = deriveTerm(exp.base);
        type0 = exp.getBaseType();
      }
      else
      {
        // name has form (Package.)*[Class.]variable
        fmrisc.ProgramExplorer.Syntax.Program.VariableValue base0 =
          (fmrisc.ProgramExplorer.Syntax.Program.VariableValue)exp.base;
        Identifier ident = base0.ident;
        Symbol symbol = ident.getSymbol();
        if (symbol instanceof ObjectVariableSymbol)
        {
          ObjectVariableSymbol symbol0 = (ObjectVariableSymbol)symbol;
          /*
          // prefix unqualified reference to object variable with "this" 
          ClassSymbol csymbol = symbol0.getClassSymbol();
          TheorySymbol tsymbol = csymbol.getClassTheory();
          SymbolTable<TypeSymbol> ttable = tsymbol.getTypes();
          Identifier tident = Identifier.construct(csymbol.getName().getString());
          TypeSymbol tysymbol = ttable.get(tident);
          RecordType rtype = (RecordType)tysymbol.getType();
          Selector sel = identSelector(rtype, ident);
          Identifier thisIdent = csymbol.getThisSymbol().getName();
          term = SelectionTerm.construct(Reference.construct(thisIdent), sel);
          */
          term = PreVariable.construct(NameUtils.uniqueName(symbol0));
          term.setType(symbol0.getLogicType());
          type0 = symbol0.getType();
        }
        else if (symbol instanceof VariableSymbol)
        {
          // variable is no object variable
          term = deriveTerm(base0);
          type0 = ((VariableSymbol)symbol).getType();
        }
        else
        {
          // name has form (Package.)*Class.variable
          while (pos < n)
          {
            if (!(exp.sels[pos] instanceof fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector))
              break;
            fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector sel0 =
              (fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector)exp.sels[pos];
            ident = sel0.name;
            symbol = sel0.name.getSymbol();
            pos = pos+1;
            if (symbol instanceof ClassVariableSymbol) break;
          }
          if (!(symbol instanceof ClassVariableSymbol))
            Main.internalError("no class variable symbol found");
          ClassVariableSymbol symbol0 = (ClassVariableSymbol)symbol;
          term = PreVariable.construct(symbol0.getLongName());
          term.setType(symbol0.getLogicType());
          type0 = symbol0.getType();
        }
      }
      
      // process true selectors 
      Type type = term.getType();
      for (int i=pos; i<n; i++)
      {
        if (type0 instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType)
        {
          addPrecondition(notNullFormula(term));
          /*
          fmrisc.ProgramExplorer.Syntax.Program.NamedType type1 =
            (fmrisc.ProgramExplorer.Syntax.Program.NamedType)type0;
          TheorySymbol tsymbol = type1.getSymbol().getClassTheory();
          SymbolTable<ValueSymbol> vtable = tsymbol.getValues();
          ValueSymbol nullSymbol = 
            vtable.get(Identifier.construct(TypeTranslator.nullName));
          addPrecondition(NotEqualsFormula.construct(term,
              theoryReference(tsymbol, nullSymbol.getName())));
          */
        }
        else if (type0 instanceof fmrisc.ProgramExplorer.Syntax.Program.ArrayType)
        {
          addPrecondition(notNullFormula(term));
          /*
          fmrisc.ProgramExplorer.Syntax.Program.ArrayType type1 =
            (fmrisc.ProgramExplorer.Syntax.Program.ArrayType)type0;
          if (type1.base instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType)
          {
            fmrisc.ProgramExplorer.Syntax.Program.NamedType type2 =
              (fmrisc.ProgramExplorer.Syntax.Program.NamedType)type1.base;
            TheorySymbol tsymbol = type2.getSymbol().getClassTheory();
            SymbolTable<ValueSymbol> vtable = tsymbol.getValues();
            ValueSymbol nullSymbol = 
              vtable.get(Identifier.construct(TypeTranslator.nullArrayName));
            addPrecondition(NotEqualsFormula.construct(term,
                theoryReference(tsymbol, nullSymbol.getName())));
          }
          else if (type1.base instanceof fmrisc.ProgramExplorer.Syntax.Program.IntType)
          {
            addPrecondition(NotEqualsFormula.construct(term,
                baseReference(nullIntArraySymbol.getName())));
          }
          else if (type1.base instanceof fmrisc.ProgramExplorer.Syntax.Program.CharType)
          {
            addPrecondition(NotEqualsFormula.construct(term,
                baseReference(nullCharArraySymbol.getName())));
          }
          else if (type1.base instanceof fmrisc.ProgramExplorer.Syntax.Program.BooleanType)
          {
            addPrecondition(NotEqualsFormula.construct(term,
                baseReference(nullBooleanArraySymbol.getName())));
          }
          else if (type1.base instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
          {
            addPrecondition(NotEqualsFormula.construct(term,
                baseReference(nullStringArraySymbol.getName())));
          }
          else
            Main.internalError("Unknown base type: " + type1.base);
          */
        }
        else if (type0 instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
        {
          TheorySymbol stringTheory = 
              baseTheory.getClassSymbol().getPackage().getTypeTranslator().getBaseTheory();
          SymbolTable<ValueSymbol> vtable = stringTheory.getValues();
          ValueSymbol nullStringSymbol = 
              vtable.get(Identifier.construct(TypeTranslator.nullStringName));
          addPrecondition(NotEqualsFormula.construct(term,
              stringReference(nullStringSymbol.getName())));
        }
        else
          Main.internalError("unexpected type: " + type0);
        baseTerm = term;
        Selector[] sels = deriveSelector(exp.sels[i]);
        if (sels != null) 
        { 
          int m = sels.length;
          for (int j=0; j<m; j++)
          {
            Selector sel = sels[j];
            term = SelectionTerm.construct(term, sel);
            if (sel instanceof SelectorIndex)
            {
              ArrayType type1 = (ArrayType)type;
              type = type1.getBase();
            }
            else if (sel instanceof SelectorNumber)
            {
              SelectorNumber sel0 = (SelectorNumber)sel;
              TupleType type1 = (TupleType)type;
              Type[] base = type1.getBase();
              int number = Integer.parseInt(sel0.getNumber().getName());
              type = base[number];
            }
            else if (sel instanceof SelectorIdentifier)
            {
              SelectorIdentifier sel0 = (SelectorIdentifier)sel;
              ValueSymbol symbol = (ValueSymbol)sel0.getIdentifier().getSymbol();
              type = symbol.getType();
            }
            else 
              Main.internalError("unknown selector: " + sel);
            term.setType(type);
          }
        }
        type0 = exp.sels[i].getType();
      }
      return term;
    }
    
    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.StringLiteral exp)
    {
      long code = Util.hash(exp.value);
      fmrisc.ProgramExplorer.Syntax.Program.TypeExpression stringType =
        baseTheory.getTypeTable().getStringType();
      if (stringType instanceof fmrisc.ProgramExplorer.Syntax.Program.StringType)
      {
        TheorySymbol stringTheory = 
            baseTheory.getClassSymbol().getPackage().getTypeTranslator().getBaseTheory();
        SymbolTable<ValueSymbol> vtable = stringTheory.getValues();
        ValueSymbol symbol = vtable.get(Identifier.construct(TypeTranslator.newStringName));
        return functionApplication(
            stringReference(symbol.getName()),
            new Term[] { intLiteral(Long.toString(code)) });
      }
      if (!(stringType instanceof fmrisc.ProgramExplorer.Syntax.Program.NamedType))
        Main.internalError("invalid string type encountered");
      fmrisc.ProgramExplorer.Syntax.Program.NamedType stringType0 =
        (fmrisc.ProgramExplorer.Syntax.Program.NamedType)stringType;
      ClassSymbol csymbol = stringType0.getSymbol();
      TheorySymbol tsymbol = csymbol.getClassTheory();
      SymbolTable<ValueSymbol> vtable = tsymbol.getValues();
      ValueSymbol symbol = vtable.get(Identifier.construct(TypeTranslator.newObjectName));
      return functionApplication(
          Reference.construct(Name.construct(tsymbol.getLongName(), symbol.getName())),
          new Term[] { intLiteral(Long.toString(code)) });
    }

    private Expression derive(fmrisc.ProgramExplorer.Syntax.Program.UnaryValueExpression exp)
    {
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NegationExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.NegationExpression)exp);
      if (exp instanceof fmrisc.ProgramExplorer.Syntax.Program.NotExpression)
        return derive((fmrisc.ProgramExplorer.Syntax.Program.NotExpression)exp);
      Main.internalError("unknown unary value expression");
      return null;
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.NegationExpression exp)
    {
      Term base = deriveTerm(exp.base);
      if (baseFunctions)
        return functionApplication(
            baseReference(intNegSymbol.getName()),
            new Term[] { base } );
      Term result = NegationTerm.construct(base);
      result.setType(intType);
      Formula pre =
        AndFormula.construct(
            LessEqualFormula.construct(
                baseReference(minIntSymbol.getName()), result),
            LessEqualFormula.construct(
                result, baseReference(maxIntSymbol.getName())));
      addPrecondition(pre);
      return result;
    }

    private Formula derive(fmrisc.ProgramExplorer.Syntax.Program.NotExpression exp)
    {
      return NotFormula.construct(deriveFormula(exp.base));
    }

    private Term derive(fmrisc.ProgramExplorer.Syntax.Program.VariableValue exp)
    {
      Name name = NameUtils.uniqueName(Name.construct(exp.ident));
      Term result = PreVariable.construct(name);
      VariableSymbol symbol = (VariableSymbol)exp.ident.getSymbol();
      result.setType(symbol.getLogicType());
      return result;
    }
  
    // the baseTerm for which the selector is constructed
    private Term baseTerm;
    
    /****************************************************************************
     * Convert program selector to one or two logic selectors.
     * Assumes that "baseTerm" is set to base expression;
     * sets "preCond" to precondition of translation (or null, if none).
     * @param sel the program selector.
     * @return the corresponding logic selectors (null, if none, i.e. 
     *         selector can be dropped)
     ***************************************************************************/
    private Selector[] derive(fmrisc.ProgramExplorer.Syntax.Program.Selector sel)
    {
      if (sel instanceof fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector)
      {
        fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector sel0 =
          (fmrisc.ProgramExplorer.Syntax.Program.IdentifierSelector) sel;
        Identifier ident = sel0.name;
        RecordType btype = (RecordType)getBaseType(sel);
        addPrecondition(notNullFormula(baseTerm));
        return new Selector[] { identSelector(btype, ident) };
      }
      if (sel instanceof fmrisc.ProgramExplorer.Syntax.Program.IndexSelector)
      {
        fmrisc.ProgramExplorer.Syntax.Program.IndexSelector sel0 =
          (fmrisc.ProgramExplorer.Syntax.Program.IndexSelector) sel;
        Term index = deriveTerm(sel0.index);
        RecordType btype = (RecordType)getBaseType(sel);
        Identifier ident = Identifier.construct(TypeTranslator.arrayValue);
        Identifier lident = Identifier.construct(TypeTranslator.arrayLength);
        Term lengthTerm = SelectionTerm.construct(baseTerm, 
            identSelector(btype, lident));
        lengthTerm.setType(intType);
        addPrecondition(
            AndFormula.construct(
                notNullFormula(baseTerm),
                AndFormula.construct(
                    LessEqualFormula.construct(intLiteral("0"), index),
                    LessFormula.construct(index, lengthTerm))));
        return new Selector[]{ identSelector(btype, ident), 
            SelectorIndex.construct(index) };
      }
      if (sel instanceof fmrisc.ProgramExplorer.Syntax.Program.MessageSelector)
      {
        // selector can be dropped (exception is interpreted as its content)
        return null;
      }
      Main.internalError("unknown selector");
      return null;
    }
  }
  
  /***************************************************************************
   * Create formula that term of object/array type is null.
   * @param term the term.
   * @return the formula.
   **************************************************************************/
  public Formula nullFormula(Term term)
  {
    return nullFormula(term, booleanType);
  }
  
  /***************************************************************************
   * Create formula that term of object/array type is not null.
   * @param term the term.
   * @return the formula.
   **************************************************************************/
  public Formula notNullFormula(Term term)
  {
    return notNullFormula(term, booleanType);
  }
  
  /***************************************************************************
   * Create formula that term of object/array type is null.
   * @param term the term.
   * @param booleanType the boolean type to be used.
   * @return the formula.
   **************************************************************************/
  public static Formula nullFormula(Term term, Type booleanType)
  {
    Type type = term.getType();
    while (type instanceof Reference)
    {
      Reference type0 = (Reference)type;
      TypeSymbol symbol = (TypeSymbol)type0.getName().getSymbol();
      type = symbol.getType();
    }
    if (type instanceof RecordType)
    {
      RecordType type0 = (RecordType)type;
      for (TypedIdentifier tid : type0.getBase())
      {
        Identifier ident = tid.getName();
        if (!ident.getString().equals(TypeTranslator.isNullName)) continue;
        SelectionTerm nullTerm = 
          SelectionTerm.construct(term, SelectorIdentifier.construct(ident));
        if (Main.firstOrder())
        {
          nullTerm.setType(BitType.construct());
          Term trueValue = BitTrue.construct();
          trueValue.setType(BitType.construct());
          Formula nullFormula = EqualsFormula.construct(nullTerm, trueValue);
          return nullFormula;
        }
        else
        {
          nullTerm.setType(booleanType);
          return nullTerm;
        }
      }
    }
    Main.internalError("unexpected type " + type);
    return null;
  }
  
  /***************************************************************************
   * Create formula that term of object/array type is not null.
   * @param term the term.
   * @param booleanType the boolean type to be used.
   * @return the formula.
   **************************************************************************/
  public static Formula notNullFormula(Term term, Type booleanType)
  {
    Type type = term.getType();
    while (type instanceof Reference)
    {
      Reference type0 = (Reference)type;
      TypeSymbol symbol = (TypeSymbol)type0.getName().getSymbol();
      type = symbol.getType();
    }
    if (type instanceof RecordType)
    {
      RecordType type0 = (RecordType)type;
      for (TypedIdentifier tid : type0.getBase())
      {
        Identifier ident = tid.getName();
        if (!ident.getString().equals(TypeTranslator.isNullName)) continue;
        SelectionTerm nullTerm = 
          SelectionTerm.construct(term, SelectorIdentifier.construct(ident));
        if (Main.firstOrder())
        {
          nullTerm.setType(BitType.construct());
          Term falseValue = BitFalse.construct();
          falseValue.setType(BitType.construct());
          Formula notNullFormula = EqualsFormula.construct(nullTerm, falseValue);
          return notNullFormula;
        }
        else
        {
          nullTerm.setType(booleanType);
          Formula notNullFormula = NotFormula.construct(nullTerm);
          return notNullFormula;
        }
      }
    }
    Main.internalError("unexpected type " + type);
    return null;
  }
}
