// ----------------------------------------------------------------------------
// Propagator.java
// Propagate pre/postconditions through a statement.
// $Id: Propagator.java,v 1.17 2012/10/18 16:12:58 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Judgements.Substitution.Value;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Syntax.Spec.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * Propagate pre/postconditions through a statement.
 *
 */
public final class Propagator
{
  /***************************************************************************
   * Propagate condition through statement.
   * @param C a statement that is annotated with statement judgements.
   * @param statement a sub-statement of C.
   * @param formula the formula to be attached to statement.
   * @param pre true iff the formula is a precondition (postcondition else).
   * @param preMap a mapping of statements to preconditions;
   *        is extended such that its domain is the set of all
   *        sub-statements of C (including C).
   * @param postMap an empty mapping of statements to postconditions;
   *        is extended such that its domain is the set of all
   *        sub-statements of C (including C).
   * @param satisfies a judgement processor.
   **************************************************************************/
  public static void execute(Statement C, 
    Statement statement, Formula formula, boolean pre,
    Map<Statement, Formula> preMap, Map<Statement,Formula> postMap, 
    Satisfies satisfies)
  {
    // normalize formula to use pre-state variable references only
    formula = Normalization.perform(formula);
    if (pre)
      put(preMap, statement, formula);
    else
      put(postMap, statement, formula);
    execute(C, preMap, postMap, satisfies);
  }
  
  /***************************************************************************
   * Propagate conditions through statement.
   * @param C a statement that is annotated with statement judgements.
   * @param preMap a partial map of statements to preconditions;
   *        is extended such that its domain is the set of all
   *        sub-statements of C (including C)
   * @param postMap a partial map of statements to postconditions;
   *        is extended such that its domain is the set of all
   *        sub-statements of C (including C)
   * @param satisfies a judgement processor.
   **************************************************************************/
  private static void execute(Statement C,
    Map<Statement, Formula> preMap, Map<Statement,Formula> postMap, 
    Satisfies satisfies)
  {
    Propagator propagator = new Propagator(preMap, postMap, satisfies);
    propagator.process(C);
  }
  
  // --------------------------------------------------------------------------
  //
  // the propagator object
  // 
  // --------------------------------------------------------------------------
  
  // mappings of statements to pre-/post-conditions
  private Map<Statement, Formula> preMap;
  private Map<Statement,Formula> postMap;
  
  // the judgement processor
  private Satisfies satisfies;
  
  // the expression and type translator
  private ExpressionLogic logic;
  
  // the state type of the command
  private StateType stype;
  
  // pre/postcondition computers
  private Pre preComp;
  private Post postComp;
  
  /****************************************************************************
   * Construct a pre-/post-condition propagator.
   * @param preMap a partial mapping of statements to pre-conditions.
   * @param postMap a partial mapping of statements to post-conditions.
   * @param satisfies a judgement processor.
   ***************************************************************************/
  private Propagator(Map<Statement, Formula> preMap, 
    Map<Statement,Formula> postMap, Satisfies satisfies)
  {
    this.preMap = preMap;
    this.postMap = postMap;
    this.satisfies = satisfies;
    this.logic = satisfies.getExpressionLogic();
    this.stype = satisfies.getStateType();
    this.preComp = new Pre(this.stype);
    this.postComp = new Post(this.stype);
  }
  
  // --------------------------------------------------------------------------
  //
  // the process functions
  // 
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Propagate statement to complete pre-/post-condition map.
   * @param C the statement to be processed.
   ***************************************************************************/
  private void process(Statement C)
  {
    // determine given pre-/post-condition from map (each may be null)
    Formula pre = preMap.get(C);
    Formula post = postMap.get(C);

    // complete pre-/post-condition such that
    // either both or none are available
    if (pre != null)
    {
      Formula post0 = postComp.derive(C, pre);
      put(postMap, C, post0);
    }
    if (post != null)
    {
      Formula pre0 = preComp.derive(C, post);
      put(preMap, C, pre0);
    }

    // process compound (non-atomic) statements
    if (C instanceof BlockStatement)
      process((BlockStatement)C);
    else if (C instanceof ConditionalStatement)
      process((ConditionalStatement)C);
    else if (C instanceof ForLoopStatement)
      process((ForLoopStatement)C);
    else if (C instanceof WhileLoopStatement)
      process((WhileLoopStatement)C);
    else if (C instanceof TryCatchStatement)
      process((TryCatchStatement)C);
  }
  
  /****************************************************************************
   * Propagate statement to complete pre/postcondition map.
   * @param C the statement to be processed.
   ***************************************************************************/
  private void process(BlockStatement C)
  {
    // determine formula precondition (if any)
    Formula pre = preMap.get(C);
    int n = C.stats.length;
    
    // we work outside-in
    if (pre != null)
    {
      Formula pre0 = pre;
      for (int i=0; i<n; i++)
      {
        Statement stat = C.stats[i];
        if (stat instanceof VariableDeclaration)
        {
          VariableDeclaration stat0 = (VariableDeclaration)stat;
          VariableSymbol vsymbol = (VariableSymbol)stat0.getName().getSymbol();

          // construct logical variable for program variable
          Collection<String> used = UsedVariables.compute(pre0);
          String name = NewVariables.newVariable(vsymbol.getName().getString(), used);
          Identifier ident = Identifier.construct(name);
          Type type = vsymbol.getLogicType();
          TypedIdentifier var = TypedIdentifier.construct(ident, type);
          ValueSymbol symbol = ValueSymbol.construct(ident, type, var);
          Reference ref = Satisfies.reference(symbol.getName());

          // perform substitution of program variable by logical value
          Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
          map.put(vsymbol, 
              new Substitution.Value(
                  Substitution.Action.term, ref,
                  Substitution.Action.term, ref));
          pre0 = (Formula)Substitution.perform(pre0, null, null, map);
        }

        // set precondition of statement
        put(preMap, stat, pre0);

        // process statement
        process(stat);

        // postcondition of this statement is precondition of the next one
        pre0 = postMap.get(stat);
      }
      
      // we are done
      return;
    }
    
    // we work inside-out
    
    // determine first and last statement with attached condition
    // and determine the local variables of the block
    int first = -1;
    int last = -1;
    Vector<VariableSymbol> variables = new Vector<VariableSymbol>();
    for (int i=0; i<n; i++)
    {
      Statement stat = C.stats[i];
      process(stat);
      if (preMap.get(stat) != null)
      {
        if (first == -1) first = i;
        last = i;
      }
      if (stat instanceof VariableDeclaration)
      {
        VariableDeclaration stat0 = (VariableDeclaration)stat;
        variables.add((VariableSymbol)stat0.getName().getSymbol());
      }
    }
    
    // there may be no condition attached
    if (first == -1) return;
    
    // propagate conditions from last statement backwards
    for (int i=last-1; i>=0; i--)
    {
      Statement stat = C.stats[i];
      Formula pre0 = preMap.get(C.stats[i+1]);
      put(postMap, stat, pre0);
      process(stat);
    }
    // propagate conditions from first statement forwards
    for (int i=first+1; i<n; i++)
    {
      Statement stat = C.stats[i];
      Formula post0 = postMap.get(C.stats[i-1]);
      put(preMap, stat, post0);
      process(stat);
    }
    
    // compute pre-condition of block from first statement of block
    {
      pre = preMap.get(C.stats[0]);
      int vn = variables.size();
      VariableSymbol[] variables0 = new VariableSymbol[vn];
      variables.toArray(variables0);
      TypedIdentifier[] vars = new TypedIdentifier[vn];
      ValueSymbol[] symbols = new ValueSymbol[vn];
      Collection<String> used = UsedVariables.compute(pre);
      for (int i=0; i<vn; i++)
      {
        VariableSymbol variable = variables0[i];
        String name = NewVariables.newVariable(variable.getName().getString(), used);
        used.add(name);
        Identifier ident = Identifier.construct(name);
        Type type = variable.getLogicType();
        vars[i] = TypedIdentifier.construct(ident, type);
        symbols[i] = ValueSymbol.construct(ident, type, vars[i]);
      }
      Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
      for (int i=0; i<vn; i++)
      {
        Reference ref = Satisfies.reference(symbols[i].getName());
        map.put(variables0[i], 
            new Substitution.Value(
                Substitution.Action.term, ref,
                Substitution.Action.term, ref));
      }
      pre = (Formula)Substitution.perform(pre, null, null, map);
      pre = ForallFormula.construct(vars, pre);
    }
    
    // compute post-condition of block from last statement of block
    Formula post = postMap.get(C.stats[n-1]);
    int vn = variables.size();
    VariableSymbol[] variables0 = new VariableSymbol[vn];
    variables.toArray(variables0);
    TypedIdentifier[] vars = new TypedIdentifier[vn];
    ValueSymbol[] symbols = new ValueSymbol[vn];
    Collection<String> used = UsedVariables.compute(post);
    for (int i=0; i<vn; i++)
    {
      VariableSymbol variable = variables0[i];
      String name = NewVariables.newVariable(variable.getName().getString(), used);
      used.add(name);
      Identifier ident = Identifier.construct(name);
      Type type = variable.getLogicType();
      vars[i] = TypedIdentifier.construct(ident, type);
      symbols[i] = ValueSymbol.construct(ident, type, vars[i]);
    }
    Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<vn; i++)
    {
      Reference ref = Satisfies.reference(symbols[i].getName());
      map.put(variables0[i], 
          new Substitution.Value(
              Substitution.Action.term, ref,
              Substitution.Action.term, ref));
    }
    post = (Formula)Substitution.perform(post, null, null, map);
    post = ExistsFormula.construct(vars, post);
    
    // set pre-and post-condition of statement
    put(preMap,  C, pre);
    put(postMap, C, post);
  }
  
  /****************************************************************************
   * Propagate statement to complete pre/postcondition map.
   * @param C the statement to be processed.
   ***************************************************************************/
  private void process(ConditionalStatement C)
  {
    // determine formula precondition (if any)
    Formula pre = preMap.get(C);
    
    // determine formula condition (ignore generated preconditions)
    Vector<Formula> preconditions = new Vector<Formula>();
    Formula cond = logic.deriveFormula(C.cond, preconditions);
    
    // we work outside in
    if (pre != null)
    {
      put(preMap, C.thenstat, 
          AndFormula.construct(pre, cond));
      if (C.elsestat != null)
        put(preMap, C.elsestat, 
            AndFormula.construct(pre, NotFormula.construct(cond)));
    } 
    
    // process statement recursively
    process(C.thenstat);
    if (C.elsestat != null) 
      process(C.elsestat);
    
    // we work inside out
    if (pre == null)
    {
      // compute precondition
      Formula pre0 = preMap.get(C.thenstat);
      Formula pre1 = C.elsestat == null ? null : preMap.get(C.elsestat);
      if (pre0 != null) 
        pre0 = ImpliesFormula.construct(cond, pre0);
      if (pre1 != null)
        pre1 = ImpliesFormula.construct(NotFormula.construct(cond), pre1);
      pre = andFormula(pre0, pre1);
      put(preMap, C, pre);
      
      // compute postcondition
      Formula post0 = postMap.get(C.thenstat);
      Formula post1 = C.elsestat == null ? null : postMap.get(C.elsestat);
      Formula post = orFormula(post0, post1);
      put(postMap, C, post);
    }
  }
  
  /****************************************************************************
   * Propagate statement to complete pre/postcondition map.
   * @param C the statement to be processed.
   ***************************************************************************/
  private void process(ForLoopStatement C)
  {
    Formula pre = preMap.get(C);
    Formula post = postMap.get(C);
    preMap.put(C.whileReplacement, pre);
    postMap.put(C.whileReplacement, post);
    process(C.whileReplacement);
    pre = preMap.get(C.whileReplacement);
    post = postMap.get(C.whileReplacement);
    preMap.put(C, pre);
    postMap.put(C, post);
  }
  
  /****************************************************************************
   * Propagate statement to complete pre/postcondition map.
   * @param C the statement to be processed.
   ***************************************************************************/
  private void process(WhileLoopStatement C)
  {
    // determine formula precondition (if any)
    Formula pre = preMap.get(C);
    
    // we may work inside-out
    if (pre == null)
    {
      // process the body
      process(C.body);
      
      // the precondition of the loop is just "true"
      pre = ExpressionLogic.booleanFormula(true);
      put(preMap, C, pre);
      
      // its postcondition can be derived from the precondition
      Formula post = postComp.derive(C, pre);
      put(postMap, C, post);
      
      // we are done
      return;
    }
    
    // we work outside-in, thus we derive the precondition of the loop body

    // determine condition formula (ignore generated preconditions)
    Vector<Formula> preconditions = new Vector<Formula>();
    Formula cond = logic.deriveFormula(C.cond, preconditions);

    // determine body judgement
    StatJudgement cSpec = C.body.getJudgement();
    if (cSpec == null) Main.internalError("no judgement");

    // determine invariant 
    Formula invariant = null;
    LoopSpec loopSpec = (LoopSpec)C.getLoopAnnotation();
    if (loopSpec != null) invariant = loopSpec.getLoopInvariant();
    
    // if there is no invariant, we only have the loop condition
    if (invariant == null)
    {
      put(preMap, C.body, cond);
      process(C.body);
      return;
    }
    
    // otherwise normalize invariant and compute used variables
    invariant = Normalization.perform(invariant, cSpec.variables);
    Collection<String> used = UsedVariables.compute(invariant);

    // logical variables for program variables modified in loop body
    int n = cSpec.variables.size();
    VariableSymbol[] vsymbols = new VariableSymbol[n];
    cSpec.variables.toArray(vsymbols);
    String[] names = NewVariables.newVariables(vsymbols, "", used);
    UsedVariables.add(names, used);

    // logical variables for quantification in invariance goal
    TypedIdentifier[] vars = new TypedIdentifier[n+1];
    ValueSymbol[] symbols = new ValueSymbol[n+1];
    for (int i=0; i<n; i++)
    {
      Type type = vsymbols[i].getLogicType();
      Identifier ident = Identifier.construct(names[i]);
      TypedIdentifier var = TypedIdentifier.construct(ident, type);
      symbols[i] = ValueSymbol.construct(ident, type, var);
      vars[i] = var;
    }
    {
      String name = NewVariables.newVariable("s", used);
      used.add(name);
      Identifier ident = Identifier.construct(name);
      TypedIdentifier var = TypedIdentifier.construct(ident, stype);
      vars[n] = var;
      symbols[n] = ValueSymbol.construct(ident, stype, var);
    }

    // substitution map
    Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
    for (int i=0; i<n; i++)
    {
      Reference ref = Satisfies.reference(symbols[i].getName());
      map.put(vsymbols[i], 
          new Substitution.Value(
              Substitution.Action.term, ref,
              Substitution.Action.flip, null));
    }

    // consequent of implication (first part)
    Reference sref = Satisfies.reference(symbols[n].getName());
    Formula invariant1 = (Formula)Substitution.perform(invariant, 
        sref, satisfies.stateLiteral(true), map);

    // finally, the precondition of the loop body
    Formula formula = 
      AndFormula.construct(cond, Satisfies.existsFormula(vars, invariant1));
    put(preMap, C.body, formula);
    
    // process the body
    process(C.body);
  }
  
  /****************************************************************************
   * Propagate statement to complete pre/postcondition map.
   * @param C the statement to be processed.
   ***************************************************************************/
  private void process(TryCatchStatement C)
  {
    // determine formula precondition (if any)
    Formula pre = preMap.get(C);
    
    // outside-in operation
    if (pre != null) 
    {
      // precondition of try statement is same as precondition of body
      put(preMap, C.body, pre); 
    }
    
    // process body and determine its pre-/post-condition
    process(C.body);
    Formula pre0 = preMap.get(C.body);
    Formula post = postMap.get(C.body);
    
    // inside-out operation
    if (pre == null)
    {
      // precondition of body is same as precondition of try statement
      put(preMap, C, pre0);
      Formula post0 = postComp.derive(C, pre0);
      put(postMap, C, post0);
      
      // also process handlers inside-out
      for (Handler handler : C.handlers)
        process(handler.body);
      
      // we are done
      return;
    }
      
    // process handlers outside-in
    for (Handler handler : C.handlers)
    {
      // determine name for exception variable and state variable
      Collection<String> used = UsedVariables.compute(post);
      String exName = NewVariables.newVariable(handler.param.name.getString(), used);
      used.add(exName);
      String stateName = NewVariables.newVariable("s", used);
      used.add(stateName);

      // quantification variables
      TypedIdentifier[] vars = new TypedIdentifier[2];
      Identifier exIdent = Identifier.construct(exName);
      Type stringType = satisfies.getStringType();
      vars[0] = TypedIdentifier.construct(exIdent, stringType);
      ValueSymbol exSymbol = ValueSymbol.construct(exIdent, stringType, vars[0]);
      Identifier stateIdent = Identifier.construct(stateName);
      vars[1] = TypedIdentifier.construct(stateIdent, stype);
      ValueSymbol stateSymbol = ValueSymbol.construct(stateIdent, stype, vars[1]);

      // perform substitution in postcondition
      Map<VariableSymbol,Value> map = new MapClass<VariableSymbol,Value>();
      map.put((VariableSymbol)handler.param.name.getSymbol(), 
            new Substitution.Value(
                Substitution.Action.term, Satisfies.reference(exSymbol.getName()),
                Substitution.Action.none, null));
      Formula formula = (Formula)Substitution.perform(post,
          Satisfies.reference(stateSymbol.getName()), null, map);
          
      // add remaining condition
      fmrisc.ProgramExplorer.Syntax.Program.TypeExpression exType = handler.param.type;
      Name exTypeName = ((NamedType)exType).name;
      formula = AndFormula.construct(formula,
          AndFormula.construct(
              StateThrowsException.construct(
                  Satisfies.reference(stateSymbol.getName()),
                  exTypeName),
              EqualsFormula.construct(
                  satisfies.stateMessage(Satisfies.reference(stateSymbol.getName())),
                  Satisfies.reference(exSymbol.getName()))));
      formula = Satisfies.existsFormula(vars, formula);
                      
      // set precondition of handler body and process it
      put(preMap, handler.body, formula);
      process(handler.body);
    }
  }
  
  // --------------------------------------------------------------------------
  //
  // auxiliary functions
  // 
  // --------------------------------------------------------------------------
  
  /***************************************************************************
   * Update in given map formula for statement C by combination with new formula.
   * @param map the map.
   * @param C the key.
   * @param formula the new formula.
   **************************************************************************/
  private static void put(Map<Statement, Formula> map, Statement C, 
    Formula formula)
  {
    Formula formula0 = map.get(C);
    Formula formula1 = andFormula(formula0, formula);
    // do not simplify here but in presentation
    // formula1 = Simplification.perform(formula1);
    map.put(C, formula1);
  }
  
  /***************************************************************************
   * Logic conjunction of two formulas both of which may be null.
   * @param formula0 one formula (may be null).
   * @param formula1 another formula (may be null).
   * @return the logical conjunction (may be null which indicates TRUE).
   **************************************************************************/
  private static Formula andFormula(Formula formula0, Formula formula1)
  {
    if (formula0 == null) return formula1;
    if (formula1 == null) return formula0;
    return AndFormula.construct(formula0, formula1);
  }
  
  /***************************************************************************
   * Logic conjunction of two formulas both of which may be null.
   * @param formula0 one formula (may be null).
   * @param formula1 another formula (may be null).
   * @return the logical disjunction (may be null which indicates TRUE).
   **************************************************************************/
  private static Formula orFormula(Formula formula0, Formula formula1)
  {
    if (formula0 == null) return formula1;
    if (formula1 == null) return formula0;
    return OrFormula.construct(formula0, formula1);
  }
}
