// ----------------------------------------------------------------------------
// Simplification2.java
// Post-processing of a specification formula to make it easier understandable.
// $Id: Simplification2.java,v 1.23 2013/11/14 21:57:52 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General private License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General private License for more details.
//
// You should have received a copy of the GNU General private License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Judgements.Simplification.*;

/**
 * 
 * Post-processing of a specification formula to make it easier understandable.
 *
 */
public final class Simplification2 extends ASTCloner
{
  /***************************************************************************
   * Create simplifier.
   **************************************************************************/
  private Simplification2()
  {
  }
  
  /**************************************************************************
   * Simplify the given expression.
   * @param exp the expression to be simplified..
   * @return a simplified version of the expression.
   *************************************************************************/
  public static Expression perform(Expression exp)
  {
    Simplification2 simplifier = new Simplification2();
    return (Expression)simplifier.clone0(exp);
  }
  
  /**************************************************************************
   * Simplify the given formula. 
   * @param formula the formula to be simplified..
   * @return a simplified version of the formula.
   *************************************************************************/
  public static Formula perform(Formula formula)
  {
    Simplification2 simplifier = new Simplification2();
    return simplifier.cloneFormula(formula);
  }
  
  /***************************************************************************
   * Simplify a formula with this simplifier.
   * @param formula a formula.
   * @return the simplified version of the formula.
   ***************************************************************************/
  private Formula simplify(Formula formula)
  {
    formula.accept(this);
    return (Formula)this.result;
  }
  
  // ------------------------------------------------------------------------
  //
  // Visitor Functions
  //
  // -------------------------------------------------------------------------
  
  /***************************************************************************
   * Simplify formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(ForallFormula formula)
  {
    result = pushQuantifiers(formula);
  }

  /***************************************************************************
   * Simplify formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(ExistsFormula formula)
  {
    result = pushQuantifiers(formula);
  }
  
  /***************************************************************************
   * Simplify formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(AndFormula formula)
  {
    Vector<Formula> formulas = new Vector<Formula>();
    Simplification.collect(Tag.AND, formula, formulas);
    result = combine(Tag.AND, formulas, true);
  }
  
  /***************************************************************************
   * Simplify formula.
   * @param formula the formula to be simplified.
   **************************************************************************/
  public void visit(OrFormula formula)
  {
    Vector<Formula> formulas = new Vector<Formula>();
    Simplification.collect(Tag.OR, formula, formulas);
    result = combine(Tag.OR, formulas, true);
  }
  
  // ------------------------------------------------------------------------
  //
  // Main Functions
  //
  // -------------------------------------------------------------------------
  
  /***************************************************************************
   * Push quantifiers inside as far as possible.
   * @param formula a closed formula.
   * @return the simplified version of the formula.
   **************************************************************************/
  private Formula pushQuantifiers(QuantifiedFormula formula)
  {
    // determine components of formula
    QuantifiedFormula formula0 = (QuantifiedFormula)formula;
    TypedIdentifier[] vars = formula0.getVariables();
    int vn = vars.length;
    Formula body = simplify(formula0.getBody());
    Tag qtag;
    if (formula instanceof ForallFormula)
      qtag = Tag.FORALL;
    else if (formula instanceof ExistsFormula)
      qtag = Tag.EXISTS;
    else
    {
      Main.internalError("unknown quantified formula");
      return null;
    }
    Tag tag;
    if (body instanceof AndFormula) 
      tag = Tag.AND;
    else if (body instanceof OrFormula)
      tag = Tag.OR;
    else if (body instanceof ImpliesFormula)
      return pushQuantifiers(qtag, vars, (ImpliesFormula)body);
    else if (body instanceof IfThenElseFormula)
      return pushQuantifiers(qtag, vars, (IfThenElseFormula)body);
    else if (body instanceof LetFormula)
      return pushQuantifiers(qtag, vars, (LetFormula)body);
    else if (body instanceof QuantifiedFormula)
      return pushQuantifiers(qtag, vars, (QuantifiedFormula)body);
    else
      return quantifiedFormula(qtag, vars, simplify(body));
    
    // decompose body into parts
    Vector<Formula> formulas = new Vector<Formula>();
    Simplification.collect(tag, body, formulas);
    int fn = formulas.size();
    if (fn == 1) return quantifiedFormula(qtag, vars, body);
    
    // determine free variable of each part
    Vector<Collection<String>> free = new Vector<Collection<String>>(fn);
    for (int i=0; i<fn; i++)
      free.add(FreeVariables.compute(formulas.get(i)));
    
    // determine clusters of formulas that are closed with respect
    // to references to the quantified variables
        
    // a mapping of variable names to variable indices
    Map<String, Integer> indices = new MapClass<String, Integer>();
    for (int i=0; i<vn; i++)
    {
      String name = vars[i].getName().getString();
      indices.put(name, i);
    }
    
    // a mapping of formula indices to sets of variable indices
    Vector<Set<Integer>> vsets = new Vector<Set<Integer>>();
    for (int i=0; i<fn; i++)
    {
      Set<Integer> vset = new SetClass<Integer>();
      Collection<String> nset = free.get(i);
      for (String name : nset)
      {
        Integer index = indices.get(name);
        if (index != null) vset.add(index);
      }
      vsets.add(vset);
    }
    
    // a mapping of variable indices to sets of formula indices
    Vector<Set<Integer>> fsets = new Vector<Set<Integer>>();
    for (int i=0; i<vn; i++)
    {
      Set<Integer> fset = new SetClass<Integer>();
      for (int j=0; j<fn; j++)
      {
        Set<Integer> vset = vsets.get(j);
        if (vset.contains(i)) fset.add(j);
      }
      fsets.add(fset);
    }
    
    // a mapping of formula indices to sets of formula indices
    Vector<Set<Integer>> clusters = new Vector<Set<Integer>>();
    for (int i=0; i<fn; i++)
    {
      Set<Integer> cluster = new SetClass<Integer>();
      cluster.add(i);
      clusters.add(cluster);
    }
    boolean merged;
    do
    {
      merged = false;
      for (int i=0; i<fn; i++)
      {
        Set<Integer> cluster = clusters.get(i);
        int n = cluster.size();
        Set<Integer> cluster0 = new SetClass<Integer>(cluster);
        for (int c : cluster)
        {
          Set<Integer> vset = vsets.get(c);
          for (int var : vset)
            cluster0.addAll(fsets.get(var));
        }
        if (n != cluster0.size()) 
        {
          clusters.set(i, cluster0);
          merged = true;
        }
      }
    }
    while (merged);
    
    // determine the set of clusters
    Vector<Set<Integer>> clusterSet = new Vector<Set<Integer>>();
    for (Set<Integer> cluster : clusters)
    {
      if (!clusterSet.contains(cluster))
        clusterSet.add(cluster);
    }
    int cn = clusterSet.size();
    if (cn == 1) return quantifiedFormula(qtag, vars, simplify(body));

    // group formulas according to clusters
    Vector<Formula> formulas0 = new Vector<Formula>();
    for (int i=0; i<cn; i++)
    {
      Set<Integer> cluster = clusterSet.get(i);
      Vector<Integer> cluster0 = new Vector<Integer>(cluster);
      Collections.sort(cluster0);
      
      // determine the formulas
      Vector<Formula> formulas1 = new Vector<Formula>();
      for (int findex : cluster0)
        formulas1.add(formulas.get(findex));

      // determine the variables
      Set<Integer> variables = new SetClass<Integer>();
      for (int findex : cluster0)
        variables.addAll(vsets.get(findex));
      Vector<Integer> variables0 = new Vector<Integer>(variables);
      Collections.sort(variables0);
      
      // construct the result
      Vector<TypedIdentifier> vars0 = new Vector<TypedIdentifier>();
      for (int vindex : variables0)
        vars0.add(vars[vindex]);
      Formula result = quantifiedFormula(qtag, vars0, tag, formulas1);
      
      // collect the result
      formulas0.add(result);
    }
    
    // combine the resulting formulas
    return combine(tag, formulas0, true);
  }
  
  private static class FormulaInfo implements Comparable<FormulaInfo>
  {
    public int index;                          // the index of the formula
    public Collection<TypedIdentifier> bound;  // the bound variables referenced
    public FormulaInfo(int i)
    {
      index = i;
      bound = new SetClass<TypedIdentifier>();
    }
    public int compareTo(FormulaInfo info) 
    { return bound.size()-info.bound.size(); } 
  }
  
  /***************************************************************************
   * Push quantifiers into implication.
   * @param qtag the quantifier tag (FORALL or EXISTS).
   * @param vars the bound variables.
   * @param body the body of the quantified formula.
   * @return a formula logically equivalent to the quantified formula.
   **************************************************************************/
  private Formula pushQuantifiers(Tag qtag, 
    TypedIdentifier[] vars, ImpliesFormula body)
  {
    Formula formula1 = body.getFirst();
    Formula formula2 = body.getSecond();
    
    // try to extract conjuncts from hypothesis: 
    // (FORALL(x2,x3): A1 /\ A2 /\ A3 => B) -> 
    // (A1 => (FORALL(x2): A2 => (FORALL(x3): A3 => B)))
    if (qtag == Tag.FORALL)
    {
      Vector<Formula> formulas = new Vector<Formula>();
      Simplification.collect(Tag.AND, formula1, formulas);
      int fn = formulas.size();
      Map<String,TypedIdentifier> bvars = new MapClass<String,TypedIdentifier>();
      for (TypedIdentifier var : vars ) 
      {
        String name = var.getName().getString();
        bvars.put(name, var);
      }
      Vector<FormulaInfo> infos = new Vector<FormulaInfo>(fn);
      for (int i=0; i<fn; i++) 
      {
        FormulaInfo info = new FormulaInfo(i);
        infos.add(info);
        Collection<String> free = FreeVariables.compute(formulas.get(i));
        for (String v : free)
        {
          TypedIdentifier var = bvars.get(v);
          if (var != null) info.bound.add(var);
        }
      }
      // don't reorder formulas
      // Collections.sort(infos);
      Collection<TypedIdentifier> closed = 
          new SetClass<TypedIdentifier>();
      Vector<Collection<TypedIdentifier>> qvars = 
          new Vector<Collection<TypedIdentifier>>();
      for (int i=0; i<fn; i++)
      {
        Collection<TypedIdentifier> diff = 
            new SetClass<TypedIdentifier>(infos.get(i).bound);
        diff.removeAll(closed);
        closed.addAll(diff);
        qvars.add(diff);
      }
      Formula result = formula2;
      Collection<TypedIdentifier> open = 
          new SetClass<TypedIdentifier>(bvars.values());
      open.removeAll(closed);
      if (!open.isEmpty())
      {
        TypedIdentifier[] v = new TypedIdentifier[open.size()];
        result = simplify(quantifiedFormula(qtag, open.toArray(v), result));
      }
      Formula assumptions = null;
      for (int i=fn-1; i>=0; i--)
      {
        Formula f = formulas.get(infos.get(i).index);
        if (assumptions == null)
          assumptions = f;
        else
          assumptions = AndFormula.construct(f,assumptions);
        Collection<TypedIdentifier> vari = qvars.get(i);
        if (!vari.isEmpty())
        {
          result = ImpliesFormula.construct(assumptions,result);
          assumptions = null;
          TypedIdentifier[] v = new TypedIdentifier[vari.size()];
          result = quantifiedFormula(qtag, vari.toArray(v), result);
          if (i > 0) result = simplify(result);
        }
      }
      if (assumptions != null)
        result = ImpliesFormula.construct(assumptions,result);
      if (!(result instanceof ForallFormula)) return result;
      ForallFormula result0 = (ForallFormula)result;
      if (result0.getVariables().length != vars.length) return result;
    }
    
    // could not extract conjunctions, so proceed as planned
    Collection<String> free1 = FreeVariables.compute(formula1);
    Collection<String> free2 = FreeVariables.compute(formula2);
    Vector<TypedIdentifier> vars0 = new Vector<TypedIdentifier>();
    Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
    for (TypedIdentifier tid: vars)
    {
      String name = tid.getName().getString();
      if (free2.contains(name))
        vars0.add(tid);
      else if (free1.contains(name))
        vars1.add(tid);
    }
    if (vars0.size() < vars.length) 
    {
      // push reverted quantifier into first formula
      TypedIdentifier[] v0 = new TypedIdentifier[vars0.size()];
      vars0.toArray(v0);
      TypedIdentifier[] v1 = new TypedIdentifier[vars1.size()];
      vars1.toArray(v1);
      Tag qtag1 = qtag == Tag.FORALL ? Tag.EXISTS : Tag.FORALL;
      Formula formula0 = simplify(quantifiedFormula(qtag1, v1, formula1));
      return quantifiedFormula(qtag, v0,
          ImpliesFormula.construct(formula0, formula2));  
    }
    vars0.clear();
    Vector<TypedIdentifier> vars2 = new Vector<TypedIdentifier>();
    for (TypedIdentifier tid: vars)
    {
      String name = tid.getName().getString();
      if (free1.contains(name))
        vars0.add(tid);
      else if (free2.contains(name))
        vars2.add(tid);
    }
    if (vars0.size() < vars.length) 
    {
      // push unchanged quantifier into second formula
      TypedIdentifier[] v0 = new TypedIdentifier[vars0.size()];
      vars0.toArray(v0);
      TypedIdentifier[] v2 = new TypedIdentifier[vars2.size()];
      vars2.toArray(v2);
      Formula formula0 = simplify(quantifiedFormula(qtag, v2, formula2));
      return quantifiedFormula(qtag, v0,
          ImpliesFormula.construct(formula1, formula0));  
    }
    // no push is possible
    return quantifiedFormula(qtag, vars, simplify(body));
  }
  
  /***************************************************************************
   * Push quantifiers into if-then-else formula.
   * @param qtag the quantifier tag (FORALL or EXISTS).
   * @param vars the bound variables.
   * @param body the body of the quantified formula.
   * @return a formula logically equivalent to the quantified formula.
   **************************************************************************/
  private Formula pushQuantifiers(Tag qtag, 
    TypedIdentifier[] vars, IfThenElseFormula body)
  {
    Formula cond = body.getCondition();
    Collection<String> cvars = FreeVariables.compute(cond);
    Vector<TypedIdentifier> vars0 = new Vector<TypedIdentifier>();
    Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
    Vector<TypedIdentifier> vars2 = new Vector<TypedIdentifier>();
    for (TypedIdentifier tid: vars)
    {
      String name = tid.getName().getString();
      if (cvars.contains(name))
        vars0.add(tid);
      else
      {
        vars1.add(tid);
        vars2.add((TypedIdentifier)ASTCloner.clone(tid));
      }
    }
    if (vars1.isEmpty()) return quantifiedFormula(qtag, vars, simplify(body));
    TypedIdentifier[] v0 = new TypedIdentifier[vars0.size()];
    vars0.toArray(v0);
    TypedIdentifier[] v1 = new TypedIdentifier[vars1.size()];
    vars1.toArray(v1);
    TypedIdentifier[] v2 = new TypedIdentifier[vars2.size()];
    vars2.toArray(v2);
    Formula branch1 = simplify(quantifiedFormula(qtag, v1, body.getThenBranch()));
    Formula branch2 = simplify(quantifiedFormula(qtag, v2, body.getElseBranch()));
    return quantifiedFormula(qtag, v0,
        IfThenElseFormula.construct(cond, branch1, branch2));    
  }
  
  /***************************************************************************
   * Push quantifiers into let-formula.
   * @param qtag the quantifier tag (FORALL or EXISTS).
   * @param vars the bound variables.
   * @param body the body of the quantified formula.
   * @return a formula logically equivalent to the quantified formula.
   **************************************************************************/
  private Formula pushQuantifiers(Tag qtag, 
    TypedIdentifier[] vars, LetFormula body)
  {
    ValueDefinition[] defs = body.getDefinitions();
    Collection<String> cvars = new SetClass<String>();
    for (ValueDefinition def : defs)
    {
      cvars.add(def.getName().getString());
      FreeVariables.add(def.getValue(), cvars);
    }
    Vector<TypedIdentifier> vars0 = new Vector<TypedIdentifier>();
    Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
    for (TypedIdentifier tid: vars)
    {
      String name = tid.getName().getString();
      if (cvars.contains(name))
        vars0.add(tid);
      else
        vars1.add(tid);
    }
    if (vars1.isEmpty()) return quantifiedFormula(qtag, vars, simplify(body));
    TypedIdentifier[] v0 = new TypedIdentifier[vars0.size()];
    vars0.toArray(v0);
    TypedIdentifier[] v1 = new TypedIdentifier[vars1.size()];
    vars1.toArray(v1);
    Formula body0 = simplify(quantifiedFormula(qtag, v1, body.getBody()));
    return quantifiedFormula(qtag, v0, LetFormula.construct(defs, body0));    
  }
  
  /***************************************************************************
   * Push quantifiers into quantified formula.
   * @param qtag the quantifier tag (FORALL or EXISTS).
   * @param vars the bound variables.
   * @param body the body of the quantified formula.
   * @return a formula logically equivalent to the quantified formula.
   **************************************************************************/
  private Formula pushQuantifiers(Tag qtag, 
    TypedIdentifier[] vars, QuantifiedFormula body)
  {
    // Assume that x is not free in P.
    // FORALL x: EXISTS y: P => Q ->
    //   EXISTS y: P => FORALL x: Q
    // FORALL x: EXISTS y: IF P THEN Q ELSE R ->
    //   EXISTS y: IF P THEN FORALL x: Q ELSE FORALL x: R
    if (qtag == Tag.FORALL && body instanceof ExistsFormula)
    {
      Formula body0 = body.getBody();
      if (body0 instanceof ImpliesFormula)
      {
        ImpliesFormula body1 = (ImpliesFormula)body0;
        Formula formula1 = body1.getFirst();
        Formula formula2 = body1.getSecond();
        Collection<String> free = FreeVariables.compute(formula1);
        Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
        Vector<TypedIdentifier> vars2 = new Vector<TypedIdentifier>();
        for (TypedIdentifier tid : vars)
        {
          String name = tid.getName().getString();
          if (free.contains(name))
            vars1.add(tid);
          else
            vars2.add(tid);
        }
        if (!vars2.isEmpty()) 
        {
          TypedIdentifier[] v1 = new TypedIdentifier[vars1.size()];
          vars1.toArray(v1);
          TypedIdentifier[] v2 = new TypedIdentifier[vars2.size()];
          vars2.toArray(v2);
          Formula f2 = simplify(quantifiedFormula(Tag.FORALL, v2, formula2));
          return quantifiedFormula(Tag.FORALL, v1,
              quantifiedFormula(Tag.EXISTS, body.getVariables(),
                  ImpliesFormula.construct(formula1, f2)));
        }
      }
      else if (body0 instanceof IfThenElseFormula)
      {
        IfThenElseFormula body1 = (IfThenElseFormula)body0;
        Formula formula1 = body1.getCondition();
        Formula formula2 = body1.getThenBranch();
        Formula formula3 = body1.getElseBranch();
        Collection<String> free = FreeVariables.compute(formula1);
        Vector<TypedIdentifier> vars1 = new Vector<TypedIdentifier>();
        Vector<TypedIdentifier> vars2 = new Vector<TypedIdentifier>();
        Vector<TypedIdentifier> vars3 = new Vector<TypedIdentifier>();
        for (TypedIdentifier tid : vars)
        {
          String name = tid.getName().getString();
          if (free.contains(name))
            vars1.add(tid);
          else
          {
            vars2.add(tid);
            vars3.add((TypedIdentifier)ASTCloner.clone(tid));
          }
        }
        if (!vars2.isEmpty()) 
        {
          TypedIdentifier[] v1 = new TypedIdentifier[vars1.size()];
          vars1.toArray(v1);
          TypedIdentifier[] v2 = new TypedIdentifier[vars2.size()];
          vars2.toArray(v2);
          TypedIdentifier[] v3 = new TypedIdentifier[vars3.size()];
          vars3.toArray(v3);
          Formula f2 = simplify(quantifiedFormula(Tag.FORALL, v2, formula2));
          Formula f3 = simplify(quantifiedFormula(Tag.FORALL, v3, formula3));
          return quantifiedFormula(Tag.FORALL, v1,
              quantifiedFormula(Tag.EXISTS, body.getVariables(),
                  IfThenElseFormula.construct(formula1, f2, f3)));
        }
      }
    }
    return quantifiedFormula(qtag, vars, simplify(body));
  }
  
  // ------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // -------------------------------------------------------------------------
  
  /***************************************************************************
   * Combine formulas according to formula tag.
   * @param tag the tag indicating the type of the formula.
   * @param formulas the formulas.
   * @param simp true if the formulas are to be simplified.
   * @return the combined formula.
   **************************************************************************/
  private Formula combine(Tag tag, Vector<Formula> formulas, boolean simp)
  {
    // avoid duplicate/redundant formulas
    Vector<Formula> formulas0 = new Vector<Formula>();
    Set<String> formulas1 = new SetClass<String>();
    for (Formula f : formulas)
    {
      Formula f0 = simp ? simplify(f) : f;
      if (f0 instanceof True)
      {
        if (tag == Tag.AND) continue;
        if (tag == Tag.OR) return f0;
      }
      else if (f0 instanceof False)
      {
        if (tag == Tag.AND) return f0;
        if (tag == Tag.OR) continue;
      }
      String f1 = f0.toString();
      if (formulas1.contains(f1)) continue;
      formulas0.add(f0);
      formulas1.add(f1);
    }
          
    // construct result right-to-left for pretty-printing
    Formula result = null;
    int n = formulas0.size();
    for (int i=n-1;  i>=0; i--)
    {
      Formula f = formulas0.get(i);
      if (result == null)
        result = f;
      else if (tag == Tag.AND)
        result = AndFormula.construct(f, result);
      else if (tag == Tag.OR)
        result = OrFormula.construct(f, result);
      else
        Main.internalError("unknown tag");
    }
    if (result != null) return result;
    if (tag == Tag.AND)
      return ExpressionLogic.booleanFormula(true);
    else
      return ExpressionLogic.booleanFormula(false);
  }
  
  /***************************************************************************
   * Construct a quantified formula from multiple body formulas.
   * @param qtag the tag indicating the quantifier (FORALL, EXISTS)
   * @param vars the variables.
   * @param tag the tag indicating the combinatin of the body formulas (AND, OR)
   * @param formulas the body formulas.
   * @return the correspondingly constructed quantified formula.
   **************************************************************************/
  private Formula quantifiedFormula(Tag qtag, Vector<TypedIdentifier> vars,
    Tag tag, Vector<Formula> formulas)
  {
    Formula combined = combine(tag, formulas, false);
    int vn = vars.size();
    if (vn == 0) return combined;
    TypedIdentifier[] vars0 = new TypedIdentifier[vn];
    vars.toArray(vars0);
    return quantifiedFormula(qtag, vars0, combined);
  }
  
  /***************************************************************************
   * Construct quantified formula.
   * @param tag the tag indicating the type of the formula.
   * @param vars the quantified variables.
   * @param body the body of the formula.
   * @return the formula.
   **************************************************************************/
  private Formula quantifiedFormula(Tag tag,
    TypedIdentifier[] vars, Formula body)
  {
    if (vars.length == 0) return body;
    Formula result = null;
    if (tag == Tag.FORALL)
      result = ForallFormula.construct(vars, body);
    else if (tag == Tag.EXISTS)
      result = existsFormula(vars, body);
    else
      Main.internalError("unknown quantified formula");
    return result;
  }
  
  /***************************************************************************
   * Construct an existentially quantified formula
   * @param vars the variables.
   * @param body the formula body.
   * @return the quantified formula.
   **************************************************************************/
  private static Formula existsFormula(TypedIdentifier[] vars, Formula body)
  {
    if (body instanceof Constant) 
    {
      // (EXISTS(x): p) <=> p
      return body;
    }
    if (body instanceof SimilarFormula || body instanceof EqualsFormula ||
        body instanceof LessEqualFormula || body instanceof GreaterEqualFormula)
    {
      // (EXISTS(x): a = x) <=> TRUE
      BinaryAtomicFormula body0 = (BinaryAtomicFormula)body;
      Term term = body0.getFirst();
      if (term instanceof Reference)
      {
        Reference term0 = (Reference)term;
        Symbol symbol = term0.getName().getSymbol();
        for (TypedIdentifier var : vars)
        {
          if (var.getName().getSymbol() == symbol) 
            return ExpressionLogic.booleanFormula(true);
        }
      }
      term = body0.getSecond();
      if (term instanceof Reference)
      {
        Reference term0 = (Reference)term;
        Symbol symbol = term0.getName().getSymbol();
        for (TypedIdentifier var : vars)
        {
          if (var.getName().getSymbol() == symbol) 
            return ExpressionLogic.booleanFormula(true);
        }
      }
      return ExistsFormula.construct(vars, body);
    }
    if (body instanceof AndFormula)
    {
      // (EXISTS(x,y,z): x = y AND y = x AND x = z) <=> TRUE
      AndFormula body0 = (AndFormula)body;
      Collection<Formula> formulas = new Vector<Formula>();
      Simplification.collect(Simplification.Tag.AND, body0, formulas);
      boolean failed = false;
      for (Formula f : formulas)
      {
        if (!(f instanceof SimilarFormula || f instanceof EqualsFormula))
        {
          failed = true;
          break;
        }
        BinaryAtomicFormula f0 = (BinaryAtomicFormula)f;
        Term term = f0.getFirst();
        if (!(term instanceof Reference))
        {
          failed = true;
          break;
        }
        Reference term0 = (Reference)term;
        Symbol symbol = term0.getName().getSymbol();
        boolean found = false;
        for (TypedIdentifier var : vars)
        {
          if (var.getName().getSymbol() == symbol) 
          {
            found = true;
            break;
          }
        }
        if (!found) 
        {
          failed = true;
          break;
        }
        term = f0.getSecond();
        if (!(term instanceof Reference))
        {
          failed = true;
          break;
        }
        term0 = (Reference)term;
        symbol = term0.getName().getSymbol();
        found = false;
        for (TypedIdentifier var : vars)
        {
          if (var.getName().getSymbol() == symbol) 
          {
            found = true;
            break;
          }
        }
        if (!found)
        {
          failed = true;
          break;
        }
      }
      if (!failed) return ExpressionLogic.booleanFormula(true);

      // (EXISTS(x1,x2,x3): x1...  /\ x2... /\ x1 ~ x2 ) <=> TRUE
      // * each ... must either denote a similarity between two locally
      //   bound states or a positive or negative mode formula referring
      //   to a locally bound state
      // * at most one mode must be positive (must have no negative occurrence)
      // * not all state modes must be ruled out
      
      // executes, continues, breaks, returns, throws, throwsex
      int modes[] = new int[6]; 
      ClassSymbol exception = null;
      for (Formula f : formulas)
      { 
        if (isBoundSimilarityFormula(vars, f)) continue;
        int sign = 1;
        if (f instanceof NotFormula)
        {
          sign = -1;
          f = ((NotFormula)f).getBase();
        }
        if (!isBoundStateFormula(vars, f))
          return ExistsFormula.construct(vars, body);
        if (f instanceof StateExecutes)
        {
          if (sign == 1 && modes[0] < 0 || sign == -1 && modes[0] > 0) 
            return ExistsFormula.construct(vars, body);
          modes[0] = sign;
        }
        if (f instanceof StateContinues)
        {
          if (sign == 1 && modes[1] < 0 || sign == -1 && modes[1] > 0) 
            return ExistsFormula.construct(vars, body);
          modes[1] = sign;
        }
        if (f instanceof StateBreaks)
        {
          if (sign == 1 && modes[2] < 0 || sign == -1 && modes[2] > 0) 
            return ExistsFormula.construct(vars, body);
          modes[2] = sign;
        }
        if (f instanceof StateReturns)
        {
          if (sign == 1 && modes[3] < 0 || sign == -1 && modes[3] > 0) 
            return ExistsFormula.construct(vars, body);
          modes[3] = sign;
        }
        if (f instanceof StateThrows)
        {
          if (sign == 1 && modes[4] < 0 || sign == -1 && modes[4] > 0) 
            return ExistsFormula.construct(vars, body);
          modes[4] = sign;
        }
        if (f instanceof StateThrowsException)
        {
          ClassSymbol exception0 = 
            (ClassSymbol)((StateThrowsException)f).getException().getSymbol();
          if (sign == 1 && modes[5] < 0 || sign == -1 && modes[5] > 0 ||
              exception != null && exception0 != exception) 
            return ExistsFormula.construct(vars, body);
          modes[5] = sign;
          exception = exception0;
        }
      }
      int pos = 0;
      int neg = 0;
      for (int i=0; i<4; i++)
      {
        if (modes[i] > 0) pos = pos+1;
        else if (modes[i] < 0) neg = neg+1;
      }
      if (modes[4] > 0 || modes[5] > 0) pos = pos+1;
      if (modes[4] < 0 || modes[5] < 0) neg = neg+1;
      if (pos > 1 || neg == 5) return ExistsFormula.construct(vars, body);
      return ExpressionLogic.booleanFormula(true);
    }
    if (body instanceof OrFormula)
    {
      // (EXISTS(x): ... \/ (NOT)x.breaks \/ ...) <=> TRUE
      Collection<Formula> formulas = new Vector<Formula>();
      Simplification.collect(Simplification.Tag.OR, body, formulas);
      for (Formula f : formulas)
      {
        if (f instanceof NotFormula)
          f = ((NotFormula)f).getBase();
        if (isBoundStateFormula(vars, f)) return ExpressionLogic.booleanFormula(true);
      }
      return ExistsFormula.construct(vars, body);
    }

    if (body instanceof NotFormula)
    {
      // (EXISTS(x): !x.breaks) <=> TRUE
      Formula core = ((NotFormula)body).getBase();
      if (isBoundStateFormula(vars, core)) 
        return ExpressionLogic.booleanFormula(true);
      // (EXISTS(x): !x.null) <=> TRUE
      else if (isBoundNullFormula(vars, core))
        return ExpressionLogic.booleanFormula(true);
      else
        return ExistsFormula.construct(vars, body);
    }
    // (EXISTS(x): x.breaks) <=> TRUE
    if (isBoundStateFormula(vars, body)) 
      return ExpressionLogic.booleanFormula(true);
    // (EXISTS(x): x.null) <=> TRUE
    else if (isBoundNullFormula(vars, body))
      return ExpressionLogic.booleanFormula(true);
    else
      return ExistsFormula.construct(vars, body);
  }
  
  /***************************************************************************
   * Determine whether formula is a state similarity whose states are bound.
   * @param vars the bound variables.
   * @param formula the formula.
   * @return true if formula is a state formula whose state appears among vars.
   ***************************************************************************/
  private static boolean 
  isBoundSimilarityFormula(TypedIdentifier[] vars, Formula formula)
  {
    if (!(formula instanceof SimilarFormula)) return false;
    SimilarFormula formula0 = (SimilarFormula)formula;
    Term arg0 = formula0.getFirst();
    Term arg1 = formula0.getSecond();
    if (arg0 instanceof Reference && arg1 instanceof Reference)
    {
      Reference a0 = (Reference)arg0;
      Reference a1 = (Reference)arg1;
      ValueSymbol symbol0 = (ValueSymbol)a0.getName().getSymbol();
      ValueSymbol symbol1 = (ValueSymbol)a1.getName().getSymbol();
      boolean found0 = false;
      boolean found1 = false;
      for (TypedIdentifier var : vars)
      {
        Symbol symbol = var.getName().getSymbol();
        if (symbol == symbol0) found0 = true;
        if (symbol == symbol1) found1 = true;
        if (found0 && found1) return true;
      }
    }
    return false;
  }
  
  /***************************************************************************
   * Determine whether formula is a state formula whose state is bound.
   * @param vars the bound variables.
   * @param formula the formula.
   * @return true if formula is a state formula whose state appears among vars.
   ***************************************************************************/
  private static boolean 
  isBoundStateFormula(TypedIdentifier[] vars, Formula formula)
  {
    if (!(formula instanceof PostfixFormula)) return false;
    PostfixFormula formula0 = (PostfixFormula)formula;
    Expression arg = formula0.getArgument();
    if (arg instanceof Reference)
    {
      Reference arg0 = (Reference)arg;
      ValueSymbol symbol = (ValueSymbol)arg0.getName().getSymbol();
      for (TypedIdentifier var : vars)
      {
        if (var.getName().getSymbol() == symbol) return true;
      }
    }
    return false;
  }
  
  /***************************************************************************
   * Determine whether formula is a null formula whose record is bound.
   * @param vars the bound variables.
   * @param formula the formula.
   * @return true if formula is a null formula whose record appears among vars.
   ***************************************************************************/
  private static boolean 
  isBoundNullFormula(TypedIdentifier[] vars, Formula formula)
  {
    if (!(formula instanceof SelectionTerm)) return false;
    SelectionTerm term = (SelectionTerm)formula;
    Selector sel = term.getSelector();
    if (!(sel instanceof SelectorIdentifier)) return false;
    SelectorIdentifier sel0 = (SelectorIdentifier)sel;
    Identifier ident = sel0.getIdentifier();
    if (!ident.getString().equals(TypeTranslator.nullName)) return false;
    Term base = term.getBase();
    if (!(base instanceof Reference)) return false;
    Reference ref = (Reference) base;
    Name name = ref.getName();
    if (name.idents.length != 1) return false;
    Symbol symbol = name.idents[0].getSymbol();
    for (TypedIdentifier var : vars)
    {
      if (var.getName().getSymbol() == symbol) return true;
    }
    return false;
  }
}
