// ----------------------------------------------------------------------------
// StatJudgement.java
// A statement judgement.
// $Id: StatJudgement.java,v 1.11 2011/09/30 15:47:38 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General private License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General private License for more details.
//
// You should have received a copy of the GNU General private License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Judgements;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A statement judgement.
 *
 */
public final class StatJudgement extends ASTBase
{
  // the command's precondition (may be null)
  private  Formula precondition;      
  
  public final Formula formula;      // the specification formula
  public final Formula sformula;     // its simplified version
  public final boolean executes;
  public final boolean continues;
  public final boolean breaks;
  public final boolean returns;
  public final Set<ClassSymbol> exceptions;
  public final Set<VariableSymbol> variables;
  
  // set by setTermination()
  private Formula termination;  // the termination formula (may be null)
  
  /**************************************************************************
   * Construct a command specification.
   * @param precondition the formula describing the precondition (may be null).
   * @param formula the formula describing the transition relation.
   * @param executes true if the command may execute normally.
   * @param continues true if the command may "continue".
   * @param breaks true if the command may "break".
   * @param returns true if the command may "return".
   * @param exceptions the set of exceptions that may be thrown by the command
   *        (null means the empty set)
   * @param variables the set of variables that may be modified by the command
   *        (null means the empty set)
   *************************************************************************/
  public StatJudgement(Formula precondition, Formula formula, 
    boolean executes, boolean continues, boolean breaks, boolean returns,
    Set<ClassSymbol> exceptions, Set<VariableSymbol> variables)
  {
    this.precondition = precondition;
    this.formula = formula;
    this.sformula = Simplification.perform(formula);
    this.executes = executes;
    this.continues = continues;
    this.breaks = breaks;
    this.returns = returns;
    if (exceptions != null)
      this.exceptions = exceptions;
    else
      this.exceptions = new SetClass<ClassSymbol>();
    if (variables != null)
      this.variables = variables;
    else
      this.variables = new SetClass<VariableSymbol>();
    this.termination = null;
  }
  
  /****************************************************************************
   * Set the termination formula.
   * @param termination the formula.
   ***************************************************************************/
  public void setTermination(Formula termination)
  {
    this.termination = termination;
  }
  
  /****************************************************************************
   * Get the termination formula.
   * @return the formula.
   ***************************************************************************/
  public Formula getTermination()
  {
    return termination;
  }
  
  /****************************************************************************
   * Get the precondition formula.
   * @param pre the formula.
   ***************************************************************************/
  public void setPrecondition(Formula pre)
  {
    this.precondition = pre;
  }
  
  /****************************************************************************
   * Get the precondition formula.
   * @return the precondition.
   ***************************************************************************/
  public Formula getPrecondition()
  {
    return precondition;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    if (precondition != null) precondition.accept(visitor);
    formula.accept(visitor);
    sformula.accept(visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    if (precondition != null) precondition.accept(visitor);
    formula.accept(visitor);
    sformula.accept(visitor);
  }
}
