// ----------------------------------------------------------------------------
// ProofNavigator.java
// Translate theories into the RISC ProofNavigator.
// $Id: ProofNavigator.java,v 1.21 2013/04/03 09:54:49 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.ewuni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Proving;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Semantics.*;

import fmrisc.ProofNavigator.Syntax.Construct;

/**
 * 
 * Translate theories into the RISC ProofNavigator.
 *
 */
public final class ProofNavigator
{  
  
  // name of string type
  private final String stringTypeName = StringType.stringKeyword;
  
  // the base name of string literals (must create names not used elsewhere)
  private final static String stringLiteralBaseName = "_literal";
  
  /****************************************************************************
   * Translate classical problem to RISC ProofNavigator Problem
   * @param task the problem to be translated.
   * @param dir the directory to which information may be written
   * @param out stream for printing messages
   * @return the translation.
   **************************************************************************/
  public static ProofNavigatorProblem translate(ClassicalProblem task, 
    File dir, ErrorStream out)
  {
    ProofNavigator pn = new ProofNavigator(task); 
    fmrisc.ProofNavigator.Syntax.Declaration[] result = 
      new fmrisc.ProofNavigator.Syntax.Declaration[pn.result.size()];
    pn.result.toArray(result);
    return new ProofNavigatorProblem(result, pn.goal, dir, out); 
  }
  
  // the generated declarations 
  private Vector<fmrisc.ProofNavigator.Syntax.Declaration> result;
  
  // the translated goal
  private fmrisc.ProofNavigator.Syntax.Expression goal;
  
  // the number of string literals generated so far 
  private int stringLiteralCounter = 0;
 
  /****************************************************************************
   * Construct a ProofNavigator translation of a flattened type checking task.
   * @param task the flattened type checking task.
   ***************************************************************************/
  private ProofNavigator(ClassicalProblem task)
  {    
    // initialize tables
    this.result = new Vector<fmrisc.ProofNavigator.Syntax.Declaration>();

    // generate declaration of string type
    add(Construct.typeDeclaration(
      Construct.typeDeclIdentifier(stringTypeName), null));
    
    // perform translation of declarations
    translate(task.getDeclarations());
 
    // translate goal
    goal = translate(task.getGoal());
  }
  
  /***************************************************************************
   * Add declaration to sequence of generated declarations.
   * @param decl the declaration to be added.
   **************************************************************************/
  private void add(fmrisc.ProofNavigator.Syntax.TypeDeclaration decl)
  {
    result.add(decl);
  }
  
  /***************************************************************************
   * Add declaration to sequence of generated declarations.
   * @param decl the declaration to be added.
   **************************************************************************/
  private void add(fmrisc.ProofNavigator.Syntax.FormulaDeclaration decl)
  {
    result.add(decl);
  }
  
  /***************************************************************************
   * Add declaration to sequence of generated declarations.
   * @param decl the declaration to be added.
   **************************************************************************/
  private void add(fmrisc.ProofNavigator.Syntax.ValueDeclaration decl)
  {
    result.add(decl);
  }
  
  /****************************************************************************
   * Translate sequence of declarations and store result in global variable
   * @param decls the declarations to be translated
   **************************************************************************/
  private void translate(Declaration[] decls)
  {
    for (Declaration d : decls)
    {
      fmrisc.ProofNavigator.Syntax.Declaration d0 = translate(d);
      if (d0 instanceof fmrisc.ProofNavigator.Syntax.TypeDeclaration) 
        add((fmrisc.ProofNavigator.Syntax.TypeDeclaration)d0);
      else if (d0 instanceof fmrisc.ProofNavigator.Syntax.ValueDeclaration) 
        add((fmrisc.ProofNavigator.Syntax.ValueDeclaration)d0);
      else if (d0 instanceof fmrisc.ProofNavigator.Syntax.FormulaDeclaration) 
        add((fmrisc.ProofNavigator.Syntax.FormulaDeclaration)d0);
      else 
        Main.internalError("unknown declaration" + d);  
    }
  }
  
  // --------------------------------------------------------------------------
  //
  // Translate declarations
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Translate declaration.
   * @param decl the declaration to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Declaration 
    translate(Declaration decl)
  {
    if (decl instanceof TypeDeclaration)
      return translate((TypeDeclaration)decl);
    if (decl instanceof ValueDeclaration)
      return translate((ValueDeclaration)decl);
    if (decl instanceof FormulaDefinition)
      return translate((FormulaDefinition)decl);
    Main.internalError("unknown declaration: " + decl);
    return null;
  }   
  
  /****************************************************************************
   * Translate declaration.
   * @param decl the declaration to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TypeDeclaration 
    translate(TypeDeclaration decl)
  {
    Identifier name = decl.getName();
    fmrisc.ProofNavigator.Syntax.Type type = null;
    if (decl instanceof TypeDefinition)
      type = translate(((TypeDefinition)decl).getValue());
    return new fmrisc.ProofNavigator.Syntax.TypeDeclaration(
      translateTypeDeclIdentifier(name), type);
  }
  
  /***************************************************************************
   * Translate identifier to type declaration identifier.
   * @param name the identifier to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TypeDeclIdentifier
    translateTypeDeclIdentifier(Identifier name)
  {
    return new fmrisc.ProofNavigator.Syntax.TypeDeclIdentifier(name.getString());
  }
  
  /****************************************************************************
   * Translate declaration.
   * @param decl the declaration to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ValueDeclaration 
    translate(ValueDeclaration decl)
  {
    Identifier name = decl.getName();
    Type type = decl.getType();
    fmrisc.ProofNavigator.Syntax.Expression exp = null;
    if (decl instanceof ValueDeclarationInitialized)
      exp = translate(((ValueDeclarationInitialized)decl).getValue());
    if (decl instanceof ValueDeclarationFormula)
      exp = translate((ValueDeclarationFormula)decl).getValue();
    return new fmrisc.ProofNavigator.Syntax.ValueDeclaration(
      translateValueDeclIdentifier(name), translate(type), exp);
  }
  
  /***************************************************************************
   * Translate identifier to type declaration identifier.
   * @param name the identifier to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ValueDeclIdentifier
    translateValueDeclIdentifier(Identifier name)
  {
    return new fmrisc.ProofNavigator.Syntax.ValueDeclIdentifier(name.getString());
  }
  
  /****************************************************************************
   * Translate declaration.
   * @param decl the declaration to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.FormulaDeclaration 
    translate(FormulaDefinition decl)
  {
    Identifier name = decl.getName();
    Formula formula = decl.getValue();
    if (decl instanceof AxiomDefinition)
      return new fmrisc.ProofNavigator.Syntax.AxiomDeclaration(
        translateFormulaDeclIdentifier(name), translate(formula));
    else
      return new fmrisc.ProofNavigator.Syntax.FormulaDeclaration(
        translateFormulaDeclIdentifier(name), translate(formula));
  }
  
  /***************************************************************************
   * Translate identifier to type declaration identifier.
   * @param name the identifier to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.FormulaDeclIdentifier
    translateFormulaDeclIdentifier(Identifier name)
  {
    return new fmrisc.ProofNavigator.Syntax.FormulaDeclIdentifier(name.getString());
  }
  
  // --------------------------------------------------------------------------
  //
  // Translate formulas
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
  translate(Formula formula)
  {
    if (formula instanceof AtomicFormula)
      return translate((AtomicFormula)formula);
    if (formula instanceof BinaryAtomicFormula)
      return translate((BinaryAtomicFormula)formula);
    if (formula instanceof BinaryFormula)
      return translate((BinaryFormula)formula);
    if (formula instanceof BooleanLiteral)
      return translateFormula((BooleanLiteral)formula);
    if (formula instanceof WritesOnlyFormula)
      return translate((WritesOnlyFormula)formula);
    if (formula instanceof IfThenElseFormula)
      return translate((IfThenElseFormula)formula);
    if (formula instanceof LetFormula)
      return translate((LetFormula)formula);
    if (formula instanceof PostfixFormula)
      return translate((PostfixFormula)formula);
    if (formula instanceof QuantifiedFormula)
      return translate((QuantifiedFormula)formula);
    if (formula instanceof Reference)
      return translateFormula((Reference)formula);
    if (formula instanceof ProgramVariableBase)
      return translateFormula((ProgramVariableBase)formula);
    if (formula instanceof SelectionTerm)
      return translateFormula((SelectionTerm)formula);
    if (formula instanceof UnaryFormula)
      return translate((UnaryFormula)formula);
    Main.internalError("Unknown kind of formula " + formula);
    return null;
  }
  
  /***************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ApplicationExpression
    translate(AtomicFormula formula)
  {
    return new fmrisc.ProofNavigator.Syntax.ApplicationExpression(
        translate(formula.getFunction()), translate(formula.getArguments()));
  }
  
  /***************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BinaryExpression
    translate(BinaryAtomicFormula formula)
  {
    Term first = formula.getFirst();
    Term second = formula.getSecond();
    if (formula instanceof EqualsFormula)
    {
      if (LogicChecking.isBooleanType(first.getType()))
        return new fmrisc.ProofNavigator.Syntax.EquivalentFormula(
          translate(first), translate(second)); 
      else
        return new fmrisc.ProofNavigator.Syntax.EqualsFormula(
          translate(first), translate(second));
    }
    if (formula instanceof GreaterEqualFormula)
      return new fmrisc.ProofNavigator.Syntax.GreaterEqualFormula(
          translate(first), translate(second));
    if (formula instanceof GreaterFormula)
      return new fmrisc.ProofNavigator.Syntax.GreaterFormula(
          translate(first), translate(second));
    if (formula instanceof LessEqualFormula)
      return new fmrisc.ProofNavigator.Syntax.LessEqualFormula(
          translate(first), translate(second));
    if (formula instanceof LessFormula)
      return new fmrisc.ProofNavigator.Syntax.LessFormula(
          translate(first), translate(second));
    if (formula instanceof NotEqualsFormula)
    {
      if (LogicChecking.isBooleanType(first.getType()))
        return new fmrisc.ProofNavigator.Syntax.NotEquivalentFormula(
          translate(first), translate(second)); 
      else
        return new fmrisc.ProofNavigator.Syntax.NotEqualsFormula(
          translate(first), translate(second));
    }
    Main.internalError("Unknown kind of binary atomic formula " + formula);
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BinaryExpression
    translate(BinaryFormula formula)
  {    
    Formula first = formula.getFirst();
    Formula second = formula.getSecond();
    if (formula instanceof AndFormula)
      return new fmrisc.ProofNavigator.Syntax.AndFormula(
          translate(first), translate(second));
    if (formula instanceof EquivalentFormula)
      return new fmrisc.ProofNavigator.Syntax.EquivalentFormula(
          translate(first), translate(second));
    if (formula instanceof ImpliesFormula)
      return new fmrisc.ProofNavigator.Syntax.ImpliesFormula(
          translate(first), translate(second));
    if (formula instanceof NotEquivalentFormula)
      return new fmrisc.ProofNavigator.Syntax.NotEquivalentFormula(
          translate(first), translate(second));
    if (formula instanceof OrFormula)
      return new fmrisc.ProofNavigator.Syntax.OrFormula(
          translate(first), translate(second));
    Main.internalError("Unknown kind of binary formula " + formula);
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Logical
    translateFormula(BooleanLiteral formula)
  {
    if (formula instanceof True)
      return new fmrisc.ProofNavigator.Syntax.Logical(true);
    if (formula instanceof False)
      return new fmrisc.ProofNavigator.Syntax.Logical(false);
    Main.internalError("Unknown kind of boolean literal " + formula);
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(WritesOnlyFormula formula)
  {
    if (formula instanceof ReadsOnlyFormula)
      return translate((ReadsOnlyFormula)formula);
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(ReadsOnlyFormula formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.IfThenElseExpression
    translate(IfThenElseFormula formula)
  {
    return new fmrisc.ProofNavigator.Syntax.IfThenElseExpression(
        translate(formula.getCondition()), 
        translate(formula.getThenBranch()),
        translate(formula.getElseBranch()));
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.LetExpression
    translate(LetFormula formula)
  {
    return new fmrisc.ProofNavigator.Syntax.LetExpression(
        translateValueDefs(formula.getDefinitions()),
        translate(formula.getBody()));
  }
  
  /****************************************************************************
   * Translate value definitions.
   * @param defs the value definitions.
   * @return the translation
   **************************************************************************/
  public fmrisc.ProofNavigator.Syntax.ValueDeclaration[] 
    translateValueDefs(ValueDefinition[] defs)
  {
    Vector<fmrisc.ProofNavigator.Syntax.ValueDeclaration> result = 
      new Vector<fmrisc.ProofNavigator.Syntax.ValueDeclaration>();
    for (ValueDefinition d : defs)
      result.add(translateValueDef(d));
    fmrisc.ProofNavigator.Syntax.ValueDeclaration[] result0 = 
      new fmrisc.ProofNavigator.Syntax.ValueDeclaration[result.size()];
    result.toArray(result0);
    return result0;   
  }
  
  /****************************************************************************
   * Translate declaration.
   * @param decl the declaration to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ValueDeclaration 
    translateValueDef(ValueDefinition def)
  {
    if (def instanceof ValueDeclarationInitialized)
    {
      ValueDeclarationInitialized def0 = (ValueDeclarationInitialized)def;
      return new fmrisc.ProofNavigator.Syntax.ValueDeclaration(
          translateValueDeclIdentifier(
              def0.getName()), translate(def0.getType()), 
              translate(def0.getValue()));
    }  
      return new fmrisc.ProofNavigator.Syntax.ValueDeclaration(
          translateValueDeclIdentifier(
              def.getName()), null, translate(def.getValue()));
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(PostfixFormula formula)
  {
    if (formula instanceof StateBreaks)
      return translate((StateBreaks)formula);
    if (formula instanceof StateContinues)
      return translate((StateContinues)formula);
    if (formula instanceof StateExecutes)
      return translate((StateExecutes)formula);
    if (formula instanceof StateReturns)
      return translate((StateReturns)formula);
    if (formula instanceof StateThrows)
      return translate((StateThrows)formula);
    if (formula instanceof StateThrowsException)
      return translate((StateThrowsException)formula);
    Main.internalError("unknown postfix formula " + formula);
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateBreaks formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateContinues formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateExecutes formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateReturns formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateThrows formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateThrowsException formula)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.QuantifiedExpression
    translate(QuantifiedFormula formula)
  {
    if (formula instanceof ExistsFormula)
      return new fmrisc.ProofNavigator.Syntax.ExistsFormula(
          translate(formula.getVariables()), translate(formula.getBody()));
    if (formula instanceof ForallFormula)
      return new fmrisc.ProofNavigator.Syntax.ForallFormula(
          translate(formula.getVariables()), translate(formula.getBody()));
    Main.internalError("Unknown quantified formula: " + formula);
    return null;
  }
  
  /****************************************************************************
   * Translate typed identifiers.
   * @param tids the typed identifiers to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TypedIdentifier[]
    translate(TypedIdentifier[] tids)
  {
    int n = tids.length;
    fmrisc.ProofNavigator.Syntax.TypedIdentifier[] result = 
      new fmrisc.ProofNavigator.Syntax.TypedIdentifier[n];
    for (int i=0; i<n; i++)
      result[i] = translate(tids[i]);
    return result;
  }
  
  /****************************************************************************
   * Translate typed identifier.
   * @param tid the typed identifier to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TypedIdentifier
    translate(TypedIdentifier tid)
  {
    return new fmrisc.ProofNavigator.Syntax.TypedIdentifier(
        translateValueDeclIdentifier(tid.getName()), translate(tid.getType()));
  }
  
  /***************************************************************************
   * Translate reference.
   * @param ref the reference to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translateFormula(Reference ref)
  {
    return translateReference(ref);
  }
  
  /***************************************************************************
   * Translate program variable.
   * @param var the program variable to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translateFormula(ProgramVariableBase var)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /***************************************************************************
   * Translate selection term as formula.
   * @param sel the term to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translateFormula(SelectionTerm term)
  {
    return translate(term);
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.UnaryExpression
    translate(UnaryFormula formula)
  {    
    if (formula instanceof NotFormula)
      return translate((NotFormula)formula);
    Main.internalError("Unknown kind of unary formula " + formula);
    return null;
  }
  
  /****************************************************************************
   * Translate formula.
   * @param formula the formula to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.NotFormula
    translate(NotFormula formula)
  {    
    return new fmrisc.ProofNavigator.Syntax.NotFormula(
        translate(formula.getBase()));
  }
  
  // --------------------------------------------------------------------------
  //
  // Translate terms
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Translate terms.
   * @param terms the terms to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression[]
    translate(Term[] terms)
  {
    int n = terms.length;
    fmrisc.ProofNavigator.Syntax.Expression[] result =
      new fmrisc.ProofNavigator.Syntax.Expression[n];
    for (int i=0; i<n; i++)
      result[i] = translate(terms[i]);
    return result;
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(Term term)
  {
    if (term instanceof BinaryTerm)
      return translate((BinaryTerm)term);
    if (term instanceof BooleanLiteral)
      return translateTerm((BooleanLiteral)term);
    if (term instanceof BitLiteral)
      return translateTerm((BitLiteral)term);
    if (term instanceof FunctionApplication)
      return translate((FunctionApplication)term);
    if (term instanceof IfThenElseTerm)
      return translate((IfThenElseTerm)term);
    if (term instanceof IntLiteral)
      return translate((IntLiteral)term);
    if (term instanceof StringLiteral)
      return translate((StringLiteral)term);
    if (term instanceof LetTerm)
      return translate((LetTerm)term);
    if (term instanceof PostfixTerm)
      return translate((PostfixTerm)term);
    if (term instanceof QuantifiedTerm)
      return translate((QuantifiedTerm)term);
    if (term instanceof Reference)
      return translateTerm((Reference)term);
    if (term instanceof StateLiteral)
      return translate((StateLiteral)term);
    if (term instanceof TermBase)
      return translate((TermBase)term);
    if (term instanceof UnaryTerm)
      return translate((UnaryTerm)term);
    Main.internalError("unknown kind of term " + term);
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BinaryExpression
    translate(BinaryTerm term)
  {
    if (term instanceof DividesTerm)
      return translate((DividesTerm)term);
    if (term instanceof MinusTerm)
      return translate((MinusTerm)term);
    if (term instanceof PlusTerm)
      return translate((PlusTerm)term);
    if (term instanceof PowerTerm)
      return translate((PowerTerm)term);
    if (term instanceof TimesTerm)
      return translate((TimesTerm)term);
    if (term instanceof BitAndTerm)
      return translate((BitAndTerm)term);
    if (term instanceof BitOrTerm)
      return translate((BitOrTerm)term);
    Main.internalError("unknown kind of binary term " + term);
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.DividesTerm
    translate(DividesTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.DividesTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.MinusTerm
    translate(MinusTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.MinusTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.PlusTerm
    translate(PlusTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.PlusTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.PowerTerm
    translate(PowerTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.PowerTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TimesTerm
    translate(TimesTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.TimesTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BitAndTerm
    translate(BitAndTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.BitAndTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param terms the term to be translated.
   * @return the translation.
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BitOrTerm
    translate(BitOrTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.BitOrTerm(
        translate(term.getFirst()), translate(term.getSecond()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Logical
    translateTerm(BooleanLiteral term)
  {
    if (term instanceof True)
      return new fmrisc.ProofNavigator.Syntax.Logical(true);
    if (term instanceof False)
      return new fmrisc.ProofNavigator.Syntax.Logical(false);
    Main.internalError("Unknown kind of boolean literal " + term);
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BitLogical
    translateTerm(BitLiteral term)
  {
    if (term instanceof BitTrue)
      return new fmrisc.ProofNavigator.Syntax.BitLogical(true);
    if (term instanceof BitFalse)
      return new fmrisc.ProofNavigator.Syntax.BitLogical(false);
    Main.internalError("Unknown kind of boolean literal " + term);
    return null;
  }
  
  /***************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation.
   **************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ApplicationExpression
    translate(FunctionApplication term)
  {
    return new fmrisc.ProofNavigator.Syntax.ApplicationExpression(
        translate(term.getFunction()), translate(term.getArguments()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.IfThenElseExpression
    translate(IfThenElseTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.IfThenElseExpression(
        translate(term.getCondition()), 
        translate(term.getThenBranch()),
        translate(term.getElseBranch()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Number
    translate(IntLiteral term)
  {
    return new fmrisc.ProofNavigator.Syntax.Number(term.getName());
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StringLiteral term)
  {
    // we replace string by new literal constant with unknown value
    String name = stringLiteralBaseName + stringLiteralCounter;
    stringLiteralCounter = stringLiteralCounter+1;
    
    // generate constant declaration
    add(Construct.valueDeclaration(
        Construct.valueDeclIdentifier(name), 
        Construct.reference(stringTypeName), null));
    
    // translate to reference to constant
    return Construct.reference(name);
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.LetExpression
    translate(LetTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.LetExpression(
        translateValueDefs(term.getDefinitions()),
        translate(term.getBody()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(PostfixTerm term)
  {
    if (term instanceof StateValue)
      return translate((StateValue)term);
    if (term instanceof StateMessage)
      return translate((StateMessage)term);
    Main.internalError("unknown postfix term " + term);
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateValue term)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateMessage term)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.QuantifiedExpression
    translate(QuantifiedTerm term)
  {
    if (term instanceof ArrayTerm)
      return translate((ArrayTerm)term);
    if (term instanceof LambdaFormula)
      return translate((LambdaFormula)term);
    if (term instanceof LambdaTerm)
      return translate((LambdaTerm)term);
    Main.internalError("unknown kind of quantified term: " + term);
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ArrayTerm
    translate(ArrayTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.ArrayTerm(
        translate(term.getVariables()), translate(term.getBody()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.LambdaTerm
    translate(LambdaFormula term)
  {
    return new fmrisc.ProofNavigator.Syntax.LambdaTerm(
        translate(term.getVariables()), translate(term.getBody()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.LambdaTerm
    translate(LambdaTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.LambdaTerm(
        translate(term.getVariables()), translate(term.getBody()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param ref the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression translateTerm(Reference ref)
  {
    return translateReference(ref);
  }
  
  /****************************************************************************
   * Translate reference.
   * @param ref the reference to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression translateReference(Reference ref)
  {
    // get rid of non-value name prefix
    Name name = ref.getName();
    int n = name.getLength();

    // name may be plain identifier
    if (n == 1)
      return new fmrisc.ProofNavigator.Syntax.Reference(
          name.idents[0].getString());
    
    // name may denote selection term
    Term term = Reference.construct(name.idents[0]);
    for (int i=1; i<n; i++)
    {
      term = SelectionTerm.construct(term,
          SelectorIdentifier.construct(name.idents[i]));
      ValueSymbol symbol = (ValueSymbol)name.idents[i].getSymbol();
      term.setType(symbol.getType());
    }
    return translate(term);
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(StateLiteral term)
  {
    if (term instanceof NewState) return translate((NewState)term);
    if (term instanceof OldState) return translate((OldState)term);
    Main.internalError("unknwon state literal");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(NewState term)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(OldState term)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translate(TermBase term)
  {
    if (term instanceof ProgramVariableBase)
      return translateTerm((ProgramVariableBase)term);
    if (term instanceof RecordTerm)
      return translate((RecordTerm)term);
    if (term instanceof SelectionTerm)
      return translate((SelectionTerm)term);
    if (term instanceof TupleTerm)
      return translate((TupleTerm)term);
    if (term instanceof UpdateTerm)
      return translate((UpdateTerm)term);
    Main.internalError("unknown term derived from TermBase");
    return null;
  }
  
  /****************************************************************************
   * Translate program variable to term.
   * @param var the variable to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Expression
    translateTerm(ProgramVariableBase var)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.RecordTerm
    translate(RecordTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.RecordTerm(
        translate(term.getBase()));
  }
  
  /****************************************************************************
   * Translate valued identifiers.
   * @param vids the valued identifiers to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ValuedIdentifier[]
    translate(ValuedIdentifier[] vids)
  {
    int n = vids.length;
    fmrisc.ProofNavigator.Syntax.ValuedIdentifier[] result =
      new fmrisc.ProofNavigator.Syntax.ValuedIdentifier[n];
    for (int i=0; i<n; i++)
      result[i] = translate(vids[i]);
    return result;
  }
  
  /****************************************************************************
   * Translate valued identifier.
   * @param vid the valued identifier to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ValuedIdentifier
    translate(ValuedIdentifier vid)
  {
    return new fmrisc.ProofNavigator.Syntax.ValuedIdentifier(
        translateReference(vid.getIdentifier()), translate(vid.getValue()));
  }
  
  /****************************************************************************
   * Translate reference
   * @param id the reference to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Reference
    translateReference(Identifier id)
  {
    return new fmrisc.ProofNavigator.Syntax.Reference(id.getString());
  }
  
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.SelectionTerm
    translate(SelectionTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.SelectionTerm(
        translate(term.getBase()),
        translate(term.getSelector()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TupleTerm
    translate(TupleTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.TupleTerm(
        translate(term.getBase()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.UpdateTerm
    translate(UpdateTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.UpdateTerm(
        translate(term.getBase()),
        translate(term.getSelectors()),
        translate(term.getValue()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.UnaryExpression
    translate(UnaryTerm term)
  {
    if (term instanceof NegationTerm)
      return translate((NegationTerm)term);
    if (term instanceof BitNegationTerm)
      return translate((BitNegationTerm)term);
    Main.internalError("unknown unary term");
    return null;
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.NegationTerm
    translate(NegationTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.NegationTerm(
        translate(term.getBase()));
  }
  
  /****************************************************************************
   * Translate term.
   * @param term the term to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.BitNotTerm
    translate(BitNegationTerm term)
  {
    return new fmrisc.ProofNavigator.Syntax.BitNotTerm(
        translate(term.getBase()));
  }
  
  // --------------------------------------------------------------------------
  //
  // Auxiliaries
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Translate selectors
   * @param sels the selectors to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Selector[]
    translate(Selector[] sels)
  {
    int n = sels.length;
    fmrisc.ProofNavigator.Syntax.Selector[] result =
      new fmrisc.ProofNavigator.Syntax.Selector[n];
    for (int i=0; i<n; i++)
      result[i] = translate(sels[i]);
    return result;
  }
  
  /****************************************************************************
   * Translate selector
   * @param sel the selector to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Selector
    translate(Selector sel)
  {
    if (sel instanceof SelectorNumber)
      return new fmrisc.ProofNavigator.Syntax.SelectorNumber(
          translate(((SelectorNumber)sel).getNumber()));
    if (sel instanceof SelectorIndex)
      return new fmrisc.ProofNavigator.Syntax.SelectorIndex(
          translate(((SelectorIndex)sel).getIndex()));
    if (sel instanceof SelectorIdentifier)
      return new fmrisc.ProofNavigator.Syntax.SelectorIdentifier(
          translateReference(((SelectorIdentifier)sel).getIdentifier()));
    Main.internalError("Unknown selector: " + sel);
    return null;
  }
  
  // --------------------------------------------------------------------------
  //
  // Translate types
  //
  // --------------------------------------------------------------------------
  
  /****************************************************************************
   * Translate types.
   * @param types the types to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Type[] 
    translate(Type[] types)
  {
    int n = types.length;
    fmrisc.ProofNavigator.Syntax.Type[] result = 
      new fmrisc.ProofNavigator.Syntax.Type[n];
    for (int i=0; i<n; i++)
      result[i] = translate(types[i]);
    return result;
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Type 
    translate(Type type)
  {
    if (type instanceof Reference) return translateType((Reference)type);
    if (type instanceof AtomicType) return translate((AtomicType)type);
    if (type instanceof FunctionType) return translate((FunctionType)type);
    if (type instanceof TupleType) return translate((TupleType)type);
    if (type instanceof RecordType) return translate((RecordType)type);
    if (type instanceof ArrayType) return translate((ArrayType)type);
    if (type instanceof StateType) return translate((StateType)type);
    if (type instanceof SubType) return translate((SubType)type);
    if (type instanceof SubrangeType) return translate((SubrangeType)type);
    Main.internalError("unknown type: " + type);
    return null;
  }
 
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Type 
    translateType(Reference type)
  {
    return new fmrisc.ProofNavigator.Syntax.Reference(
        type.getName().getBase().toString());
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Type 
    translate(AtomicType type)
  {
    if (type instanceof BooleanType)
      return fmrisc.ProofNavigator.Semantics.Checking.getBooleanType();
    if (type instanceof BitType)
      return fmrisc.ProofNavigator.Semantics.Checking.getBitType();
    if (type instanceof NatType)
      return fmrisc.ProofNavigator.Semantics.Checking.getNatType();
    if (type instanceof IntType)
      return fmrisc.ProofNavigator.Semantics.Checking.getIntType();
    if (type instanceof RealType)
      return fmrisc.ProofNavigator.Semantics.Checking.getRealType();
    if (type instanceof StringType)
      return Construct.reference(stringTypeName);
    return Construct.reference(type.getName());
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.FunctionType
    translate(FunctionType type)
  {
    return new fmrisc.ProofNavigator.Syntax.FunctionType(
        translate(type.getDomain()), translate(type.getRange()));
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.TupleType
    translate(TupleType type)
  {
    return new fmrisc.ProofNavigator.Syntax.TupleType(
        translate(type.getBase()));
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.RecordType
    translate(RecordType type)
  {
    TypedIdentifier[] base = type.getBase();
    if (base.length == 0)
    {
      // CVCL does not allow empty records
      base = new TypedIdentifier[]{
          TypedIdentifier.construct(
              Identifier.construct("r"),
              IntType.construct())
      };
    }
    return new fmrisc.ProofNavigator.Syntax.RecordType(translate(base));
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.ArrayType
    translate(ArrayType type)
  {
    return new fmrisc.ProofNavigator.Syntax.ArrayType(
        translate(type.getIndex()), translate(type.getBase()));
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.Type
    translate(StateType type)
  {
    Main.internalError("unexpected construct");
    return null;
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.SubType
    translate(SubType type)
  {
    return new fmrisc.ProofNavigator.Syntax.SubType(
        translate(type.getPredicate()));
  }
  
  /****************************************************************************
   * Translate type.
   * @param type the type to be translated.
   * @return the translation
   ***************************************************************************/
  private fmrisc.ProofNavigator.Syntax.SubrangeType
    translate(SubrangeType type)
  {
    return new fmrisc.ProofNavigator.Syntax.SubrangeType(
        translate(type.getFrom()), translate(type.getTo()));
  }
}
