// ----------------------------------------------------------------------------
// PackageSymbol.java
// A symbol denoting a package.
// $Id: PackageSymbol.java,v 1.21 2010/10/29 16:31:10 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Tasks.*;

import java.io.*;
import java.util.*;

/**
 * 
 * A symbol denoting a value.
 *
 */
public final class PackageSymbol extends GlobalSymbolBase 
{    
  // the parent package (null for the top-level package)
  private final PackageSymbol parent;
  
  // the packages contained in the package
  private final SymbolTable<PackageSymbol> packages;
    
  // the classes contained in the package
  private final SymbolTable<ClassSymbol> classes;
  
  // the theories contained in the package
  private final SymbolTable<TheorySymbol> theories;
  
  // the type tables used for making names in this package unique
  private final TypeTable ttable;
  private final LogicTypeTable lttable;
  
  // the translator from program types to logical types
  private final TypeTranslator translator;
  
  // the stream used for printing errors
  private ErrorStream error;
  
  // the directory associated to the package (may be null)
  private File dir = null;
  
  // the folder of tasks related to this package
  private TaskFolder tasks = null;
  
  /***************************************************************************
   * Construct symbol for top-level package.
   * @param name the name of the package
   * @param error the stream for printing error messages
   * @param dir directory for writing information
   **************************************************************************/
  public PackageSymbol(Identifier name, ErrorStream error, File dir)
  {
    // a top-level package creates a new type table
    super(name, Name.construct(name));
    this.parent = null;
    this.packages = new SymbolTable<PackageSymbol>(error);
    this.classes = new SymbolTable<ClassSymbol>(error);
    this.theories = new SymbolTable<TheorySymbol>(error);
    this.ttable = new TypeTable();
    this.lttable = new LogicTypeTable(error);
    this.error = error;
    this.tasks = new TaskFolder(this, "package " + name, dir);
    this.translator = new TypeTranslator();
  }
  
  /***************************************************************************
   * Construct symbol for subpackage of another package.
   * @param parent the parent package.
   * @param name the short name of the package.
   * @param error the stream for printing error messages
   **************************************************************************/
  public PackageSymbol(PackageSymbol parent, Identifier name,
    ErrorStream error)
  {
    // a subpackage inherits the type table from the parent
    super(parent, name);
    this.parent = parent;
    this.packages = new SymbolTable<PackageSymbol>(error);
    this.classes = new SymbolTable<ClassSymbol>(error);
    this.theories = new SymbolTable<TheorySymbol>(error);
    this.ttable = parent.getTypeTable();
    this.lttable = parent.getLogicTypeTable();
    this.tasks = new TaskFolder(this, "package " + name, parent.getTasks());
    this.translator = parent.getTypeTranslator();
  }
  
  /****************************************************************************
   * Get parent package.
   * @return parent package (null for top-level package).
   ***************************************************************************/
  public PackageSymbol getParent()
  {
    return parent;
  }
  
  /****************************************************************************
   * Get top-level package of package
   * @return the top-level package for this package
   ***************************************************************************/
  public PackageSymbol getTopPackage()
  {
    PackageSymbol psymbol = this;
    PackageSymbol parent = psymbol.getParent();
    while (parent != null)
    {
      psymbol = parent;
      parent = psymbol.getParent();
    }
    return psymbol;
  }
  
  /****************************************************************************
   * Get type table of package.
   * @return the package's type table.
   ***************************************************************************/
  public TypeTable getTypeTable()
  {
    return ttable;
  }
  
  /****************************************************************************
   * Get logic type table of package.
   * @return the package's logic type table.
   ***************************************************************************/
  public LogicTypeTable getLogicTypeTable()
  {
    return lttable;
  }
  
  /****************************************************************************
   * Get type translator of package.
   * @return the package's type translator
   ***************************************************************************/
  public TypeTranslator getTypeTranslator()
  {
    return translator;
  }
  
  /****************************************************************************
   * Get set of packages held by package.
   * @return the set of packages.
   ***************************************************************************/
  public SymbolTable<PackageSymbol> getPackages()
  {
    return packages;
  }
  
  /****************************************************************************
   * Get set of packages held by package as collection.
   * @return the set of packages.
   ***************************************************************************/
  public Collection<PackageSymbol> getPackageSet()
  {
    return packages.getSymbols();
  }
  
  /****************************************************************************
   * Get set of classes held by package.
   * @return the set of classes.
   ***************************************************************************/
  public SymbolTable<ClassSymbol> getClasses()
  {
    return classes;
  }
  
  /****************************************************************************
   * Get set of classes held by package as collection.
   * @return the set of classes.
   ***************************************************************************/
  public Collection<ClassSymbol> getClassSet()
  {
    return classes.getSymbols();
  }
   
  /****************************************************************************
   * Get set of theories held by package.
   * @return the set of theories.
   ***************************************************************************/
  public SymbolTable<TheorySymbol> getTheories()
  {
    return theories;
  }
  
  /****************************************************************************
   * Get set of theories held by package as collection.
   * @return the set of theories.
   ***************************************************************************/
  public Collection<TheorySymbol> getTheorySet()
  {
    return theories.getSymbols();
  }
  
  /***************************************************************************
   * Set directory associated to package.
   * @param dir the directory.
   **************************************************************************/
  public void setDirectory(File dir)
  {
    this.dir = dir;
  }
  
  /***************************************************************************
   * Print symbol on output stream.
   * @param out the output stream.
   **************************************************************************/
  public void print(PrintWriter out)
  {
    out.print("package ");
    getLongName().print(out);
    if (dir != null)
    {
      out.print(" (directory ");
      out.print(dir.getAbsolutePath());
      out.print(")");
    }
  }
  
  /****************************************************************************
   * Return symbol for package denoted by name in current package;
   * an error message is printed if the lookup fails.
   * @param name a name which should denote a package.
   * @return the symbol or null if the name could not be looked up
   ***************************************************************************/
  public PackageSymbol getPackage(Name name)
  {
    return getPackage(name, true);
  }
  
  /****************************************************************************
   * Return symbol for package denoted by name in current package.
   * @param name a name which should denote a package.
   * @param verbose true if an error message is printed if the lookup fails.
   * @return the symbol or null if the name could not be looked up
   ***************************************************************************/
  public PackageSymbol getPackage(Name name, boolean verbose)
  {
    Identifier[] idents = name.idents;
    int n = idents.length;
    if (n == 0) Main.internalError("empty name encountered");
    Identifier ident = idents[0];
    PackageSymbol psymbol = packages.get(ident);
    if (psymbol == null)
    {
      if (verbose)
        error.newError(ident, "there is no top-level package " + ident);
      return null;
    }
    for (int i=1; i<n; i++)
    {
      ident = idents[i];
      PackageSymbol p = psymbol.getPackages().get(ident);
      if (p == null)
      {
        if (verbose)
          error.newError(ident, "there is no subpackage " + ident + 
              " in package " + psymbol.getLongName());
        return null;
      }
      psymbol = p;
    }
    return psymbol;
  }
  
  /***************************************************************************
   * Get the folder of tasks related to this package.
   * @return the task folder.
   **************************************************************************/
  public TaskFolder getTasks()
  {
    return tasks;
  }
}
