// ----------------------------------------------------------------------------
// ParamSymbols.java
// A set of parameterized symbols.
// $Id: ParamSymbols.java,v 1.10 2017/07/04 08:04:37 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2008, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Semantics;

import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;

/**
 * 
 * A set of parameterized symbols.
 *
 */
public final class ParamSymbols<S extends ParamSymbol> 
{
  // our symbol collection
  private final Map<TypeExpression[],S> table = 
    new MapClass<TypeExpression[], S>();
  
  // the stream to print messages on
  private ErrorStream error;
  
  /****************************************************************************
   * Construct symbol table with denoted error stream.
   * @param error the error stream.
   ***************************************************************************/
  public ParamSymbols(ErrorStream error)
  {
    this.error = error;
  }
  
  /****************************************************************************
   * Return collection of symbols in table.
   * @return the collection of all symbols.
   ***************************************************************************/
  public Collection<S> getSymbols()
  {
    return table.values();
  }
  
  /****************************************************************************
   * Puts symbol into table to be looked up for the symbol's argument types.
   * @param value the symbol.
   * @return true, if symbol could be added; 
   *         false, if symbol with same argument types was already in table
   *                (in this case the symbol is not added)
   ***************************************************************************/
  public boolean put(S value)
  {
    TypeExpression[] key = value.getArgumentTypes();
    if (get(key) != null) 
    {
      StringBuffer b = new StringBuffer();
      b.append("symbol " + value.getName() + "(");
      int n = key.length;
      if (n > 0)
      {
        for (int i=0; i<n-1; i++)
          b.append(key[i] + ", ");
        b.append(key[n-1]);
      }
      b.append(") is already defined in current environment");
      error.newError(b.toString());
      return false;
    }
    table.put(key, value);
    return true;
  }
  
  /****************************************************************************
   * Returns value associated to argument types.
   * @param atypes the array of argument types serving as key.
   * @return the associated symbol.
   ***************************************************************************/
  public S get(TypeExpression[] atypes)
  {
    TypeExpression[] key = atypes;
    return table.get(key);
  }
}