// ----------------------------------------------------------------------------
// MethodSpec.java
// A method specification.
// $Id: MethodSpec.java,v 1.20 2014/12/02 17:25:34 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Syntax.Spec;

import org.antlr.runtime.*;
import java.util.Vector;

import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Parser.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;

/**
 * 
 * A method specification.
 *
 */
public final class MethodSpec extends SpecificationBase
{
  private Name[] variables;
  private Name[] exceptions;
  private Formula preCondition;
  private Formula divergesCondition;
  private Formula postCondition;
  private Term[] terminationTerm;
  private boolean helper;
  
  public final static String variablesKeyword =  "assignable";
  public final static String exceptionsKeyword = "signals";
  public final static String preKeyword = "requires";
  public final static String divergesKeyword = "diverges";
  public final static String postKeyword ="ensures";
  public final static String terminationKeyword = "decreases";
  
  /****************************************************************************
   * Construct method specification from denoted text.
   * @param text the text of the specification.
   * @param error the stream for printing error messages
   * @param pos the source position of the specification
   * @return the method specification (null, if an error ocurred in parsing)
   ***************************************************************************/
  public static MethodSpec construct(String text, 
    ErrorStream error, SourcePosition pos)
  {
    MethodSpec spec = new MethodSpec(text);
    try
    {
      
      CharStream input = new ANTLRStringStream(text);
      SpecLangLexer lexer = new SpecLangLexer(input, error, pos);
      CommonTokenStream tokens = new CommonTokenStream(lexer);
      SpecLangParser parser = new SpecLangParser(tokens, error, pos, text);
      parser.methodspec(spec);  
      if (error.getErrorNumber() == 0)
      {
        // normalize precondition
        Formula pre = spec.getPreCondition();
        if (pre != null)
        {
          Formula pre0 = (Formula)PreStateCloner.clone(pre);
          pre0.setPosition(pre.getPosition());
          spec.setPreCondition(pre0);
        }
      }
    }
    catch (RecognitionException e)
    {
      error.newError("ill-formed specification (" + e.getMessage() + ")");
      return null;
    }
    return spec;
  }
  
  /***************************************************************************
   * Construct method specification from text (other fields are not set).
   * @param text the specification text.
   **************************************************************************/
  private MethodSpec(String text)
  {
    super(text);
  }
  
  /***************************************************************************
   * Get method frame i.e. list of variables that may be modified by method.
   * @return list of global variables modified by method (null, if none).
   **************************************************************************/
  public Name[] getVariables() { return variables; }
  
  /***************************************************************************
   * Set specification variables.
   * @param vs vector of variable names (may be null)
   **************************************************************************/
  public void setVariables(Vector<Name> vs)
  {
    if (vs == null) return;
    variables = vs.toArray(new Name[vs.size()]);
  }
  
  /***************************************************************************
   * Get list of exceptions that may be thrown by method.
   * @return list of exceptions that may be modified by method (null, if none).
   **************************************************************************/
  public Name[] getExceptions() { return exceptions; }
  
  /***************************************************************************
   * Set exceptions
   * @param es vector of exception names (may be null)
   **************************************************************************/
  public void setExceptions(Vector<Name> es)
  {
    if (es == null) return;
    exceptions = es.toArray(new Name[es.size()]);
  }
  
  /***************************************************************************
   * Get method precondition.
   * @return the method precondition (null, if none).
   **************************************************************************/
  public Formula getPreCondition() { return preCondition; }
  
  /***************************************************************************
   * Set precondition.
   * @param pre the method precondition (may be null).
   **************************************************************************/
  public void setPreCondition(Formula pre) 
  { 
    preCondition = pre; 
  }
  
  /***************************************************************************
   * Get method diverges condition.
   * @return the method diverges condition (null, if none).
   **************************************************************************/
  public Formula getDivergesCondition() { return divergesCondition; }
  
  /***************************************************************************
   * Set diverges condition.
   * @param div the method diverges condition (may be null).
   **************************************************************************/
  public void setDivergesCondition(Formula div)
  {
    divergesCondition = (Formula)PreStateCloner.clone(div);
  }
  
  /***************************************************************************
   * Get method postcondition.
   * @return the method postcondition (null, if none).
   **************************************************************************/
  public Formula getPostCondition() { return postCondition; }
  
  /***************************************************************************
   * Set postcondition.
   * @param post the method postcondition (may be null).
   **************************************************************************/
  public void setPostCondition(Formula post)
  {
    postCondition = post;
  }
  
  /***************************************************************************
   * Get termination term vector.
   * @return the loop termination term vector (null, if none).
   **************************************************************************/
  public Term[] getTerminationTerm() 
  { 
    return terminationTerm; 
  }
  
  /***************************************************************************
   * Set termination term. vector.
   * @param term the loop termination term vector (may be null).
   **************************************************************************/
  public void setTerminationTerm(Vector<Term> term)
  {
    int n = term.size();
    terminationTerm = new Term[n];
    term.toArray(terminationTerm);
    for (int i=0; i<n; i++)
      terminationTerm[i] = (Term)PreStateCloner.clone(terminationTerm[i]);
  }
  
  /***************************************************************************
   * Set helper status of a method.
   * @param helper true iff the method is a helper.
   **************************************************************************/
  public void setHelper(boolean helper)
  {
    this.helper = helper;
  }
  
  /***************************************************************************
   * Get helper status of a method.
   * @return true iff the method is a helper.
   **************************************************************************/
  public boolean isHelper()
  {
    return helper;
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitorOld visitor)
  {
    boolean done = visitor.visit(this);
    if (done) return;
    super.accept(visitor);
    if (preCondition != null) preCondition.accept(visitor);
    if (variables != null) accept(variables, visitor);
    if (exceptions != null) accept(exceptions, visitor);
    if (postCondition != null) postCondition.accept(visitor);
    if (divergesCondition != null) divergesCondition.accept(visitor);
    if (terminationTerm != null) accept(terminationTerm, visitor);
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   **************************************************************************/
  public void accept(ASTVisitor visitor)
  {
    visitor.visit(this);
  }
  
  /***************************************************************************
   * Accept visitor for a visit to the children of this node.
   * @param visitor the visitor who is accepted by the children of this node.
   **************************************************************************/
  public void acceptChildren(ASTVisitor visitor)
  {
    super.acceptChildren(visitor);
    if (preCondition != null) preCondition.accept(visitor);
    if (variables != null) accept(variables, visitor);
    if (exceptions != null) accept(exceptions, visitor);
    if (postCondition != null) postCondition.accept(visitor);
    if (divergesCondition != null) divergesCondition.accept(visitor);
    if (terminationTerm != null) accept(terminationTerm, visitor);
  }
}
