// ----------------------------------------------------------------------------
// InvariantTask.java
// The task of proving the correctness of a loop invariant.
// $Id: InvariantTask.java,v 1.14 2017/07/04 08:04:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.util.*;
import java.io.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Judgements.Simplification;
import fmrisc.ProgramExplorer.Proving.StateProblem;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.Statement;
import fmrisc.ProgramExplorer.Semantics.*;

/**
 * 
 * The task of proving the correctness of a loop invariant.
 *
 */
public class InvariantTask extends StateTaskBase
{
  private ParamSymbol method;
  // private Statement stat;
  private Formula goal;
  private Formula invariant;
  //private Formula cond;
  //private Formula body;
  private StateType stateType;
  private Set<VariableSymbol> variables;
  private Set<ClassSymbol> exceptions;
  private Vector<TheorySymbol> theories;
  private Declaration[][] decls;
  private ErrorStream out;
  
  /***************************************************************************
   * Construct an invariant verification task.
   * @param method the method in which the invariant arises.
   * @param stat the loop in which the invariant arises.
   * @param goal the goal to be proved (constructed outside)
   * @param invariant the invariant.
   * @param pos the position of the corresponding loop
   * @param cond the loop condition.
   * @param body the specification of the loop body.
   * @param parent the parent file.
   * @param stateType the type of the current state (null, if none)
   * @param variables the names of the variables modified in the loop body.
   * @param exceptions the names of the exceptions thrown in the loop body.
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public InvariantTask(ParamSymbol method, Statement stat,
    Formula goal, Formula invariant, SourcePosition pos,
    Formula cond, Formula body,
    File parent, StateType stateType, 
    Set<VariableSymbol> variables, Set<ClassSymbol> exceptions,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    super(taskName(method, stat), false, pos, parent, new ManualStrategy());
    this.method = method;
    // this.stat = stat;
    this.goal = Simplification.perform(goal);
    this.invariant = invariant;
    //this.cond = cond;
    //this.body = body;
    this.stateType = stateType;
    this.variables = variables;
    this.exceptions = exceptions;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
  }
  
  /****************************************************************************
   * Compute a specific task name.
   * @param method the method for the task.
   * @param stat the statement for the task.
   * @return a name that is probably different for different invariants.
   ***************************************************************************/
  private static String taskName(ParamSymbol method, Statement stat)
  {
    String gname = stat.toStringCore();
    String key = Util.hash(gname, 3);
    return "[" + method.getLongName() + ":" + key + "] invariant is preserved";
  }
  
  /****************************************************************************
   * Print linear representation of task.
   * @param out the stream to write to.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
    out.println("Type: verify invariant");
    out.println("Invariant: " + shortDescription(invariant.toString()));
  }
  
  /****************************************************************************
   * Translate task into a proving problem in state logic.
   * @return the translation result
   ***************************************************************************/
  public StateProblem stateProblem()
  {
    ClassSymbol csymbol = method.getClassSymbol();
    FlattenDeclarations.Result result = 
      FlattenDeclarations.flatten(theories, decls, goal, csymbol.getLocalTheory());
    Type stringType = csymbol.getPackage().getTypeTranslator().getStringType();
    return new StateProblem(result.decls, result.goal, result.map,
        stateType, stringType, variables, exceptions, out); 
  }
}
