// ----------------------------------------------------------------------------
// PreconditionTasks.java
// The tasks of proving that the preconditions of a method body are satisfied.
// $Id: PreconditionTasks.java,v 1.15 2011/09/30 15:47:31 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.jku.at>
// Copyright (C) 2010-, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.jku.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Syntax.Program.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * The tasks of proving that the preconditions of a method body are satisfied.
 *
 */
public final class PreconditionTasks extends ASTVisitorBase
{
  /***************************************************************************
   * Construct the precondition tasks for a given method.
   * @param method the symbol of a method with a specification.
   * @param preMap a mapping of body statements to pre-state knowledge.
   * @param stateType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  public static void generate(ParamSymbol method,
    Map<Statement, Formula> preMap,
    StateType stateType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    PreconditionTasks visitor =
      new PreconditionTasks(method, preMap, stateType, stateVars, stateExcs,
          theories, decls, out);
    Statement body = method.getDeclaration().getBody();
    body.accept(visitor);
  }

  private final ParamSymbol method;
  private final Map<Statement, Formula> preMap;
  private final StateType stateType;
  private final Name[] stateVars;
  private final Name[] stateExcs;
  private final Vector<TheorySymbol> theories;
  private final Declaration[][] decls;
  private final ErrorStream out;
  
  /***************************************************************************
   * Construct the precondition tasks for a given method.
   * @param method the symbol of a method with a specification.
   * @param preMap a mapping of body statements to pre-state knowledge.
   * @param stateType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories the theories on which the following depend in dependence order.
   * @param decls a sequence of declaration lists.
   * @param out the stream to which to write error messages.
   **************************************************************************/
  private PreconditionTasks(ParamSymbol method,
    Map<Statement, Formula> preMap,
    StateType stateType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls,
    ErrorStream out)
  {
    this.method = method;
    this.preMap = preMap;
    this.stateType = stateType;
    this.stateVars = stateVars;
    this.stateExcs = stateExcs;
    this.theories = theories;
    this.decls = decls;
    this.out = out;
  }
  
  // concrete statements
  public void visit(AssertionStatement stat) { visitCore(stat); }
  public void visit(AssignmentStatement stat) { visitCore(stat); }
  public void visit(BlockStatement stat) { visitCore(stat); }
  public void visit(BreakStatement stat) { visitCore(stat); }
  public void visit(ConditionalStatement stat) { visitCore(stat); }
  public void visit(ContinueStatement stat) { visitCore(stat); }
  public void visit(EmptyStatement stat) { visitCore(stat); }
  public void visit(ReturnStatement stat) { visitCore(stat); }
  public void visit(ThrowStatement stat) { visitCore(stat); }
  public void visit(TryCatchStatement stat) { visitCore(stat); }
  public void visit(VariableStatement stat) { visitCore(stat); }
  
  // groups of statements
  public void visit(WhileLoopStatement stat) { visitCore(stat); }
  public void visit(CallStatement stat) { visitCore(stat); }
  public void visit(ForLoopStatement stat) 
  { 
    // need to implement recursion over while replacement
    Statement[] stats = stat.whileReplacement.stats;
    stats[0].accept(this);
    stats[1].accept(this);
  }
  
  // counter for enumerating generated tasks
  private int counter = 0;
  
  /***************************************************************************
   * Generate precondition task for statement (if necessary)
   * @param stat the statement.
   * @return false, i.e. the statement is further recursively processed.
   **************************************************************************/
  private void visitCore(Statement stat)
  {
    StatJudgement judgement = stat.getJudgement();
    if (judgement == null) Main.internalError("no judgement");
    Formula goal = judgement.getPrecondition();
    if (goal == null) 
    {
      stat.acceptChildren(this);
      return;
    }
    Term state = OldState.construct();
    state.setType(stateType);
    goal = ImpliesFormula.construct(StateExecutes.construct(state), goal);
    Formula knowledge = preMap.get(stat);
    if (knowledge == null) Main.internalError("no knowledge: " + stat + " " + goal);
    knowledge = Simplification.perform(knowledge);
    goal = Simplification.perform(goal);
    TaskFolder folder = method.getPreconditions();
    Task task = new PreconditionTask(method, counter, stat, knowledge, goal,
        folder.getDirectory(),
        stateType, stateVars, stateExcs, theories, decls, out);
    counter = counter+1;
    folder.add(task);
    stat.acceptChildren(this);
  }
  
  /**
   * 
   * The task of proving that a statement precondition is satisfied.
   *
   */
  private static final class PreconditionTask extends StateTaskBase
  {
    private final ParamSymbol method;
    private final Formula knowledge;
    private final Formula goal;
    private final StateType stateType;
    private final Name[] stateVars;
    private final Name[] stateExcs;
    private final Vector<TheorySymbol> theories;
    private final Declaration[][] decls;
    private final ErrorStream out;

    /***************************************************************************
     * Construct a precondition task for a given method.
     * @param method the symbol of a method with a specification.
     * @param counter a unique counter for the precondition within the method.
     * @param statement a statement in the body of the method.
     * @param parent the parent of the new directory for this task
     * @param stateType the type of the current state (null, if none)
     * @param stateVars the names of the modifiable variables (null, if none)
     * @param stateExcs the names of the throwable exceptions (null, if none)
     * @param theories the theories on which the following depend in dependence order.
     * @param decls a sequence of declaration lists.
     * @param out the stream to which to write error messages.
     **************************************************************************/
    public PreconditionTask(ParamSymbol method, int counter,
      Statement statement, Formula knowledge, Formula goal,
      File parent, StateType stateType, Name[] stateVars, Name[] stateExcs,
      Vector<TheorySymbol> theories, Declaration[][] decls,
      ErrorStream out)
    {
      super("[" + method.getLongName() + ":" + counter + "] " + 
          statement.getKind() + " precondition", false,
          statement.getPosition(), parent, 
          new ManualStrategy());
      this.method = method;
      this.knowledge = knowledge;
      this.goal = goal;
      this.stateType = stateType;
      this.stateVars = stateVars;
      this.stateExcs = stateExcs;
      this.theories = theories;
      this.decls = decls;
      this.out = out;
    }
    
    /*************************************************************************
     * Print linear representation of task.
     * @param out the stream to write to.
     ************************************************************************/
    public void print(PrintWriter out)
    {
      super.print(out);
      out.println("Type: verify statement precondition");
    }
    
    /*************************************************************************
     * Translate task into a proving problem in state logic.
     * @return the translation result
     ************************************************************************/
    public StateProblem stateProblem()
    {
      Formula formula = ImpliesFormula.construct(knowledge, goal);
      ClassSymbol csymbol = method.getClassSymbol();
      FlattenDeclarations.Result result = 
        FlattenDeclarations.flatten(theories, decls, formula, csymbol.getLocalTheory());
      Type stringType = csymbol.getPackage().getTypeTranslator().getStringType();
      return new StateProblem(result.decls, result.goal, result.map,
          stateType, stringType, Normalization.modifiableVariables(stateVars), 
          NameUtils.getClassSymbols(stateExcs), out);    
    }
  }
}
