// ----------------------------------------------------------------------------
// TypeCheckingTask.java
// The task of proving a type-checking condition.
// $Id: TypeCheckingTask.java,v 1.29 2011/06/29 15:48:31 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------
package fmrisc.ProgramExplorer.Tasks;

import java.io.*;
import java.util.*;

import fmrisc.ProgramExplorer.*;
import fmrisc.ProgramExplorer.Syntax.*;
import fmrisc.ProgramExplorer.Syntax.Logic.*;
import fmrisc.ProgramExplorer.Semantics.*;
import fmrisc.ProgramExplorer.InOut.*;
import fmrisc.ProgramExplorer.Proving.*;
import fmrisc.ProgramExplorer.Judgements.*;

/**
 * 
 * The task of proving a type-checking condition.
 *
 */
public class TypeCheckingTask extends StateTaskBase 
{
  private TheorySymbol tsymbol;
  private StateType resultType;
  private Name[] stateVars;
  private Name[] stateExcs;
  private Vector<TheorySymbol> theories;
  private Declaration[][] decls;
  private Formula goal;
  private ErrorStream out;
  
  /****************************************************************************
   * Create task of proving a type-checking condition.
   * @param tsymbol the theory of the task.
   * @param name the name of the task.
   * @param position a related source position (may be null)
   * @param dir a directory for writing persistent information
   * @param resultType the type of the current state (null, if none)
   * @param stateVars the names of the modifiable variables (null, if none)
   * @param stateExcs the names of the throwable exceptions (null, if none)
   * @param theories a vector of theories on which the following depend.
   * @param decls a sequence of declaration lists.
   * @param goal the goal to be proved.
   * @param out the stream to which to write error messages.
   ***************************************************************************/
  public TypeCheckingTask(TheorySymbol tsymbol,
    String name, SourcePosition position, File dir,
    StateType resultType, Name[] stateVars, Name[] stateExcs,
    Vector<TheorySymbol> theories, Declaration[][] decls, Formula goal,
    ErrorStream out)
  {
    // create task with auto strategy
    super(taskName(tsymbol, name, goal), false, position, dir, new AutoStrategy());
    this.tsymbol = tsymbol;
    this.resultType = resultType;
    this.stateVars = stateVars;
    this.stateExcs = stateExcs;
    this.theories = theories;
    this.decls = decls;
    this.goal = goal;
    this.out = out;
  }
  
  /****************************************************************************
   * Compute from base name and goal formula a specific task name.
   * @param tsymbol the theory symbol
   * @param base the base name of the task.
   * @param goal the goal formula.
   * @return a name that is probably different for different goal formulas.
   ***************************************************************************/
  private static String taskName(TheorySymbol tsymbol, String base, Formula goal)
  {
    String gname = goal.toString();
    String key = Util.hash(gname, 3);
    String tkey = tsymbol == null ? "(no theory)" : tsymbol.getLongName().toString();
    return "[" + tkey + ":" + key + "] " + base;
  }
  
  /****************************************************************************
   * Determine whether new task is to be resumed.
   * @return true if new task is to be resumed.
   ***************************************************************************/
  public boolean resumeOnNew()
  {
    return true;
  }
  
  /****************************************************************************
   * Print linear representation of task.
   * @param out the stream to write to.
   ***************************************************************************/
  public void print(PrintWriter out)
  {
    super.print(out);
    out.println("Type: verify type checking condition");
    out.println("Goal formula: " + shortDescription(goal.toString()));
  }
     
  /****************************************************************************
   * Translate task into a proving problem in state logic.
   * @return the translation result
   ***************************************************************************/
  public StateProblem stateProblem()
  {
    FlattenDeclarations.Result result = 
      FlattenDeclarations.flatten(theories, decls, goal, tsymbol);
    Type stringType = tsymbol.getPackage().getTypeTranslator().getStringType();
    return new StateProblem(result.decls, result.goal, result.map,
        resultType, stringType, Normalization.modifiableVariables(stateVars), 
        NameUtils.getClassSymbols(stateExcs), out);    
  }
}
