// ----------------------------------------------------------------------------
// Auto.java
// Automatically generate instantiations of quantified formulas.
// $Id: Auto.java,v 1.4 2008/10/13 09:32:00 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import org.w3c.dom.*;

import fmrisc.ProofNavigator.State;
import fmrisc.ProofNavigator.Proving.*;

/**
 * 
 * Command "auto": try to close state by automatic instantiation 
 * of quantified formulas.
 *
 */
public final class Auto extends ProofCommandBase
{
  public final static String Name = "auto";
  private String[] labels; // labels of formulas to be instantiated (may be null)
  
  /***************************************************************************
   * Create a "auto" command.
   * @param labels labels of formulas to be simplified (may be null).
   **************************************************************************/
  public Auto(String[] labels)
  {
    super(Name);
    this.labels = labels;
  }
  
  /****************************************************************************
   * Process command "auto": 
   * @param state the proof state to which the command is applied
   * @return null or array of single new proof state.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    // it is the command itself that simplfies the child state 
    // to decide whether the command was applicable and
    // then extends the current state correspondingly
    // (see also its use in command "autostar")
    
    // instantiate all appropriate formulas in state 
    // and check whether state can be closed
    if (labels == null)
    {
      ProofState state1 = Instantiating.instantiate(state);
      if (state1 != null) simplify(state, state1); 
      return null;
    }
    
    // instantiate single selected formula in state
    // and check whether state can be closed
    int n = labels.length;
    if (n == 1)
    {
      Formula formula = state.getFormula(labels[0]);
      if (formula == null)
      {
        error("there is no formula with label [" + labels[0] + "]");
        return null;
      }
      ProofState state1 = Instantiating.instantiate(formula);
      if (state1 != null) simplify(state, state1); 
      return null;
    }
    
    // instantiate multiple selected formulas in state
    // and check whether state can be closed
    Formula[] formulas = new Formula[n];
    for (int i=0; i<n; i++)
    {
      formulas[i] = state.getFormula(labels[i]);
      if (formulas[i] == null)
      {
        error("there is no formula with label [" + labels[i] + "]");
        return null;
      }
    }
    ProofState state1 = Instantiating.instantiate(formulas);
    if (state1 != null) simplify(state, state1); 
    return null;
  }
  
  /****************************************************************************
   * Check whether state1 can be closed; if yes, extend state by state1.
   * @param state parent state.
   * @param state1 potential child of state.
   ***************************************************************************/
  private void simplify(ProofState state, ProofState state1)
  {
    ProofState[] children = State.getReplayChildren();
    State.setReplayChildren(null);
    state1.simplify(false);
    State.setReplayChildren(children);
    if (!State.isAborted() && state1.isClosed()) 
      state.extend(this, new ProofState[] { state1 });
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    if (labels == null) return;
    int n = labels.length;
    for (int i=0; i<n; i++)
    {
      Node node0 = newNode("label");
      node.appendChild(node0);
      node0.appendChild(newTextNode(labels[i]));
    }
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n == 0) return new Auto(null);
    String[] labels = new String[n];
    for (int i=0; i<n; i++)
    {
      NodeList child = children.item(i).getChildNodes();
      if (child.getLength() != 1) return null;
      labels[i] = toString(child.item(0));
      if (labels[i] == null) return null;
    }
    return new Auto(labels);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    if (labels == null) return;
    int n = labels.length;
    out.print(" ");
    for (int i=0; i<n; i++)
    {
      out.print(labels[i]);
      if (i<n-1) out.print(", ");
    }
  }
}
