// ----------------------------------------------------------------------------
// TypeAxiom.java
// The "typeaxiom" command.
// $Id: TypeAxiom.java,v 1.8 2012/10/16 15:49:36 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Commands;

import java.io.*;
import org.w3c.dom.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Proving.*;
import fmrisc.ProofNavigator.Semantics.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * The "typeaxiom" command: instantiate a quantified type axiom.
 *
 */
public final class TypeAxiom extends ProofCommandBase
{
  public final static String Name = "typeaxiom";
  private Expression exp;      // expression whose type axiom is instantiated
  private Expression[] values; // the instantiation values
  
  /***************************************************************************
   * Create an "typeaxiom" command.
   * @param exp expression whose type axiom is instantiated.
   * @param values the instantiation values.
   **************************************************************************/
  public TypeAxiom(Expression exp, Expression[] values)
  {
    super(Name);
    this.exp = exp;
    this.values = values;
  }
  
  /****************************************************************************
   * Process command "typeaxiom": 
   * instantiate universal type axiom of denoted expression.
   * @param state the proof state to which the command is applied
   * @return null or array of one or two new proof states.
   ***************************************************************************/
  protected ProofState[] processCore(ProofState state)
  {
    Type etype = Checking.processDefinitionTerm(exp);
    if (etype == null) return null;
    TypeExpression te = Checking.clearSubtypes(etype, exp);
    if (te == null)
    {
      error("no type axiom associated to value " + exp.toString());
      return null;
    }    
    Expression exp0 = te.getExpression();
    if (exp0 == null)
    {
      error("no type axiom associated to value " + exp.toString());
      return null;
    }
    if (!(exp0 instanceof ForallFormula))
    {
      error("type axiom " + exp0.toString() + " is not universally quantified");
      return null;
    }
    
    ForallFormula f = (ForallFormula)exp0;
    TypedIdentifier[] vars = f.getVariables();
    int n = vars.length;
    if (n != values.length)
    {
      error ("wrong number of instantiation values (" +
        n + " expected, " + values.length + " given)");
      return null;
    }
    Expression base = f.getBase();

    // need to match types according to their CVCL semantics
    Environment env = getEnvironment();
    Checking.setWeakerTypes(true);
    ValueDeclIdentifier[] idents = new ValueDeclIdentifier[n];
    Context context = State.getContext();
    context.open();
    for (int i=0; i<n; i++)
    {
      idents[i] = vars[i].getIdentifier();
      Expression value = values[i];
      Type type = typeCheckTerm(env, value);
      if (type == null) 
      {
        context.abort();
        Checking.setWeakerTypes(false);
        return null;
      }
      TypedIdentifier var = vars[i];
      Type type0 = var.getIdentifier().getSymbol().getType();
      if (!Checking.hasType(value, type, type0, true)) 
      {
        context.abort();
        error("type " + type.toString() + " of value " + value.toString() +
          " does not match variable declaration " + var.toString());
        Checking.setWeakerTypes(false);
        return null;
      }
    }
    Checking.setWeakerTypes(false);
    Expression tcc = context.getFormula();
    context.abort();
    
    Formula[] assumptions = state.getAssumptions();
    Formula[] goals = state.getGoals();
    
    ProofState state1 = null;
    if (tcc != null)
    {
      // create child state to prove tcc
      state1 = new ProofState(state, env);
      state1.setAssumptions(state1.copy(assumptions));
      Formula goals0[] = { new Formula(state1, tcc, true) };
      state1.setGoals(goals0);
    }
    
    // create substitution
    Expression base0 = Substitute.substitute(base, idents, values, false);
    
    ProofState state0 = new ProofState(state, state.getEnvironment());
    Formula formula0 = new Formula(state0, base0, false);
    state0.setAssumptions(state0.add(assumptions, formula0));
    state0.setGoals(state0.copy(goals));
      
    if (tcc != null)
      state0.setAxioms(state0.add(state0.getAxioms(),
        new Formula(state0, tcc, false)));
    if (state1 == null)
      return new ProofState[] { state0 };
    else
      return new ProofState[] { state0, state1 };
  }
  
  /***************************************************************************
   * Set external representation of command in a DOM node.
   * @param node the node in which the representation is to be set
   *        (by default none, to be overridden by command class)
   **************************************************************************/
  protected void setRepresentation(Node node)
  {
    node.appendChild(toNode(exp));
    int n = values.length;
    for (int i=0; i<n; i++)
      node.appendChild(toNode(values[i]));
  }
  
  /***************************************************************************
   * Convert DOM node to command.
   * @param node the DOM representation of the command.
   * @return the command (null, if an error occured).
   **************************************************************************/
  public static Command newCommand(Node node)
  {
    NodeList children = node.getChildNodes();
    int n = children.getLength();
    if (n == 0)
    {
      State.setError("child expected: " + node.toString());
      return null;
    }
    Expression exp = toExpression(children.item(0));
    if (exp == null) return null;
    Expression[] values = new Expression[n-1];
    for (int i=1; i<n; i++)
    {
      values[i-1] = toExpression(children.item(i));
      if (values[i-1] == null) return null;
    }
    return new TypeAxiom(exp, values);
  }
  
  /***************************************************************************
   * Prints text representation on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print(getName());
    out.print(" ");
    int n = values.length;
    for (int i=0; i<n; i++)
    {
      values[i].print(out);
      if (i+1 < n) out.print(", ");
    }
    out.print(" in ");
    exp.print(out);
  }
}
