// ----------------------------------------------------------------------------
// MathML.java
// Conversion of AST to HTML/MathML presentation markup.
// $Id: MathML.java,v 1.28 2013/04/06 11:29:40 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Syntax.*;

import org.w3c.dom.*;

/**
 *
 * Conversion of AST to HTML/MathML presentation markup.
 *
 **/
public final class MathML
{
  private Document document; // the document in which to create new nodes
  
  /***************************************************************************
   * Create MathML node in document.
   * @param document the document in which to create the node.
   * @return the MathML node.
   **************************************************************************/
  public static Element createMathNode(Document document)
  {
    Element mnode = document.createElement("math");
    mnode.setAttribute("xmlns", "http://www.w3.org/1998/Math/MathML");
    return mnode;
  }
  
  /****************************************************************************
   * Create converter that creates DOM nodes in the specified document.
   * @param document the document in which to create the DOM nodes.
   ***************************************************************************/
  public MathML(Document document)
  {
    this.document = document;
  }
  
  /***************************************************************************
   * Convert declaration to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the declaration to be converted.
   ***************************************************************************/
  public void appendDecl(Node parent, Declaration object)
  {
    appendDeclCore(parent, object);
  }
  
  /***************************************************************************
   * Convert declaration to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the declaration to be converted.
   ***************************************************************************/
  private void appendDeclCore(Node parent, Declaration object)
  {
    if (object instanceof TypeDeclaration)
    {
      TypeDeclaration decl = (TypeDeclaration)object;
      Identifier ident = decl.getIdentifier();
      Type type = decl.getType();
      appendIdentifier(parent, ident);
      appendConnector(parent, "\u2208"); // &isin;
      appendKeyword(parent, "type");
      if (type == null) return;
      appendInfixPredicate(parent, "=");
      appendType(parent, type);
      return;
    }
    if (object instanceof ValueDeclaration)
    {
      ValueDeclaration decl = (ValueDeclaration)object;
      Identifier ident = decl.getIdentifier();
      Type type = decl.getType();
      Expression value = decl.getValue();
      appendIdentifier(parent, ident);
      if (type != null)
      {
        appendConnector(parent, "\u2208"); // &isin;
        appendType(parent, type);
      }
      if (value != null)
      {
        appendInfixPredicate(parent, "=");
        appendExp(parent, value);
      }
      return;
    } 
    if (object instanceof FormulaDeclaration)
    {
      FormulaDeclaration decl = (FormulaDeclaration)object;
      Identifier ident = decl.getIdentifier();
      Expression formula = decl.getFormula();
      if (decl instanceof AxiomDeclaration)
        appendPrefixConnective(parent, "axiom");
      appendIdentifier(parent, ident);
      appendInfixConnective(parent, "\u2261"); // &equiv;
      appendExp(parent, formula);
      return;
    }
    System.out.println("Unknown declaration: " + object.toString());
    System.exit(-1);
  }
 
  /***************************************************************************
   * Convert type to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the type to be converted.
   ***************************************************************************/
  public void appendType(Node parent, Type object)
  {
    Node node = parent; // appendNode(parent, "mrow");
    appendTypeCore(node, object);
  }
  
  /***************************************************************************
   * Convert type to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the type to be converted.
   ***************************************************************************/
  private void appendTypeCore(Node parent, Type object)
  {
    if (object instanceof Identifier)
    {
      Identifier ident = (Identifier)object;
      appendIdentifier(parent, ident);
      return;
    }
    if (object instanceof ArrayType)
    {
      ArrayType type = (ArrayType) object;
      Type index = type.getIndex();
      Type base = type.getBase();
      appendPrefixConnective(parent, "array");
      appendType(parent, index);
      appendInfixConnective(parent, "of");
      appendType(parent, base);
      return;
    }
    if (object instanceof AtomicType)
    {
      AtomicType type = (AtomicType) object;
      String name = type.getName();
      if (name.equals("BOOLEAN"))
      {
        Element node = document.createElement("mo");
        node.setAttribute("mathvariant", "double-struck");
        parent.appendChild(node);
        if (State.booleanAsLogical())
          appendTextNode(node, "LOGICAL");
        else
          appendTextNode(node, "B");
      }
      else if (name.equals("BIT"))
      {
        Element node = document.createElement("mo");
        node.setAttribute("mathvariant", "double-struck");
        parent.appendChild(node);
        if (State.booleanAsLogical())
          appendTextNode(node, "B");
        else
          appendTextNode(node, "BIT");
      }
      else if (name.equals("NAT"))
        appendOperator(parent, "\u2115"); // &naturals;
      else if (name.equals("INT"))
        appendOperator(parent, "\u2124"); // &integers;
      else if (name.equals("REAL"))
        appendOperator(parent, "\u211D"); // &reals;
      else
        appendIdentifier(parent, name);
      return;
    } 
    if (object instanceof FunctionType)
    {
      FunctionType type = (FunctionType) object;
      Type domain[] = type.getDomain();
      int n = domain.length;
      Type range = type.getRange();
      if (n == 1)
      {
        if (domain[0] instanceof FunctionType ||
            domain[0] instanceof ArrayType)
        {
          Node node = parent; // appendNode(parent, "mrow");
          appendOperator(node, "(");
          appendType(node, domain[0]);
          appendOperator(node, ")");
        }
        else
          appendType(parent, domain[0]);
      }
      else
      {
        appendOperator(parent, "("); 
        for (int i=0; i<n; i++)
        {
          appendType(parent, domain[i]);
          if (i+1 < n) appendOperator(parent, ",");
        }
        appendOperator(parent, ")"); 
      }
      appendInfixFunction(parent, "\u2192"); // &rightarrow;
      appendType(parent, range);
      return;
    }
    if (object instanceof RecordType)
    {
      RecordType type = (RecordType) object;
      TypedIdentifier[] base = type.getBase();
      appendOperator(parent, "[#");
      Node node = parent; // appendNode(parent, "mrow");
      int n = base.length;
      for (int i=0; i<n; i++)
      {
        appendTypedIdentifier(node, base[i]);
        if (i+1 < n) appendOperator(node, ",");
      }
      appendOperator(parent, "#]");
      return;
    }
    if (object instanceof SubrangeType)
    {
      SubrangeType type = (SubrangeType) object;
      Expression from = type.getFrom();
      Expression to = type.getTo();
      appendOperator(parent, "[");
      appendExp(parent, from);
      appendOperator(parent, ",");
      appendExp(parent, to);
      appendOperator(parent, "]");
      return;
    }
    if (object instanceof SubType)
    {
      SubType type = (SubType) object;
      Expression pred = type.getPredicate();
      appendOperator(parent, "subtype");
      appendOperator(parent, "(");
      appendExp(parent, pred);
      appendOperator(parent, ")");
      return;
    }
    if (object instanceof TupleType)
    {
      TupleType type = (TupleType) object;
      Type[] base = type.getBase();
      appendOperator(parent, "[");
      Node node = parent; // appendNode(parent, "mrow");
      int n = base.length;
      for (int i=0; i<n; i++)
      {
        appendType(node, base[i]);
        if (i+1 < n) appendOperator(node, ",");
      }
      appendOperator(parent, "]"); 
      return;
    }
    
    if (object instanceof LetType)
    {
      LetType type = (LetType)object;
      TypeDeclaration[] decls = type.getDeclarations();
      Type body = type.getBase();
      appendPrefixConnective(parent, "let");
      Node node = parent; // appendNode(parent, "mrow");
      int n = decls.length;
      for (int i=0; i<n; i++)
      {
        appendDecl(node, decls[i]);
        if (i+1 < n) appendOperator(node, ",");
      }
      appendInfixConnective(parent, "in");
      appendType(parent, body);
      return;
    }
    
    System.out.println("Unknown type: " + object.toString());
    System.exit(-1);
  }

  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup
   * surrounded by parentheses.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to convert.
   **************************************************************************/
  private void appendParens(Node parent, Expression exp)
  {
    appendOperator(parent, "(");
    appendExp(parent, exp);
    appendOperator(parent, ")");
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup
   * surrounded by parentheses, if necessary.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to convert.
   * @param the priority level of the parent expression. If exp has a higher
   *        priority, then parentheses are added to surround the markup.
   **************************************************************************/
  private void appendExp(Node parent, Expression exp, int priority)
  {
    int p = exp.getPriority();
    if (p <= priority)
      appendExp(parent, exp);
    else
      appendParens(parent, exp);
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to convert.
   **************************************************************************/
  public void appendExp(Node parent, Expression exp)
  {
    Node node = parent; // appendNode(parent, "mrow");
    appendExpCore(node, exp);
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendExpCore(Node parent, Expression exp)
  {
    if (exp instanceof fmrisc.ProofNavigator.Syntax.Number)
    {
      fmrisc.ProofNavigator.Syntax.Number object = (fmrisc.ProofNavigator.Syntax.Number)exp;
      appendNumber(parent, object);
      return;
    }
    if (exp instanceof Identifier)
    {
      Identifier object = (Identifier)exp;
      appendIdentifier(parent, object);
      return;
    }
    if (exp instanceof ApplicationExpression)
    {
      ApplicationExpression object = (ApplicationExpression)exp;
      int priority = object.getPriority();
      Expression fun = object.getFunction();
      appendExp(parent, fun, priority);
      // appendOperator(parent, "\u2061"); // &ApplyFunction;
      Node anode = parent; // appendNode(parent, "mrow");
      appendOperator(anode, "(");
      Node bnode = anode; // appendNode(anode, "mrow");
      Expression[] args = object.getArguments();
      int n = args.length;
      for (int i=0; i<n; i++)
      {
        appendExp(bnode, args[i]);
        if (i+1 < n) appendOperator(bnode, ",");
      }
      appendOperator(anode, ")");
      return;
    }
    if (exp instanceof BinaryExpression)
    {
      BinaryExpression object = (BinaryExpression)exp;
      int priority = object.getPriority();
      Expression arg1 = object.getFirst();
      Expression arg2 = object.getSecond();
      if (object instanceof DividesTerm)
      {
        Node node = appendNode(parent, "mfrac");
        Node num = appendNode(node, "mrow");
        appendExp(num, arg1);
        Node denom = appendNode(node, "mrow");
        appendExp(denom, arg2);
        return;
      }
      if (object instanceof PowerTerm)
      {
        Node node = appendNode(parent, "msup");
        Node base = appendNode(node, "mrow");
        appendExp(base, arg1, priority);
        Node power = appendNode(node, "mrow");
        appendExp(power, arg2, priority);
        return;
      }
      if (object instanceof MinusTerm &&
          (arg2 instanceof PlusTerm || arg2 instanceof MinusTerm))
      {
        appendExp(parent, arg1, priority);
        appendInfixFunction(parent, "-");
        appendParens(parent, arg2);
        return;
      }
      if (object instanceof EquivalentFormula ||
          object instanceof NotEquivalentFormula)
      {
        if (arg1 instanceof EquivalentFormula ||
            arg1 instanceof NotEquivalentFormula ||
            arg1 instanceof ImpliesFormula)
          appendParens(parent, arg1);
        else
          appendExp(parent, arg1, priority);
        if (object instanceof EquivalentFormula)
          appendInfixConnective(parent, "\u21D4"); // &iff;
        else
          appendInfixConnective(parent, "xor");
        if (arg2 instanceof EquivalentFormula ||
            arg2 instanceof NotEquivalentFormula ||
            arg2 instanceof ImpliesFormula)
          appendParens(parent, arg2);
        else
          appendExp(parent, arg2, priority);
        return;
      }
      if (object instanceof ImpliesFormula)
      {
        if (arg1 instanceof EquivalentFormula ||
            arg1 instanceof NotEquivalentFormula ||
            arg1 instanceof ImpliesFormula)
          appendParens(parent, arg1);
        else
          appendExp(parent, arg1, priority);
        appendInfixConnective(parent, "\u21D2"); // &Implies;
        if (arg2 instanceof EquivalentFormula ||
            arg2 instanceof NotEquivalentFormula)
          appendParens(parent, arg2);
        else
          appendExp(parent, arg2, priority);
        return;
      }
      appendExp(parent, arg1, priority);
      if (object instanceof AndFormula)
       appendInfixConnective(parent, "\u2227"); // &and;
      else if (object instanceof BitAndTerm)
        appendInfixConnective(parent, "&"); 
      else if (object instanceof EqualsFormula)
        appendInfixPredicate(parent, "=");
      else if (object instanceof GreaterEqualFormula)
        appendInfixPredicate(parent, "\u2265"); // &ge;
      else if (object instanceof GreaterFormula)
        appendInfixPredicate(parent, ">");
      else if (object instanceof LessFormula)
        appendInfixPredicate(parent, "<");
      else if (object instanceof MinusTerm)
        appendInfixFunction(parent, "-");
      else if (object instanceof LessEqualFormula)
        appendInfixPredicate(parent, "\u2264"); // &le;
      else if (object instanceof NotEqualsFormula)
        appendInfixPredicate(parent, "\u2260"); // &ne;
      else if (object instanceof NotEquivalentFormula)
        appendInfixConnective(parent, "xor");
      else if (object instanceof OrFormula)
        appendInfixConnective(parent, "\u2228"); // &or;
      else if (object instanceof BitOrTerm)
        appendInfixConnective(parent, "|"); 
      else if (object instanceof PlusTerm)
        appendInfixFunction(parent, "+");
      else if (object instanceof TimesTerm)
        // appendInfixFunction(parent, "\u00D7"); // &times;
        appendInfixFunction(parent, "\u00B7"); // &centerdot;
      else
      {
        System.out.println("Unknown binary expression: " + exp.toString());
        System.exit(-1);
      }
      appendExp(parent, arg2, priority);
      return;
    }
    if (exp instanceof IfThenElseExpression)
    {
      IfThenElseExpression object = (IfThenElseExpression)exp;
      Expression exp1 = object.getCondition();
      Expression exp2 = object.getThenBranch();
      Expression exp3 = object.getElseBranch();
      appendInfixConnective(parent, "if");
      appendExp(parent, exp1);
      appendInfixConnective(parent, "then");
      appendExp(parent, exp2);
      while (exp3 instanceof IfThenElseExpression)
      {
        object = (IfThenElseExpression)exp3;
        exp1 = object.getCondition();
        exp2 = object.getThenBranch();
        exp3 = object.getElseBranch();
        appendInfixConnective(parent, "elsif");
        appendExp(parent, exp1);
        appendInfixConnective(parent, "then");
        appendExp(parent, exp2);
      }
      appendInfixConnective(parent, "else");
      appendExp(parent, exp3);
      appendInfixConnective(parent, "endif");
      return;
    }
    if (exp instanceof LetExpression)
    {
      LetExpression object = (LetExpression)exp;
      ValueDeclaration[] decls = object.getDeclarations();
      Expression body = object.getBase();
      appendPrefixConnective(parent, "let");
      Node node = parent; // appendNode(parent, "mrow");
      int n = decls.length;
      for (int i=0; i<n; i++)
      {
        appendDecl(node, decls[i]);
        if (i+1 < n) appendOperator(node, ",");
      }
      appendInfixConnective(parent, "in");
      appendExp(parent, body);
      return;
    }
    if (exp instanceof Logical)
    {
      Logical object = (Logical)exp;
      if (State.booleanAsLogical())
      {
        if (object.getValue())
          appendKeyword(parent, "logicaltrue");
        else
          appendKeyword(parent, "logicalfalse");
      }
      else
      {
        if (object.getValue())
          appendKeyword(parent, "true");
        else
          appendKeyword(parent, "false");
      }
      return;
    }
    if (exp instanceof BitLogical)
    {
      BitLogical object = (BitLogical)exp;
      if (State.booleanAsLogical())
      {
        if (object.getValue())
          appendKeyword(parent, "true");
        else
          appendKeyword(parent, "false");
      }
      else
      {
        if (object.getValue())
          appendKeyword(parent, "bittrue");
        else
          appendKeyword(parent, "bitfalse");
      }
      return;
    }
    if (exp instanceof QuantifiedExpression)
    {
      QuantifiedExpression object = (QuantifiedExpression)exp;
      int priority = object.getPriority();
      TypedIdentifier[] vars = object.getVariables();
      Expression body = object.getBase();
      String quantifier = null;
      if (exp instanceof ForallFormula)
        quantifier = "\u2200"; // &forall;
      else if (exp instanceof ExistsFormula)
        quantifier = "\u2203"; // &exists;
      else if (exp instanceof LambdaTerm)
        quantifier = "\u03BB"; // &lambda;
      else if (exp instanceof ArrayTerm)
        quantifier = "array";
      else
      {
        System.out.println("Unknown quantifier: " + exp.toString());
        System.exit(-1);
      }
      appendQuantifier(parent, quantifier);
      Node node = parent; // appendNode(parent, "mrow");
      int n = vars.length;
      for (int i=0; i<n; i++)
      {
        appendTypedIdentifier(node, vars[i]);
        if (i+1<n) appendOperator(node, ",");
      }
      appendPunctuation(parent, ":");
      appendExp(parent, body, priority);
      return;
    }
    if (exp instanceof RecordTerm)
    {
      RecordTerm object = (RecordTerm)exp;
      appendOperator(parent, "(#");
      Node node = parent; // appendNode(parent, "mrow");
      ValuedIdentifier[] idents = object.getBase();
      int n = idents.length;
      for (int i=0; i<n; i++)
      {
        appendValuedIdentifier(node, idents[i]);
        if (i+1 < n) appendOperator(node, ",");
      }
      appendOperator(parent, "#)");
      return;
    }
    if (exp instanceof SelectionTerm)
    {
      SelectionTerm object = (SelectionTerm)exp;
      int priority = object.getPriority();
      Expression base = object.getBase();
      Selector selector = object.getSelector();
      appendExp(parent, base, priority);
      appendSelector(parent, selector);
      return;
    }
    if (exp instanceof TupleTerm)
    {
      TupleTerm object = (TupleTerm)exp;
      appendOperator(parent, "(");
      Node node = parent; // appendNode(parent, "mrow");
      Expression[] exps = object.getBase();
      int n = exps.length;
      for (int i=0; i<n; i++)
      {
        appendExp(node, exps[i]);
        if (i+1 < n) appendOperator(node, ",");
      }
      appendOperator(parent, ")");
      return;
    }
    if (exp instanceof UnaryExpression)
    {
      UnaryExpression object = (UnaryExpression)exp;
      int priority = object.getPriority();
      Expression base = object.getBase();
      if (exp instanceof NegationTerm)
        appendPrefixFunction(parent, "-");
      else if (exp instanceof NotFormula)
        appendPrefixConnective(parent,"\u00AC"); //&not;
      else if (object instanceof BitNotTerm)
        appendInfixConnective(parent, "~"); 
      else
      {
        System.out.println("Unknown unary expression: " + exp.toString());
        System.exit(-1);
      }
      appendExp(parent, base, priority);
      return;
    }
    if (exp instanceof UpdateTerm)
    {
      UpdateTerm object = (UpdateTerm)exp;
      int priority = object.getPriority();
      Expression base = object.getBase();
      Expression update = object.getUpdate();
      Selector[] selectors = object.getSelectors();
      appendExp(parent, base, priority);
      appendInfixConnective(parent, "with");
      int n = selectors.length;
      for (int i=0; i<n; i++)
        appendSelector(parent, selectors[i]);
      appendInfixFunction(parent, ":=");
      appendExp(parent, update, priority);
      return;
    }
    System.out.println("Unknown expression: " + exp.toString());
    System.exit(-1);
  }
    
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendOperator(Node parent, String text)
  {
    Element node = appendNode(parent, "mo");
    appendTextNode(node, text);
    return node;
  }
 
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendPrefixFunction(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "0px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendInfixFunction(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendInfixPredicate(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "3px");
    node.setAttribute("rspace", "3px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendKeyword(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "0px");
    node.setAttribute("mathcolor", "blue");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendInfixConnective(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "4px");
    node.setAttribute("rspace", "4px");
    node.setAttribute("mathcolor", "blue");
    //node.setAttribute("mathvariant", "bold");
    //Element snode = appendNode(parent, "mspace");
    //snode.setAttribute("linebreak", "goodbreak");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendPrefixConnective(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "3px");
    node.setAttribute("mathcolor", "blue");
    //node.setAttribute("mathvariant", "bold");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendQuantifier(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("rspace", "0px");
    node.setAttribute("mathcolor", "red");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  public void appendPunctuation(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "2px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendConnector(Node parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
  }
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendNumber(Node parent, fmrisc.ProofNavigator.Syntax.Number object)
  {
    Element node = appendNode(parent, "mn");
    appendTextNode(node, object.getDigits());
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the identifier to be converted.
   ***************************************************************************/
  public void appendIdentifier(Node parent, Identifier object)
  {
    String name = object.toString();
    int p = name.indexOf('_');
    if (p == -1 || p == 0 || p == name.length()-1)
    {
      appendIdentifier(parent, name);
      return;
    }
    String base = name.substring(0, p);
    String index = name.substring(p+1);
    Node node = appendNode(parent, "msub");
    appendIdentifier(node, base);
    appendIdentifier(node, index);
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param name the name of the identifier to be converted.
   ***************************************************************************/
  private void appendIdentifier(Node parent, String name)
  {
    Element node = appendNode(parent, "mi");
    appendTextNode(node, name);
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendValuedIdentifier(Node parent, ValuedIdentifier object)
  {
    Identifier ident = object.getIdentifier();
    appendIdentifier(parent, ident);
    appendOperator(parent, "=");
    Expression value = object.getValue();
    appendExp(parent, value);
  }
  
  /***************************************************************************
   * Convert typed identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the typed identifier to be converted.
   ***************************************************************************/
  public void appendTypedIdentifier(Node parent, TypedIdentifier object)
  {
    Identifier ident = object.getIdentifier();
    appendIdentifier(parent, ident);
    appendConnector(parent, "\u2208"); // &isin;
    Type type = object.getType();
    appendType(parent, type);
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendSelector(Node parent, Selector object)
  {
    if (object instanceof SelectorIdentifier)
    {
      SelectorIdentifier sel = (SelectorIdentifier)object;
      Identifier ident = sel.getIdentifier();
      appendOperator(parent, ".");
      appendIdentifier(parent, ident.getName());
      return;
    }
    if (object instanceof SelectorIndex)
    {
      SelectorIndex sel = (SelectorIndex)object;
      Expression index = sel.getIndex();
      appendOperator(parent, "[");
      appendExp(parent, index);
      appendOperator(parent, "]");
      return;
    }
    if (object instanceof SelectorNumber)
    {
      SelectorNumber sel = (SelectorNumber)object;
      fmrisc.ProofNavigator.Syntax.Number number = sel.getNumber();
      appendOperator(parent, ".");
      appendNumber(parent, number);
      return;
    }
    System.out.println("Unknown selector: " + object.toString());
    System.exit(-1);
  }
  
  /***************************************************************************
   * Create new OMDoc node with denoted name and append it to parent node.
   * @param parent the parent of the new node.
   * @param name the name of the new node.
   * @return the new node.
   **************************************************************************/
  private Element appendNode(Node parent, String name)
  {
    Element node = document.createElement(name);
    parent.appendChild(node);
    return node;
  }
  
  /***************************************************************************
   * Create new DOM text node with given text and append it into parent node.
   * @param parent the parent of the new node.
   * @param text the text of the node.
   **************************************************************************/
  private void appendTextNode(Node parent, String text)
  {
    Node node = document.createTextNode(text);
    parent.appendChild(node);
  }
}
