// ----------------------------------------------------------------------------
// PrettyMathML.java
// Conversion of AST to HTML/MathML presentation markup (pretty-printed).
// $Id: PrettyMathML.java,v 1.19 2013/04/06 11:29:40 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Communication;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Syntax.*;

import java.util.*;

import org.w3c.dom.*;

/**
 * 
 * Conversion of AST to HTML/MathML presentation markup (pretty-printed).
 *
 **/
public final class PrettyMathML
{
  // see class PrettyPrinter for more information.
  
  // the default width for printing
  public static final int WIDTH = 85;
  
  private final int MAXLEVEL = 7; // maximum indentation level
  private final int MINFREE = 20; // no further breaks if this is reached
  private final int INDENT = 5;   // amount of indentation per level
  
  private int width;         // the suggested line width
  private Document document; // the document in which to create new nodes
  private Node root;         // the root to append math nodes to
  private Parent mroot;      // the current root math node
  
  private boolean lineMode = false;  // we are in line mode
  private int level = 0;             // the current indentation level
  private int pos = 0;               // the current column position
  
  // a reference to the current parent node
  private final static class Parent
  {
    Node node;
    Parent (Node node)
    {
      this.node = node;
    }
  }
  
  /****************************************************************************
   * Create converter that creates DOM nodes in the specified document.
   * @param width the maximum line width (suggestion, not a hard bound).
   * @param document the document in which to create the DOM nodes.
   ***************************************************************************/
  public PrettyMathML(Document document, int width)
  {
    this.document = document;
    this.width = width;
  }

  /****************************************************************************
   * Create converter that creates DOM nodes in the specified document.
   * @param document the document in which to create the DOM nodes.
   ***************************************************************************/
  public PrettyMathML(Document document)
  {
    this(document, WIDTH);
  }

  /****************************************************************************
   * Set pretty printer to initial state
   ***************************************************************************/
  private void reset()
  {
    lineMode = false;
    pos = 0;
  }
  
  /***************************************************************************
   * Convert declaration to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the declaration to be converted.
   ***************************************************************************/
  public void appendDecl(Node parent, Declaration object)
  {
    reset();
    root = parent;
    mroot = new Parent(MathML.createMathNode(document));
    parent.appendChild(mroot.node);
    appendDecl(mroot, BreakInfo.compute(object, true));
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to convert.
   **************************************************************************/
  public void appendExp(Node parent, Expression exp)
  {
    reset();
    root = parent;
    mroot = new Parent(MathML.createMathNode(document));
    parent.appendChild(mroot.node);
    appendExp(mroot, BreakInfo.compute(exp, true));
  }
  
  /***************************************************************************
   * Convert typed identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the typed identifier to be converted.
   ***************************************************************************/
  public void appendTypedIdentifier(Node parent, TypedIdentifier object)
  {
    reset();
    root = parent;
    mroot = new Parent(MathML.createMathNode(document));
    parent.appendChild(mroot.node);
    appendTypedIdentifier(mroot, BreakInfo.compute(object, true));
  }
  
  /***************************************************************************
   * 
   * Printing text.
   * 
   **************************************************************************/
  
  /***************************************************************************
   * Start new line and indent to the appropriate position.
   * @param overflow true iff line is an overflow line
   **************************************************************************/
  private void resetLine(Parent parent, boolean overflow)
  {
    root.appendChild(document.createElement("br"));
    int indent = level*INDENT;
    if (overflow) indent = indent+2*INDENT;
    StringBuffer buffer = new StringBuffer(indent);
    for (int i=0; i<indent; i++)
      buffer.append('\u00A0');
    root.appendChild(document.createTextNode(buffer.toString()));
    mroot.node = MathML.createMathNode(document);
    root.appendChild(mroot.node);
    pos = indent;
  }
  
  /***************************************************************************
   * Create new DOM text node with given text and append it into parent node.
   * @param parent the parent of the new node.
   * @param text the text of the node.
   * @param len the length which the text is assumed to consume
   *        (typically text.length() but may be larger to make up for
   *         surrounding space).
   **************************************************************************/
  private void appendTextNode(Parent parent, String text, int len)
  {
    if (width-pos < len) 
    {
      resetLine(parent, true);
    }
    Node node = document.createTextNode(text);
    parent.node.appendChild(node);
    pos = pos+len;
  }
  
  /***************************************************************************
   * If in block mode, indent to the apropriate position
   * by appending space to the parent node.
   **************************************************************************/
  private void indent(Parent parent)
  {
    int indent = level*INDENT;
    StringBuffer buffer = new StringBuffer(Math.max(0,indent-pos));
    for (int i=pos; i<indent; i++)
      buffer.append('\u00A0');
    String s = buffer.toString();
    appendTextNode(parent, s, s.length());
    pos = indent;
  }
  
  /****************************************************************************
   * Change indentation level.
   * @param increase true if level is increased (decreased otherwise)
   * @param parent the parent node
   ***************************************************************************/
  public void changeLevel(boolean increase, Parent parent)
  {
    if (lineMode) return;
    if (increase) 
    {
      level++;
      indent(parent);
    }
    else
      level--;
  }
  
  /***************************************************************************
   * Start new (non-overflow)line and indent it appropriately.
   * @param parent the parent node.
   ***************************************************************************/
  public void newLine(Parent parent)
  {
    if (lineMode) return;
    resetLine(parent, false);
  }
  
  /***************************************************************************
   * 
   * Auxiliary methods.
   * 
   **************************************************************************/
  
  /***************************************************************************
   * Convert declaration to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param info the declaration to be converted.
   ***************************************************************************/
  private void appendDecl(Parent parent, BreakInfo info)
  {
    Vector<Object> children = info.getChildren();
    Declaration object = (Declaration)info.getTree();
    
    int len = info.getLength();
    int free = width-pos;
    
    // enough space left, we can print the simple way
    if (len <= free)
    {
      new MathML(document).appendDecl(parent.node, object);
      pos = pos+len;
      return;
    }
    
    // we switch to line mode
    boolean oldMode = lineMode;
    if (level > MAXLEVEL || free < MINFREE)
      lineMode = true;
    
    if (object instanceof TypeDeclaration)
    {
      TypeDeclaration decl = (TypeDeclaration)object;
      appendIdentifier(parent, object.getIdentifier());
      appendWideConnector(parent, "\u2208"); // &isin;
      appendKeyword(parent, "type");
      if (decl.getType() == null) return;
      appendInfixPredicate(parent, "=");
      changeLevel(true, parent);
      newLine(parent);
      appendType(parent, (BreakInfo)children.elementAt(1));
      changeLevel(false, parent);
    }
    else if (object instanceof ValueDeclaration)
    {
      ValueDeclaration decl = (ValueDeclaration)object;
      appendIdentifier(parent, decl.getIdentifier());
      int pos = 1;
      if (decl.getType() != null)
      {
        appendWideConnector(parent, "\u2208"); // &isin;
        appendType(parent, (BreakInfo)children.elementAt(pos));
        pos = pos+1;
      }
      if (decl.getValue() != null)
      {
        appendInfixPredicate(parent, "=");
        changeLevel(true, parent);
        newLine(parent);
        appendExp(parent, (BreakInfo)children.elementAt(pos));        
        changeLevel(false, parent);
      }
    } 
    else if (object instanceof FormulaDeclaration)
    {
      if (object instanceof AxiomDeclaration)
        appendPrefixConnective(parent, "axiom");
      appendIdentifier(parent, object.getIdentifier());
      appendInfixConnective(parent, "\u2261"); // &equiv;
      changeLevel(true, parent);
      newLine(parent);
      appendExp(parent, (BreakInfo)children.elementAt(1));
      changeLevel(false, parent);
      return;
    }
    else
    {
      System.out.println("Unknown declaration: " + object.toString());
      System.exit(-1);
    }
    
    // restore mode
    lineMode = oldMode;
  }
  
  /***************************************************************************
   * Convert type to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param info the type to be converted.
   ***************************************************************************/
  private void appendType(Parent parent, BreakInfo info)
  {
    Vector<Object> children = info.getChildren();
    Type object = (Type)info.getTree();
    boolean mode = lineMode;
    lineMode = true;
    if (object instanceof Identifier)
    {
      appendIdentifier(parent, (Identifier)object);
    }
    else if (object instanceof ArrayType)
    {
      appendPrefixConnective(parent, "array");
      appendType(parent, (BreakInfo)children.elementAt(0));
      appendInfixConnective(parent, "of");
      appendType(parent, (BreakInfo)children.elementAt(1));
    }
    else if (object instanceof AtomicType)
    {
      AtomicType type = (AtomicType) object;
      String name = type.getName();
      if (name.equals("BOOLEAN"))
      {
        Element node = document.createElement("mo");
        node.setAttribute("mathvariant", "double-struck");
        parent.node.appendChild(node);
        if (State.booleanAsLogical())
          appendTextNode(new Parent(node), "LOGICAL", 1);
        else
          appendTextNode(new Parent(node), "B", 1);
      }
      else if (name.equals("BIT"))
      {
        Element node = document.createElement("mo");
        node.setAttribute("mathvariant", "double-struck");
        parent.node.appendChild(node);
        if (State.booleanAsLogical())
          appendTextNode(new Parent(node), "B", 1);
        else
          appendTextNode(new Parent(node), "BIT", 1);
      }
      else if (name.equals("NAT"))
        appendOperator(parent, "\u2115"); // &naturals;
      else if (name.equals("INT"))
        appendOperator(parent, "\u2124"); // &integers;
      else if (name.equals("REAL"))
        appendOperator(parent, "\u211D"); // &reals;
      else
        appendIdentifier(parent, name);
    } 
    else if (object instanceof FunctionType)
    {
      FunctionType type = (FunctionType) object;
      Type[] domain = type.getDomain();
      int n = domain.length;
      BreakInfo[] dinfo = (BreakInfo[])children.elementAt(0);
      if (n == 1)
      {
        if (domain[0] instanceof FunctionType ||
            domain[0] instanceof ArrayType)
        {
          appendOperator(parent, "(");
          appendType(parent, dinfo[0]);
          appendOperator(parent, ")");
        }
        else
          appendType(parent, dinfo[0]);
      }
      else
      {
        appendOperator(parent, "(");
        for (int i=0; i<n; i++)
        {
          appendType(parent, dinfo[i]);
          if (i<n-1) appendWideOperator(parent, ",");
        }
        appendOperator(parent, ")");
      }
      appendWideInfixFunction(parent, "\u2192"); // &rightarrow;
      appendType(parent, (BreakInfo)children.elementAt(1));
    }
    else if (object instanceof RecordType)
    {
      BreakInfo[] base = (BreakInfo[])children.elementAt(0);
      appendOperator(parent, "[#");
      int n = base.length;
      for (int i=0; i<n; i++)
      {
        appendTypedIdentifier(parent, base[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      appendOperator(parent, "#]");
    }
    else if (object instanceof SubrangeType)
    {
      appendOperator(parent, "[");
      appendExp(parent, (BreakInfo)children.elementAt(0));
      appendOperator(parent, ",");
      appendExp(parent, (BreakInfo)children.elementAt(1));
      appendOperator(parent, "]");
    }
    else if (object instanceof SubType)
    {
      appendOperator(parent, "subtype");
      appendOperator(parent, "(");
      appendExp(parent,(BreakInfo)children.elementAt(0));
      appendOperator(parent, ")");
    }
    else if (object instanceof TupleType)
    {
      BreakInfo[] base = (BreakInfo[])children.elementAt(0);
      appendOperator(parent, "[");
      int n = base.length;
      for (int i=0; i<n; i++)
      {
        appendType(parent, base[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      appendOperator(parent, "]"); 
    }
    else if (object instanceof LetType)
    {
      BreakInfo[] decls = (BreakInfo[])children.elementAt(0);
      appendPrefixConnective(parent, "let");
      changeLevel(true, parent);
      int n = decls.length;
      for (int i=0; i<n; i++)
      {
        newLine(parent);
        appendDecl(parent, decls[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      changeLevel(false, parent);
      appendInfixConnective(parent, "in");
      changeLevel(true, parent);
      newLine(parent);
      appendType(parent, (BreakInfo)children.elementAt(1));
      changeLevel(false, parent);
    }
    else
    {
      System.out.println("Unknown type: " + object.toString());
      System.exit(-1);
    }
    lineMode = mode;
  }

  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup
   * surrounded by parentheses.
   * @param parent the node to which the representation is appended.
   * @param info the expression to convert.
   **************************************************************************/
  private void appendParens(Parent parent, BreakInfo info)
  {
    appendOperator(parent, "(");
    appendExp(parent, info);
    appendOperator(parent, ")");
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup
   * surrounded by parentheses, if necessary.
   * @param parent the node to which the representation is appended.
   * @param info the expression to convert.
   * @param the priority level of the parent expression. If exp has a higher
   *        priority, then parentheses are added to surround the markup.
   **************************************************************************/
  private void appendExp(Parent parent, BreakInfo info, int priority)
  {
    Expression exp = (Expression)info.getTree();
    int p = exp.getPriority();
    if (p <= priority)
      appendExp(parent, info);
    else
      appendParens(parent, info);
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendExp(Parent parent, BreakInfo info)
  {
    Vector<Object> children = info.getChildren();
    Expression exp = (Expression)info.getTree();
    
    int len = info.getLength();
    int free = width-pos;
    
    // enough space left, we can print the simple way
    if (len <= free)
    {
      new MathML(document).appendExp(parent.node, exp);
      pos = pos+len;
      return;
    }
    
    // we switch to line mode
    boolean oldMode = lineMode;
    if (level > MAXLEVEL || free < MINFREE)
      lineMode = true;
    
    if (exp instanceof fmrisc.ProofNavigator.Syntax.Number)
    {
      fmrisc.ProofNavigator.Syntax.Number object = (fmrisc.ProofNavigator.Syntax.Number)exp;
      appendNumber(parent, object);
    }
    else if (exp instanceof Identifier)
    {
      Identifier object = (Identifier)exp;
      appendIdentifier(parent, object);
    }
    else if (exp instanceof ApplicationExpression)
    {
      lineMode = true;
      int priority = exp.getPriority();
      appendExp(parent, (BreakInfo)children.elementAt(0), priority);
      // appendOperator(parent, "\u2061"); // &ApplyFunction;
      appendOperator(parent, "(");
      BreakInfo[] args = (BreakInfo[])children.elementAt(1);
      int n = args.length;
      for (int i=0; i<n; i++)
      {
        appendExp(parent, args[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      appendOperator(parent, ")");
    }
    else if (exp instanceof BinaryExpression)
    {
      BinaryExpression object = (BinaryExpression)exp;
      int priority = object.getPriority();
      if (object instanceof DividesTerm)
      {
        lineMode = true;
        Node node = appendNode(parent, "mfrac");
        Node num = appendNode(new Parent(node), "mrow");
        appendExp(new Parent(num), (BreakInfo)children.elementAt(0));
        Node denom = appendNode(new Parent(node), "mrow");
        appendExp(new Parent(denom), (BreakInfo)children.elementAt(1));
      }
      else if (object instanceof PowerTerm)
      {
        lineMode = true;
        Node node = appendNode(parent, "msup");
        Node base = appendNode(new Parent(node), "mrow");
        appendExp(new Parent(base), (BreakInfo)children.elementAt(0), priority);
        Node power = appendNode(new Parent(node), "mrow");
        appendExp(new Parent(power), (BreakInfo)children.elementAt(1), priority);
      }
      else
      {
        Expression first = object.getFirst();
        Expression second = object.getSecond();
        BreakInfo firstInfo = (BreakInfo)children.elementAt(0);
        BreakInfo secondInfo = (BreakInfo)children.elementAt(1);
        if (object instanceof MinusTerm &&
              (second instanceof PlusTerm || second instanceof MinusTerm))
        {
          appendExp(parent, firstInfo, priority);
          appendInfixFunction(parent, "-");
          appendParens(parent, secondInfo);
        }
        else if (object instanceof EquivalentFormula ||
            object instanceof NotEquivalentFormula)
        {
          changeLevel(true, parent);
          if (first instanceof EquivalentFormula ||
              first instanceof NotEquivalentFormula ||
              first instanceof ImpliesFormula)
            appendParens(parent, firstInfo);
          else
            appendExp(parent, firstInfo, priority);
          changeLevel(false, parent);
          newLine(parent);
          if (object instanceof EquivalentFormula)
            appendInfixConnective(parent, "\u21D4"); // &iff;
          else
            appendInfixConnective(parent, "xor");
          changeLevel(true, parent);
          newLine(parent);
          if (second instanceof EquivalentFormula ||
              second instanceof NotEquivalentFormula ||
              second instanceof ImpliesFormula)
            appendParens(parent, secondInfo);
          else
            appendExp(parent, secondInfo, priority);
          changeLevel(false, parent);
        }
        else if (object instanceof ImpliesFormula)
        {
          changeLevel(true, parent);
          while (true)
          {
            if (first instanceof EquivalentFormula ||
                first instanceof NotEquivalentFormula ||
                first instanceof ImpliesFormula)
              appendParens(parent, firstInfo);
            else
              appendExp(parent, firstInfo, priority);
            changeLevel(false, parent);
            newLine(parent); 
            appendInfixConnective(parent, "\u21D2"); // &Implies;
            if (!(object.getSecond() instanceof ImpliesFormula)) break;
            if (secondInfo.getLength()+INDENT <= width-pos) break;
            changeLevel(true, parent);
            newLine(parent);
            info = secondInfo;
            children = info.getChildren();
            object = (BinaryExpression)info.getTree();
            priority = object.getPriority();
            first = object.getFirst();
            second = object.getSecond();
            firstInfo = (BreakInfo)children.elementAt(0);
            secondInfo = (BreakInfo)children.elementAt(1);
          }
          changeLevel(true, parent);
          newLine(parent);
          if (second instanceof EquivalentFormula ||
              second instanceof NotEquivalentFormula)
            appendParens(parent, secondInfo);
          else
            appendExp(parent, secondInfo, priority);
          changeLevel(false, parent);
        }
        else if (object instanceof AndFormula)
        {
          changeLevel(true, parent);
          while (true)
          {
            while (true)
            {
              appendExp(parent, firstInfo, priority);
              if (!(object.getSecond() instanceof AndFormula)) break;
              Vector<Object> children0 = secondInfo.getChildren();
              BreakInfo thirdInfo = (BreakInfo)children0.elementAt(0);
              if (3+thirdInfo.getLength()+INDENT > width-pos) break;
              appendInfixConnective(parent, "\u2227"); // &and;
              object = (BinaryExpression)object.getSecond();
              firstInfo = thirdInfo;
              secondInfo = (BreakInfo)children0.elementAt(1);
            }
            changeLevel(false, parent);
            newLine(parent);
            appendInfixConnective(parent, "\u2227"); // &and;
            if (!(object.getSecond() instanceof AndFormula)) break;
            if (secondInfo.getLength()+INDENT <= width-pos) break;
            changeLevel(true, parent);
            newLine(parent);
            info = secondInfo;
            children = info.getChildren();
            object = (BinaryExpression)info.getTree();
            priority = exp.getPriority();
            first = object.getFirst();
            second = object.getSecond();
            firstInfo = (BreakInfo)children.elementAt(0);
            secondInfo = (BreakInfo)children.elementAt(1);
          }
          changeLevel(true, parent);
          newLine(parent);
          appendExp(parent, secondInfo, priority);
          changeLevel(false, parent);
        }
        else if (object instanceof OrFormula)
        {
          changeLevel(true, parent);
          while (true)
          {
            while (true)
            {
              appendExp(parent, firstInfo, priority);
              if (!(object.getSecond() instanceof OrFormula)) break;
              Vector<Object> children0 = secondInfo.getChildren();
              BreakInfo thirdInfo = (BreakInfo)children0.elementAt(0);
              if (3+thirdInfo.getLength()+INDENT > width-pos) break;
              appendInfixConnective(parent, "\u2228"); // &or;
              object = (BinaryExpression)object.getSecond();
              firstInfo = thirdInfo;
              secondInfo = (BreakInfo)children0.elementAt(1);
            }
            changeLevel(false, parent);
            newLine(parent);
            appendInfixConnective(parent, "\u2228"); // &or;
            if (!(object.getSecond() instanceof OrFormula)) break;
            if (secondInfo.getLength()+INDENT <= width-pos) break;
            changeLevel(true, parent);
            newLine(parent);
            info = secondInfo;
            children = info.getChildren();
            object = (BinaryExpression)info.getTree();
            priority = exp.getPriority();
            first = object.getFirst();
            second = object.getSecond();
            firstInfo = (BreakInfo)children.elementAt(0);
            secondInfo = (BreakInfo)children.elementAt(1);
          }
          changeLevel(true, parent);
          newLine(parent);
          appendExp(parent, secondInfo, priority);
          changeLevel(false, parent);
        }
        else 
        {
          // expressions printed in lineMode
          lineMode = true;
          appendExp(parent, (BreakInfo)children.elementAt(0), priority);
          if (object instanceof EqualsFormula)
            appendInfixPredicate(parent, "=");
          else if (object instanceof GreaterEqualFormula)
            appendInfixPredicate(parent, "\u2265"); // &ge;
          else if (object instanceof GreaterFormula)
            appendInfixPredicate(parent, ">");
          else if (object instanceof LessFormula)
            appendInfixPredicate(parent, "<");
          else if (object instanceof MinusTerm)
            appendInfixFunction(parent, "-");
          else if (object instanceof LessEqualFormula)
            appendInfixPredicate(parent, "\u2264"); // &le;
          else if (object instanceof NotEqualsFormula)
            appendInfixPredicate(parent, "\u2260"); // &ne;
          else if (object instanceof PlusTerm)
            appendInfixFunction(parent, "+");
          else if (object instanceof BitOrTerm)
            appendInfixFunction(parent, "|");
          else if (object instanceof TimesTerm)
            // appendInfixFunction(parent, "\u00D7"); // &times;
            appendInfixFunction(parent, "\u00B7"); // &centerdot;
          else if (object instanceof BitAndTerm)
            appendInfixFunction(parent, "&");
          else
          {
            System.out.println("Unknown binary expression: " + exp.toString());
            System.exit(-1);
          }
          appendExp(parent, (BreakInfo)children.elementAt(1), priority);
        }
      }
    }
    else if (exp instanceof IfThenElseExpression)
    {
      IfThenElseExpression cexp = (IfThenElseExpression)exp;
      appendInfixConnective(parent, "if");
      boolean mode = lineMode;
      lineMode = true;
      appendExp(parent, (BreakInfo)children.elementAt(0));
      lineMode = mode;
      appendInfixConnective(parent, "then");
      changeLevel(true, parent);
      newLine(parent); 
      appendExp(parent, (BreakInfo)children.elementAt(1));
      changeLevel(false, parent);
      Expression elseBranch = cexp.getElseBranch();
      while (elseBranch instanceof IfThenElseExpression)
      {
        info = (BreakInfo)children.elementAt(2);
        children = info.getChildren();
        cexp = (IfThenElseExpression)elseBranch;
        newLine(parent);
        appendInfixConnective(parent, "elsif");
        lineMode = true;
        appendExp(parent, (BreakInfo)children.elementAt(0));
        lineMode = mode;
        appendInfixConnective(parent, "then");
        changeLevel(true, parent);
        newLine(parent); 
        appendExp(parent, (BreakInfo)children.elementAt(1));
        changeLevel(false, parent);
        elseBranch = cexp.getElseBranch();
      }
      newLine(parent);
      appendInfixConnective(parent, "else");
      changeLevel(true, parent);
      newLine(parent);
      appendExp(parent, (BreakInfo)children.elementAt(2));
      changeLevel(false, parent);
      newLine(parent);
      appendInfixConnective(parent, "endif");
    }
    else if (exp instanceof LetExpression)
    {
      BreakInfo[] decls = (BreakInfo[])children.elementAt(0);
      appendPrefixConnective(parent, "let");
      changeLevel(true, parent);
      int n = decls.length;
      // check whether sequence fits in single line
      int l = BreakInfo.length(decls);
      if (n>1) l = l+(n-1)*2;
      boolean single = (pos+l <= width);
      boolean mode = lineMode;
      if (single) 
      {
        newLine(parent);
        lineMode = true;
      }
      for (int i=0; i<n; i++)
      {
        if (!single) newLine(parent);
        appendDecl(parent, decls[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      lineMode = mode;
      changeLevel(false, parent);
      newLine(parent);
      appendInfixConnective(parent, "in");
      changeLevel(true, parent);
      newLine(parent);
      appendExp(parent, (BreakInfo)children.elementAt(1));
      changeLevel(false, parent);
    }
    else if (exp instanceof Logical)
    {
      Logical object = (Logical)exp;
      if (State.booleanAsLogical())
      {
        if (object.getValue())
          appendKeyword(parent, "logicaltrue");
        else
          appendKeyword(parent, "logicalfalse");
      }
      else
      {
        if (object.getValue())
          appendKeyword(parent, "true");
        else
          appendKeyword(parent, "false");
      }
    }
    else if (exp instanceof BitLogical)
    {
      BitLogical object = (BitLogical)exp;
      if (State.booleanAsLogical())
      {
        if (object.getValue())
          appendKeyword(parent, "true");
        else
          appendKeyword(parent, "false");
      }
      else
      {
        if (object.getValue())
          appendKeyword(parent, "bittrue");
        else
          appendKeyword(parent, "bitfalse");
      }
    }
    else if (exp instanceof QuantifiedExpression)
    {
      QuantifiedExpression object = (QuantifiedExpression)exp;
      int priority = object.getPriority();
      String quantifier = null;
      if (exp instanceof ForallFormula)
        quantifier = "\u2200"; // &forall;
      else if (exp instanceof ExistsFormula)
        quantifier = "\u2203"; // &exists;
      else if (exp instanceof LambdaTerm)
        quantifier = "\u03BB"; // &lambda;
      else if (exp instanceof ArrayTerm)
        quantifier = "array";
      else
      {
        System.out.println("Unknown quantifier: " + exp.toString());
        System.exit(-1);
      }
      appendQuantifier(parent, quantifier);
      boolean mode = lineMode;
      lineMode = true;
      BreakInfo[] vars = (BreakInfo[])children.elementAt(0);
      int n = vars.length;
      for (int i=0; i<n; i++)
      {
        appendTypedIdentifier(parent, vars[i]);
        if (i+1<n) appendWideOperator(parent, ",");
      }
      lineMode = mode;
      appendPunctuation(parent, ":");
      changeLevel(true, parent);
      newLine(parent);
      appendExp(parent, (BreakInfo)children.elementAt(1), priority);
      changeLevel(false, parent);
    }
    else if (exp instanceof RecordTerm)
    {
      lineMode = true;
      appendOperator(parent, "(#");
      BreakInfo[] idents = (BreakInfo[])children.elementAt(0);
      int n = idents.length;
      for (int i=0; i<n; i++)
      {
        appendValuedIdentifier(parent, idents[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      appendOperator(parent, "#)");
    }
    else if (exp instanceof SelectionTerm)
    {
      lineMode = true;
      SelectionTerm object = (SelectionTerm)exp;
      int priority = object.getPriority();
      appendExp(parent, (BreakInfo)children.elementAt(0), priority);
      appendSelector(parent, (BreakInfo)children.elementAt(1));
    }
    else if (exp instanceof TupleTerm)
    {
      lineMode = true;
      appendOperator(parent, "(");
      BreakInfo[] exps = (BreakInfo[])children.elementAt(0);
      int n = exps.length;
      for (int i=0; i<n; i++)
      {
        appendExp(parent, exps[i]);
        if (i+1 < n) appendWideOperator(parent, ",");
      }
      appendOperator(parent, ")");
    }
    else if (exp instanceof UnaryExpression)
    {
      int priority = exp.getPriority();
      if (exp instanceof NegationTerm)
      {
        lineMode = true;
        appendPrefixFunction(parent, "-");
      }
      else if (exp instanceof NotFormula)
        appendPrefixConnective(parent,"\u00AC"); //&not;
      else if (exp instanceof BitNotTerm)
        appendPrefixConnective(parent,"~");
      else
      {
        System.out.println("Unknown unary expression: " + exp.toString());
        System.exit(-1);
      }
      appendExp(parent, (BreakInfo)children.elementAt(0), priority);
    }
    else if (exp instanceof UpdateTerm)
    {
      lineMode = true;
      int priority = exp.getPriority();
      appendExp(parent, (BreakInfo)children.elementAt(0), priority);
      appendInfixConnective(parent, "with");
      BreakInfo[] selectors = (BreakInfo[])children.elementAt(1);
      int n = selectors.length;
      for (int i=0; i<n; i++)
        appendSelector(parent, selectors[i]);
      appendInfixFunction(parent, ":=");
      appendExp(parent, (BreakInfo)children.elementAt(2), priority);
    }
    else
    {
      System.out.println("Unknown expression: " + exp.toString());
      System.exit(-1);
    }
    
    // restore mode
    lineMode = oldMode;
  }
  
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param info the expression to be converted.
   ***************************************************************************/
  private void appendValuedIdentifier(Parent parent, BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object> children = info.getChildren();
    ValuedIdentifier object = (ValuedIdentifier)info.getTree();
    Identifier ident = object.getIdentifier();
    appendIdentifier(parent, ident);
    appendOperator(parent, "=");
    appendExp(parent, (BreakInfo)children.elementAt(1));
    lineMode = mode;
  }
  
  /***************************************************************************
   * Convert typed identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param info the typed identifier to be converted.
   ***************************************************************************/
  private void appendTypedIdentifier(Parent parent, BreakInfo info)
  {
    boolean mode = lineMode;
    lineMode = true;
    Vector<Object> children = info.getChildren();
    TypedIdentifier object = (TypedIdentifier)info.getTree();
    Identifier ident = object.getIdentifier();
    appendIdentifier(parent, ident);
    appendWideConnector(parent, "\u2208"); // &isin;
    appendType(parent, (BreakInfo)children.elementAt(1));
    lineMode = mode;
  }
  
  /***************************************************************************
   * Convert selector to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param info the selctor to be converted.
   ***************************************************************************/
  private void appendSelector(Parent parent, BreakInfo info)
  {
    Vector<Object> children = info.getChildren();
    Selector object = (Selector)info.getTree();
    if (object instanceof SelectorIdentifier)
    {
      SelectorIdentifier sel = (SelectorIdentifier)object;
      Identifier ident = sel.getIdentifier();
      appendOperator(parent, ".");
      appendIdentifier(parent, ident.getName());
      return;
    }
    if (object instanceof SelectorIndex)
    {
      appendOperator(parent, "[");
      appendExp(parent, (BreakInfo)children.elementAt(0));
      appendOperator(parent, "]");
      return;
    }
    if (object instanceof SelectorNumber)
    {
      SelectorNumber sel = (SelectorNumber)object;
      fmrisc.ProofNavigator.Syntax.Number number = sel.getNumber();
      appendOperator(parent, ".");
      appendNumber(parent, number);
      return;
    }
    System.out.println("Unknown selector: " + object.toString());
    System.exit(-1);
  }
  
  /***************************************************************************
   * 
   * Append individual nodes.
   * 
   **************************************************************************/
  
  /***************************************************************************
   * Create new OMDoc node with denoted name and append it to parent node.
   * @param parent the parent of the new node.
   * @param name the name of the new node.
   * @return the new node.
   **************************************************************************/
  private Element appendNode(Parent parent, String name)
  {
    Element node = document.createElement(name);
    parent.node.appendChild(node);
    return node;
  }
   
  /***************************************************************************
   * Convert expression to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param exp the expression to be converted.
   ***************************************************************************/
  private void appendNumber(Parent parent, fmrisc.ProofNavigator.Syntax.Number object)
  {
    Node node = appendNode(parent, "mn");
    String s = object.getDigits();
    appendTextNode(new Parent(node), s, s.length());
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param name the name of the identifier to be converted.
   ***************************************************************************/
  private void appendIdentifier(Parent parent, String name)
  {
    Element node = appendNode(parent, "mi");
    appendTextNode(new Parent(node), name, name.length());
  }
  
  /***************************************************************************
   * Convert identifier to DOM representation of HTML/MathML markup.
   * @param parent the node to which the representation is appended.
   * @param object the identifier to be converted.
   ***************************************************************************/
  private void appendIdentifier(Parent parent, Identifier object)
  {
    String name = object.toString();
    int p = name.indexOf('_');
    if (p == -1 || p == 0 || p == name.length()-1)
    {
      appendIdentifier(parent, name);
      return;
    }
    String base = name.substring(0, p);
    String index = name.substring(p+1);
    Element node = appendNode(parent, "msub");
    appendIdentifier(new Parent(node), base);
    appendIdentifier(new Parent(node), index);
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendOperator(Parent parent, String text)
  {
    Element node = appendNode(parent, "mo");
    appendTextNode(new Parent(node), text, text.length());
    return node;
  }
 
  /***************************************************************************
   * Append to parent operator node with denoted text assuming that it
   * takes some more space than denoted by the text itself (2 chars extra).
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendWideOperator(Parent parent, String text)
  {
    Element node = appendNode(parent, "mo");
    appendTextNode(new Parent(node), text, text.length()+1);
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendPrefixFunction(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "0px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendInfixFunction(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text, wide version.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   * @return the created node.
   **************************************************************************/
  private Element appendWideInfixFunction(Parent parent, String text)
  {
    Element node = appendWideOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
    return node;
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendInfixPredicate(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "3px");
    node.setAttribute("rspace", "3px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendKeyword(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "0px");
    node.setAttribute("mathcolor", "blue");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendInfixConnective(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "4px");
    node.setAttribute("rspace", "4px");
    node.setAttribute("mathcolor", "blue");
    //node.setAttribute("mathvariant", "bold");
    //Element snode = appendNode(parent, "mspace");
    //snode.setAttribute("linebreak", "goodbreak");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendPrefixConnective(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "3px");
    node.setAttribute("mathcolor", "blue");
    //node.setAttribute("mathvariant", "bold");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendQuantifier(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("rspace", "0px");
    node.setAttribute("mathcolor", "red");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text, wide version.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendWideConnector(Parent parent, String text)
  {
    Element node = appendWideOperator(parent, text);
    node.setAttribute("lspace", "1px");
    node.setAttribute("rspace", "1px");
  }
  
  /***************************************************************************
   * Append to parent operator node with denoted text.
   * @param parent the node to which the operator node is appended.
   * @param text the text of the operator node.
   **************************************************************************/
  private void appendPunctuation(Parent parent, String text)
  {
    Element node = appendOperator(parent, text);
    node.setAttribute("lspace", "0px");
    node.setAttribute("rspace", "2px");
  }
}
