// ----------------------------------------------------------------------------
// Rewriting.java
// A visitor for rewriting expressions to a more human-friendly form.
// $Id: Rewriting.java,v 1.26 2015/09/17 09:31:03 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Proving;

import java.util.Vector;
import fmrisc.ProofNavigator.Syntax.Number;
import fmrisc.ProofNavigator.Syntax.*;
import fmrisc.ProofNavigator.Semantics.*;

/**
 * 
 * A visitor for rewriting expressions to a more human-friendly form.
 *
 */
public final class Rewriting extends ASTVisitorBase
{
  /***************************************************************************
   * Rewrite expression to more human-friendly form.
   * @param exp an expression.
   * @return its rewritten form.
   *************************************************************************/
  public static Expression rewrite(Expression exp)
  {
    return (Expression)exp.accept(new Rewriting());
  }
  
  /***************************************************************************
   * 
   * Visitor functions.
   *
   **************************************************************************/
  
  public boolean isCloned()
  {
    return true;
  }
  
  /***************************************************************************
   * Visit not formula, apply rule "toNotEqualsFormula" and
   * draw negation to innermost symbols.
   **************************************************************************/
  protected AST visit(NotFormula exp) 
  { 
    Expression base0 = (Expression)exp.getBase().accept(this);
    Expression result = toNotEqualsFormula(ProofUtil.negate(base0));
    result = toIfThenElseFormula(result);
    // further rewriting may be possible
    if (!(result instanceof NotFormula))
      result = (Expression)result.accept(this);
    return result;
  }
 
  /***************************************************************************
   * Visit and formula, apply rule "rearrangeAndFormula".
   **************************************************************************/
  protected AST visit(AndFormula exp) 
  { 
    Expression first = (Expression)exp.getFirst().accept(this);
    Expression second = (Expression)exp.getSecond().accept(this);
    return rearrangeAndFormula(Construct.andFormula(first, second));
  }
  
  /***************************************************************************
   * Visit or formula, apply rules "rearrangeOrFormula" and "toImpliesFormula".
   **************************************************************************/
  protected AST visit(OrFormula exp) 
  { 
    // apply rule before drawing negations inside
    Expression exp0 = toImpliesFormula(Construct.orFormula(
      exp.getFirst(), exp.getSecond()));
    if (!(exp0 instanceof BinaryExpression)) return exp0;
    BinaryExpression exp1 = (BinaryExpression)exp0;
    Expression first0 = (Expression)exp1.getFirst().accept(this);
    Expression second0 = (Expression)exp1.getSecond().accept(this);
    Expression result = exp1.construct(first0, second0);
    if (result instanceof OrFormula) 
      result = rearrangeOrFormula((OrFormula)result);
    return result;
  }
  
  /***************************************************************************
   * Visit or formula, apply rule "toImpliesFormula".
   **************************************************************************/
  protected AST visit(ImpliesFormula exp) 
  { 
    // apply rule before drawing negations inside
    Expression exp0 = toImpliesFormula(Construct.impliesFormula(
      exp.getFirst(), exp.getSecond()));
    if (!(exp0 instanceof BinaryExpression)) return exp0;
    BinaryExpression exp1 = (BinaryExpression)exp0;
    Expression first0 = (Expression)exp1.getFirst().accept(this);
    Expression second0 = (Expression)exp1.getSecond().accept(this);
    Expression result = exp1.construct(first0, second0);
    if (result instanceof OrFormula) 
      result = rearrangeOrFormula((OrFormula)result);
    return result;
  }
  
  /***************************************************************************
   * Visit application expression, apply rules "toUntupledApplication" and
   * "processLambdaApplication".
   **************************************************************************/
  protected AST visit(ApplicationExpression exp) 
  { 
    Expression fun = exp.getFunction();
    Expression fun0 = (Expression)fun.accept(this);
    Expression[] args = exp.getArguments();
    int n = args.length;
    Expression[] args0 = new Expression[n];
    for (int i=0; i<n; i++)
      args0[i] = (Expression)args[i].accept(this);
    return processLambdaApplication(
      Construct.applicationExpression(fun0, args0));
  }
  
  /***************************************************************************
   * Visit let expression, apply rule "processLetExpression".
   **************************************************************************/
  protected AST visit(LetExpression exp) 
  { 
    ValueDeclaration[] decls = exp.getDeclarations();
    int n = decls.length;
    ValueDeclaration[] decls0 = new ValueDeclaration[n];
    for (int i=0; i<n; i++)
    {
      decls0[i] = (ValueDeclaration)decls[i].accept(this);
      decls[i].getValueDeclIdentifier().setSubstitution(
        decls0[i].getValueDeclIdentifier());
    }
    Expression base0 = (Expression)exp.getBase().accept(this);
    for (int i=0; i<n; i++)
      decls[i].getValueDeclIdentifier().resetSubstitution();
    base0 = processLetExpression(Construct.letExpression(decls0, base0));
    // further rewriting may be possible
    return base0.accept(this);
  }
  
  /****************************************************************************
   * Visit plus term, apply rule "toSum".
   ***************************************************************************/
  protected AST visit(PlusTerm exp)
  {
    Expression first0 = (Expression)exp.getFirst().accept(this);
    Expression second0 = (Expression)exp.getSecond().accept(this);
    return toSum(Construct.plusTerm(first0, second0));
  }
  
  /****************************************************************************
   * Visit products, quotients and powers, apply rule "toFraction".
   ***************************************************************************/
  protected AST visit(TimesTerm exp)
  {
    return visitFraction(exp);
  }
  protected AST visit(DividesTerm exp)
  {
    return visitFraction(exp);
  }
  protected AST visit(PowerTerm exp)
  {
    return visitFraction(exp);
  }
  private AST visitFraction(BinaryExpression exp)
  {
    Expression first0 = (Expression)exp.getFirst().accept(this);
    Expression second0 = (Expression)exp.getSecond().accept(this);
    return toFraction(exp.construct(first0, second0));
  }
  
  /****************************************************************************
   * Visit some binary expressions, apply rule "toEquation".
   ***************************************************************************/
  protected AST visit(EqualsFormula exp)
  {
    return visitEquation(exp);
  }
  protected AST visit(NotEqualsFormula exp)
  {
    return visitEquation(exp);
  }
  protected AST visit(LessFormula exp)
  {
    return visitEquation(exp);
  }
  protected AST visit(LessEqualFormula exp)
  {
    return visitEquation(exp);
  }
  protected AST visit(GreaterFormula exp)
  {
    return visitEquation(exp);
  }
  protected AST visit(GreaterEqualFormula exp)
  {
    return visitEquation(exp);
  }
  private AST visitEquation(BinaryExpression exp)
  {
    Expression first0 = (Expression)exp.getFirst().accept(this);
    Expression second0 = (Expression)exp.getSecond().accept(this);
    return toEquation(exp.construct(first0, second0));
  }
  
  /***************************************************************************
   * Visit selection term, apply rule "processArraySelection".
   **************************************************************************/
  protected AST visit(SelectionTerm exp) 
  { 
    Expression base = (Expression)exp.getBase().accept(this);
    Selector sel = (Selector)exp.getSelector().accept(this);
    return processArraySelection(Construct.selectionTerm(base, sel));
  }
  
  /***************************************************************************
   * 
   * Auxiliary functions
   *
   **************************************************************************/
  
  /***************************************************************************
   * If exp = NOT a=b, rewrite exp to a /= b.
   * @param exp an expression
   * @return transformed exp (exp itself, if rule is not applicable).
   **************************************************************************/
  private static Expression toNotEqualsFormula(Expression exp)
  {
    if (!(exp instanceof NotFormula)) return exp;
    NotFormula exp0 = (NotFormula)exp;
    Expression base = exp0.getBase();
    if (!(base instanceof EqualsFormula)) return exp;
    EqualsFormula base0 = (EqualsFormula)base;
    return Construct.notEqualsFormula(base0.getFirst(), base0.getSecond());
  }
  
  /***************************************************************************
   * If exp = NOT IF a THEN b ELSE c ENDIF, 
   * rewrite exp to IF a THEN NOT b ELSE NOT c ENDIF.
   * @param exp an expression
   * @return transformed exp (exp itself, if rule is not applicable).
   **************************************************************************/
  private static Expression toIfThenElseFormula(Expression exp)
  {
    if (!(exp instanceof NotFormula)) return exp;
    NotFormula exp0 = (NotFormula)exp;
    Expression base = exp0.getBase();
    if (!(base instanceof IfThenElseExpression)) return exp;
    IfThenElseExpression base0 = (IfThenElseExpression)base;
    return Construct.ifThenElseExpression(base0.getCondition(), 
        Construct.notFormula(base0.getThenBranch()),
        Construct.notFormula(base0.getElseBranch()));
  }
  
  /***************************************************************************
   * Rewrite disjunctions to implications, as far as possible, e.g.
   * ~A OR B OR ~C OR D becomes (A AND C) => (B OR D).
   * @param exp an expression
   * @return transformed exp (exp itself, if rule is not applicable).
   **************************************************************************/
  private static Expression toImpliesFormula(Expression exp)
  {
    Vector<Expression> pos = new Vector<Expression>();
    Vector<Expression> neg = new Vector<Expression>();
    separatePolarity(exp, pos, neg);
    if (neg.isEmpty()) return combine(pos, true, false);
    if (pos.isEmpty()) return combine(neg, true, true);
    return Construct.impliesFormula(
      combine(neg, false, false), combine(pos, true, false));
  }
  
  /***************************************************************************
   * Combine formulas to conjunction or disjunction.
   * @param v a non-empty vector of Expression objects denoting formulas.
   * @param or if true, formulas are disjoined, else formulas are conjoined.
   * @param negate, if true, formulas are negated before combination.
   * @return the combined formula.
   **************************************************************************/
  private static Expression combine(Vector<Expression> v, boolean or, boolean negate)
  {
    int n = v.size();
    Expression result = (Expression)v.elementAt(n-1);
    if (negate) result = Construct.notFormula(result);
    for (int i=n-2; i>=0; i--)
    {
      Expression elem = (Expression)v.elementAt(i);
      if (negate) elem = Construct.notFormula(elem);
      if (or)
        result = Construct.orFormula(elem, result);
      else
        result = Construct.andFormula(elem, result);
    }
    return result;
  }
  
  /**************************************************************************
   * Separate formulas in disjunction accoring to polarity.
   * @param exp an expression.
   * @param pos vector to which all unnegated formulas of exp are added.
   * @param neg vector to which all negated summands of exp are added.
   ************************************************************************/
  private static void separatePolarity(Expression exp, Vector<Expression> pos, Vector<Expression> neg)
  {
    if (exp instanceof OrFormula)
    {
      OrFormula exp0 = (OrFormula)exp;
      separatePolarity(exp0.getFirst(), pos, neg);
      separatePolarity(exp0.getSecond(), pos, neg);
    }
    else if (exp instanceof ImpliesFormula)
    {
      ImpliesFormula exp0 = (ImpliesFormula)exp;
      separatePolarityNeg(exp0.getFirst(), neg, pos);
      separatePolarity(exp0.getSecond(), pos, neg);
    }
    else if (exp instanceof NotFormula)
      separatePolarityNeg(((NotFormula)exp).getBase(), neg, pos);
    else if (exp instanceof NotEqualsFormula)
    {
      NotEqualsFormula exp0 = (NotEqualsFormula)exp;
      neg.add(Construct.equalsFormula(exp0.getFirst(), exp0.getSecond()));
    }
    else if (exp instanceof NotEquivalentFormula)
    {
      NotEquivalentFormula exp0 = (NotEquivalentFormula)exp;
      neg.add(Construct.equivalentFormula(exp0.getFirst(), exp0.getSecond()));
    }
    else
      pos.add(exp);
  }
  
  /**************************************************************************
   * Separate formulas in conjunction according to polarity.
   * @param exp an expression.
   * @param pos vector to which all unnegated formulas of exp are added.
   * @param neg vector to which all negated summands of exp are added.
   ************************************************************************/
  private static void separatePolarityNeg(Expression exp, Vector<Expression> pos, Vector<Expression> neg)
  {
    if (exp instanceof AndFormula)
    {
      AndFormula exp0 = (AndFormula)exp;
      separatePolarityNeg(exp0.getFirst(), pos, neg);
      separatePolarityNeg(exp0.getSecond(), pos, neg);
    }
    else if (exp instanceof NotFormula)
      separatePolarity(((NotFormula)exp).getBase(), neg, pos);
    else if (exp instanceof NotEqualsFormula)
    {
      NotEqualsFormula exp0 = (NotEqualsFormula)exp;
      neg.add(Construct.equalsFormula(exp0.getFirst(), exp0.getSecond()));
    }
    else if (exp instanceof NotEquivalentFormula)
    {
      NotEquivalentFormula exp0 = (NotEquivalentFormula)exp;
      neg.add(Construct.equivalentFormula(exp0.getFirst(), exp0.getSecond()));
    }
    else
      pos.add(exp);
  }

  /**************************************************************************
   * If exp = exp1 + 0 or exp = 0 + exp1, rewrite to exp1.
   * @return transformed exp (exp itself, if rule is not applicable).
   *************************************************************************/
  private static Expression toSum(PlusTerm exp)
  {
    Expression first = exp.getFirst();
    Expression second = exp.getSecond();
    if (first instanceof Number && ((Number)first).getDigits().equals("0"))
      return second;
    if (second instanceof Number && ((Number)second).getDigits().equals("0"))
      return first;
    return exp;
  }
  
  /**************************************************************************
   * If exp = exp1 op exp, where op is a comparison operator (=,/=,<,<=,>,>=)
   * and exp1 or exp2 is a sum, bring all negative summands to the other
   * side of the inequality.
   * @param exp an expression
   * @return transformed exp (exp itself, if rule is not applicable).
   *************************************************************************/
   private static Expression toEquation(Expression exp)
   {
     Expression exp0 = toDefinition(exp);
     if (exp0 != null) return exp0;
     ExpressionPair pair = decomposeEquation(exp);
     if (pair == null) return exp;
     Vector<Expression> left = new Vector<Expression>();
     Vector<Expression> right = new Vector<Expression>();
     separateSigns(pair.first, left, right);
     separateSigns(pair.second, right, left);
     if (right.isEmpty())
     {
       if (left.size() == 2)
       {
         Expression e0 = (Expression)left.elementAt(0);
         Expression e1 = (Expression)left.elementAt(1);
         if (e0 instanceof Identifier && e1 instanceof Number)
           return toEquation(exp, e0, Construct.negationTerm(e1));
         if (e1 instanceof Identifier && e0 instanceof Number)
           return toEquation(exp, e1, Construct.negationTerm(e0));
       }
     }
     else if (left.isEmpty())
     {
       if (right.size() == 2)
       {
         Expression e0 = (Expression)right.elementAt(0);
         Expression e1 = (Expression)right.elementAt(1);
         if (e0 instanceof Identifier && e1 instanceof Number)
           return toEquation(exp, Construct.negationTerm(e1), e0);
         if (e1 instanceof Identifier && e0 instanceof Number)
           return toEquation(exp, Construct.negationTerm(e0), e1);
       }
     }
     return toEquation(exp, makeSum(left), makeSum(right));
   }

   /**************************************************************************
    * Convert equation to definition.
    * @param exp an expression
    * @return the definition form of the equation (or null, if none)
    *************************************************************************/
   private static Expression toDefinition(Expression exp)
   {
     if (!(exp instanceof EqualsFormula)) return null;
     EqualsFormula exp0 = (EqualsFormula)exp; 
     Expression exp1 = exp0.getFirst();
     if (exp1 instanceof Identifier) return exp;
     Expression exp2 = exp0.getSecond();
     if (exp2 instanceof Identifier) 
       return Construct.equalsFormula(exp2, exp1);
     return null;
   }
   
   /**************************************************************************
    * Pair of expressions
    *************************************************************************/
   private static final class ExpressionPair
   {
     Expression first;
     Expression second;
     public ExpressionPair(Expression first, Expression second)
     {
       this.first = first;
       this.second = second;
     }
   }
   
   /**************************************************************************
    * Decompose equation into pair of expressions.
    * @param exp an expression
    * @return the pair of the equation parts (null, if exp is not an equation)
    *************************************************************************/
   private static ExpressionPair decomposeEquation(Expression exp)
   {
     Expression exp1 = null;
     Expression exp2 = null;
     if (exp instanceof EqualsFormula)
     {
       EqualsFormula exp0 = (EqualsFormula)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }
     else if (exp instanceof NotEqualsFormula)
     {
       NotEqualsFormula exp0 = (NotEqualsFormula)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     } 
     else if (exp instanceof LessFormula)
     {
       LessFormula exp0 = (LessFormula)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }
     else if (exp instanceof LessEqualFormula)
     {
       LessEqualFormula exp0 = (LessEqualFormula)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }
     else if (exp instanceof GreaterFormula)
     {
       GreaterFormula exp0 = (GreaterFormula)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }
     else if (exp instanceof GreaterEqualFormula)
     {
       GreaterEqualFormula exp0 = (GreaterEqualFormula)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }
     else
       return null;
     return new ExpressionPair(exp1, exp2);
   }
   
   /**************************************************************************
    * Separate summands in exp according to signs.
    * @param exp an expression.
    * @param pos vector to which all positive summands of exp are added.
    * @param neg vector to which all negative summands of exp are added.
    ************************************************************************/
   private static void separateSigns(Expression exp, Vector<Expression> pos, Vector<Expression> neg)
   {
     if (exp instanceof PlusTerm)
     {
       PlusTerm exp0 = (PlusTerm)exp;
       separateSigns(exp0.getFirst(), pos, neg);
       separateSigns(exp0.getSecond(), pos, neg);
       return;
     }
     if (exp instanceof MinusTerm)
     {
       MinusTerm exp0 = (MinusTerm)exp;
       separateSigns(exp0.getFirst(), pos, neg);
       separateSigns(exp0.getSecond(), neg, pos);
       return;
     }
     if (exp instanceof NegationTerm)
     {
       NegationTerm exp0 = (NegationTerm)exp;
       separateSigns(exp0.getBase(), neg, pos);
       return;
     }
     if (exp instanceof Number && ((Number)exp).getDigits().equals("0"))
       return;
     Expression e = extractSign(exp);
     if (e instanceof NegationTerm)
       neg.add(((NegationTerm)e).getBase());
     else
       pos.add(e);
   }
   
   /**************************************************************************
    * Extract from a product/quotient the sign to the outermost level.
    * @param exp an expression
    * @return an equivalent expression with the sign on the outermost
    *         level (if possible, otherwise exp itself)
    ************************************************************************/
   private static Expression extractSign(Expression exp)
   {
     Expression exp1 = null;
     Expression exp2 = null;
     if (exp instanceof TimesTerm)
     {
       TimesTerm exp0 = (TimesTerm)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }
     else if (exp instanceof DividesTerm)
     {
       DividesTerm exp0 = (DividesTerm)exp;
       exp1 = exp0.getFirst();
       exp2 = exp0.getSecond();
     }  
     else
       return exp;
     Expression e1 = extractSign(exp1);
     Expression e2 = extractSign(exp2);
     if (e1 instanceof NegationTerm && e2 instanceof NegationTerm)
       return toProduct(exp, 
         ((NegationTerm)e1).getBase(), ((NegationTerm)e2).getBase());
     if (e1 instanceof NegationTerm)
       return Construct.negationTerm(toProduct(exp, 
         ((NegationTerm)e1).getBase(), e2));
     if (e2 instanceof NegationTerm)
       return Construct.negationTerm(toProduct(exp, 
         e1, ((NegationTerm)e2).getBase()));
     return toProduct(exp, e1, e2);
   }
   
   /**************************************************************************
    * Decide if exp is a number with specific digit representation.
    * @param exp an expression
    * @param digits a sequence of digits
    * @return true iff exp is a number with the denoted digits
    ************************************************************************/
   private static boolean isNumber(Expression exp, String digits)
   {
     if (!(exp instanceof Number)) return false;
     Number n = (Number)exp;
     return n.getDigits().equals(digits);
   }
   
   /**************************************************************************
    * Combine exp1 and exp2 to the same kind of expression as exp.
    * @param exp a product or quotient expression.
    * @param exp1 an expression.
    * @param exp2 an expression.
    * @return the combination of exp1 and exp2 of the same kind as exp.
    *************************************************************************/
   private static Expression toProduct(Expression exp, 
     Expression exp1, Expression exp2)
   {
     if (exp instanceof TimesTerm)
     {
       if (isNumber(exp1, "1")) return exp2;
       if (isNumber(exp2, "1")) return exp1;
       TimesTerm result = Construct.timesTerm(exp1, exp2);
       result.setType(((TimesTerm)exp).getType());
       return result;
     }
     if (exp instanceof DividesTerm)
     {
       if (isNumber(exp2, "1")) return exp1;
       return Construct.dividesTerm(exp1, exp2);
     }
     return null;
   }
   
   /**************************************************************************
    * Construct sum of the expressions in denoted vector.
    * @param v a vector of summands.
    * @return the sum of the elements of v.
    *************************************************************************/
   private static Expression makeSum(Vector<Expression> v)
   {
     Expression exp = null;
     int n = v.size();
     for (int i=0; i<n; i++)
     {
       Expression e = (Expression)(v.elementAt(i));
       if (isNumber(e, "0")) continue;
       if (exp == null)
         exp = e;
       else
         // take care to preserve term order 
         // otherwise formula masks may change
         exp = Construct.plusTerm(exp, e);
     }
     if (exp == null) exp = Construct.number("0");
     return exp;
   }
   
   /**************************************************************************
    * combine exp1 and exp2 to the same kind of equation as exp.
    * @param exp an equation or inequation.
    * @param exp1 an expression.
    * @param exp2 an expresison.
    * @return the combination of exp1 and exp2 of the same kind as exp.
    *************************************************************************/
   private static Expression toEquation(Expression exp, 
     Expression exp1, Expression exp2)
   {
     if (exp instanceof EqualsFormula)
       return Construct.equalsFormula(exp1, exp2);
     else if (exp instanceof NotEqualsFormula)
       return Construct.notEqualsFormula(exp1, exp2);
     else if (exp instanceof LessFormula)
       return Construct.lessFormula(exp1, exp2);
     else if (exp instanceof LessEqualFormula)
       return Construct.lessEqualFormula(exp1, exp2);
     else if (exp instanceof GreaterFormula)
       return Construct.greaterFormula(exp1, exp2);
     else if (exp instanceof GreaterEqualFormula)
       return Construct.greaterEqualFormula(exp1, exp2);
     return null;
   }
   
   /***************************************************************************
    * If exp = (lambda x. e)(a), then rewrite exp to e[a/x]
    * @param exp an expression
    * @return transformed exp (exp itself, if rule is not applicable).
    **************************************************************************/
   private static Expression processLambdaApplication(Expression exp)
   {
     // check whether rule is applicable
     if (!(exp instanceof ApplicationExpression)) return exp;
     ApplicationExpression exp0 = (ApplicationExpression)exp;
     Expression fun = exp0.getFunction();
     if (!(fun instanceof LambdaTerm)) return exp;
     
     // determine substitution
     LambdaTerm fun0 = (LambdaTerm)fun;
     TypedIdentifier[] vars = fun0.getVariables();
     Expression[] args = exp0.getArguments();
     int n = vars.length;
     if (n != args.length)
     {
       System.out.println("processLambdaApplication: " +
         "number of arguments differs from number of parameters");
       System.exit(-1);
     }
     ValueDeclIdentifier[] idents = new ValueDeclIdentifier[n];
     for (int i=0; i<n; i++)
       idents[i] = vars[i].getIdentifier();
     
     // apply substitution
     Expression base = ((LambdaTerm)fun).getBase();
     return Substitute.substitute(base, idents, args, false);
   }
   
   /***************************************************************************
    * If exp = (array x. e)[a], then rewrite exp to e[a/x]
    * @param exp an expression
    * @return transformed exp (exp itself, if rule is not applicable).
    **************************************************************************/
   private static Expression processArraySelection(Expression exp)
   {
     // check whether rule is applicable
     if (!(exp instanceof SelectionTerm)) return exp;
     SelectionTerm exp0 = (SelectionTerm)exp;
     Expression fun = exp0.getBase();
     if (!(fun instanceof ArrayTerm)) return exp;
     Selector sel = exp0.getSelector();
     if (!(sel instanceof SelectorIndex)) return exp;
     
     // determine substitution
     ArrayTerm fun0 = (ArrayTerm)fun;
     SelectorIndex sel0 = (SelectorIndex)sel;
     TypedIdentifier[] vars = fun0.getVariables();
     Expression[] args = new Expression[] { sel0.getIndex() };
     ValueDeclIdentifier[] idents = 
       new ValueDeclIdentifier[] { vars[0].getIdentifier() };
     
     // apply substitution
     return Substitute.substitute(fun0.getBase(), idents, args, false);
   }
   
   /***************************************************************************
    * If exp = let x = e in a, then rewrite exp to e[a/x]
    * @param exp an expression
    * @return transformed exp (exp itself, if rule is not applicable).
    **************************************************************************/
   private static Expression processLetExpression(Expression exp)
   { 
     // check whether rule is applicable
     if (!(exp instanceof LetExpression)) return exp;
     LetExpression exp0 = (LetExpression)exp;
     
     // we must apply one substitution after the other because the bound 
     // identifiers may occur in the substitution expressions
     Expression base = exp0.getBase();
     ValueDeclaration[] decls = exp0.getDeclarations();
     int n = decls.length;
     for (int i=n-1; i>=0; i--)
     {
       ValueDeclIdentifier ident = decls[i].getValueDeclIdentifier();
       Expression value = decls[i].getValue();
       base = Substitute.substitute(base,
         new ValueDeclIdentifier[] { ident }, 
         new Expression[] { value }, false);
     }
     return base;
   }
   
   /***************************************************************************
    * If exp of form like (e1 AND (e2 AND e3)) AND e4) then
    * rewrite exp to e1 AND (e2 AND (e3 AND e4)).
    * @param exp an expression
    * @return transformed exp (exp itself, if rule is not applicable).
    **************************************************************************/
   private Expression rearrangeAndFormula(AndFormula exp) 
   { 
     Vector<Expression> result = new Vector<Expression>();
     collect(exp, true, result);
     return ASTUtil.toConjunction(ASTUtil.toExpressionArray(result));
   }
   
   /***************************************************************************
    * If exp of form like (e1 OR (e2 OR e3)) OR e4) then
    * rewrite exp to e1 OR (e2 OR (e3 OR e4)).
    * @param exp an expression
    * @return transformed exp (exp itself, if rule is not applicable).
    **************************************************************************/
   private Expression rearrangeOrFormula(OrFormula exp) 
   { 
     Vector<Expression> result = new Vector<Expression>();
     collect(exp, false, result);
     return ASTUtil.toDisjunction(ASTUtil.toExpressionArray(result));
   }
   
   /***************************************************************************
    * Collect parts of conjunction/disjunction.
    * @param exp an expression
    * @param and true if conjunction is decomposed, false, if disjunction.
    * @param result a vector to which those elements are appended whose 
    *        conjunction (if and) resp. disjunction (if not and) 
    *        is logically equivalent to exp.
    **************************************************************************/
   private void collect(Expression exp, boolean and, Vector<Expression> result)
   {
     // recursive structure to preserve order of formulas
     if ((and && !(exp instanceof AndFormula)) ||
         (!and && !(exp instanceof OrFormula)))
     {
       result.add(exp);
       return;
     }
     collect(((BinaryExpression)exp).getFirst(), and, result);
     collect(((BinaryExpression)exp).getSecond(), and, result);
   }
   
   /**************************************************************************
    * Normalize fraction to a single quotient of products.
    * @param exp an expression.
    * @return the normalized expression.
    *************************************************************************/
   private Expression toFraction(Expression exp)
   {
     Vector<Expression> numerators = new Vector<Expression>();
     Vector<Type> ntypes = new Vector<Type>();
     Vector<Expression> denominators = new Vector<Expression>();
     Vector<Type> dtypes = new Vector<Type>();
     separateFactors(exp, numerators, ntypes, denominators, dtypes);
     if (denominators.isEmpty()) return makeProduct(numerators, ntypes);
     return Construct.dividesTerm(
       makeProduct(numerators, ntypes), makeProduct(denominators, dtypes));
   }
   
   /**************************************************************************
    * Construct product of the expressions in denoted vector.
    * @param v a vector of factors.
    * @param t a vector of types for each multiplication of factors.
    * @return the product of the elements of v.
    *************************************************************************/
   private static Expression makeProduct(Vector<Expression> v,
     Vector<Type> t)
   {
     Expression exp = null;
     int n = v.size();
     for (int i=0; i<n; i++)
     {
       Expression e = (Expression)(v.elementAt(i));
       if (isNumber(e, "0")) return e;
       if (isNumber(e, "1")) continue;
       if (exp == null)
         exp = e;
       else
       {
         // take care to preserve term order 
         // otherwise formula masks may change
         exp = Construct.timesTerm(exp, e);
         ((TimesTerm)exp).setType(t.get(i-1));
       }
     }
     if (exp == null) exp = Construct.number("1");
     return exp;
   }
   
   /**************************************************************************
    * Separate factors in exp
    * @param exp an expression; this expression shall after the call equal
    * the product of all elements of numerators divided by the product
    * of all elements of denominators.
    * @param numerators vector to which all factors are added that should
    *        appear as numerators.
    * @param ntypes vector to which for each multiplication of a numerator
    *        the type of the result is added.
    * @param denominators vector to which all factors are added that should
    *        appear as denominators.
    * @param dints vector to which for each multiplication of a denominator
    *        a boolean is added that indicates whether the result is an integer
    * @param dtypes vector to which for each multiplication of a denominator
    *        the type of the result is added.     
    *************************************************************************/
   private static void separateFactors(Expression exp, 
     Vector<Expression> numerators, Vector<Type> ntypes,
     Vector<Expression> denominators, Vector<Type> dtypes)
   {
     if (exp instanceof TimesTerm)
     {
       TimesTerm exp0 = (TimesTerm)exp;
       separateFactors(exp0.getFirst(), 
           numerators, ntypes, denominators, dtypes);
       ntypes.add(exp0.getType());
       separateFactors(exp0.getSecond(), 
           numerators, ntypes, denominators, dtypes);
       return;
     }
     if (exp instanceof DividesTerm)
     {
       DividesTerm exp0 = (DividesTerm)exp;
       separateFactors(exp0.getFirst(), 
           numerators, ntypes, denominators, dtypes);
       separateFactors(exp0.getSecond(), 
           denominators, dtypes, numerators, ntypes);
       return;
     }
     if (exp instanceof PowerTerm)
     {
       PowerTerm exp0 = (PowerTerm)exp;
       Expression base = exp0.getFirst();
       Expression power = extractSign(exp0.getSecond());
       if (power instanceof NegationTerm)
       {
         Expression pbase = ((NegationTerm)power).getBase();
         if (isNumber(pbase, "1"))
           denominators.add(base);
         else
           denominators.add(Construct.powerTerm(base, pbase));
         return;
       }
     }
     numerators.add(exp);
   }
}
