// ----------------------------------------------------------------------------
// Top.java
// The top panel of the GUI.
// $Id: Top.java,v 1.68 2011/11/16 13:42:29 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------

package fmrisc.ProofNavigator.SWT;

import org.eclipse.swt.*;
import org.eclipse.swt.browser.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.layout.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.custom.*;
import org.eclipse.swt.events.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Commands.*;
import fmrisc.ProofNavigator.Proving.*;

import java.io.*;
import java.nio.charset.Charset;

/**
 * 
 * The top panel of the GUI.
 *
 */
public final class Top extends SashForm
{
  private final static String aboutText = 
    fmrisc.ProgramExplorer.SWT.TopWindow.aboutText;

  // URL of help document
  private String helpURL = 
    fmrisc.ProgramExplorer.SWT.TopWindow.helpURL;
  private String helpURLOld = 
      fmrisc.ProgramExplorer.SWT.TopWindow.helpURLOld;
  
  // various buttons
  private String stopImageName = "process-stop.png";
  private Image stopImage = null;
  private ToolItem stopButton = null;
  private String scatterImageName = "go-bottom.png";
  private Image scatterImage = null;
  private ToolItem scatterButton = null;
  private String decomposeImageName = "go-down.png";
  private Image decomposeImage = null;
  private ToolItem decomposeButton = null;
  private String previousImageName = "go-previous.png";
  private Image previousImage = null;
  private ToolItem previousButton = null;
  private String nextImageName = "go-next.png";
  private Image nextImage = null;
  private ToolItem nextButton = null;
  private String undoImageName = "edit-undo.png";
  private Image undoImage = null;
  private ToolItem undoButton = null;
  private String redoImageName = "edit-redo.png";
  private Image redoImage = null;
  private ToolItem redoButton = null;
  private String quitImageName = "system-log-out.png";
  private Image quitImage = null;
  private ToolItem quitButton = null;
  private String counterImageName = "help-browser.png";
  private Image counterImage = null;
  private ToolItem counterButton = null;
  private String autoImageName = "view-refresh.png";
  private Image autoImage = null;
  private ToolItem autoButton = null;
  private String autostarImageName = "edit-clear.png";
  private Image autostarImage = null;
  private ToolItem autostarButton = null;
  private String splitImageName = "edit-cut.png";
  private Image splitImage = null;
  private ToolItem splitButton = null;
  private String commandsImageName = "edit-find.png";
  private Image commandsImage = null;
  private ToolItem commandsButton = null;
  private String simplifyImageName = "emblem-system.png";
  private Image simplifyImage = null;
  private ToolItem simplifyButton = null;
  
  private Top top = this;
  private Shell shell = null;
  
  private Menu menuBar = null;
  private SashForm sashForm = null;
  
  private Group treeGroup = null;
  private SashForm infoGroup = null;
  private Composite declGroup = null;
  private Group stateGroup = null;
  private Group ioGroup = null;
  private ToolBar buttonGroup = null;
  
  private ProofTree tree = null;
  private Browser browser = null;
  private Button viewButton = null;
  private Text output = null;
  private Text input = null;
  
  private Menu commandsMenu = null;
  private MenuItem simplifyItem;
  
  // the streams used by the text widgets
  private PrintWriter writer; 
  private TextInputStream istream;
  
  // the various kinds of cursors
  private final Cursor arrowCursor =  
    new Cursor(getDisplay(), SWT.CURSOR_ARROW);
  private final Cursor ibeamCursor = 
    new Cursor(getDisplay(), SWT.CURSOR_IBEAM);
  private final Cursor waitCursor = 
    new Cursor(getDisplay(), SWT.CURSOR_WAIT);
  
  // are declarations visible?
  private boolean declVisible = true;
  
  // the character set name (fix endian since UTF-16 creates "byte order mark")
  private final static String charSetName = "UTF-16LE";
  
  // the ProgramExplorer window
  private fmrisc.ProgramExplorer.SWT.TopWindow topWindow = null;
  
  // a question that is currently asked
  private Question question = null;
  
  /****************************************************************************
   * Construct panel for denoted shell and with denoted style.
   * @param shell the shell to which the panel is attached.
   * @param style the style of the panel.
   ***************************************************************************/
  public Top(Shell shell, int style)
  {
    super(shell, style);
    this.shell = shell;
    initializeShell();
    createMenu();
    shell.setMenuBar(menuBar);
    initialize();
  }
  
  /****************************************************************************
   * Construct panel for denoted parent, shell, and style.
   * @param topWindow the ProgramExplorer top window
   * @param parent the parent of the panel.
   * @param shell the shell to which the panel is attached.
   * @param style the style of the panel.
   ***************************************************************************/
  public Top(fmrisc.ProgramExplorer.SWT.TopWindow topWindow, 
    Composite parent, Shell shell, int style)
  {
    super(parent, style);
    this.shell = shell;
    this.topWindow = topWindow;
    createMenu();
    initialize();
    Main.mainInternal(this);
  }
  
  /****************************************************************************
   * Get menu bar
   * @return the menu bar
   ***************************************************************************/
  public Menu getCurrentMenu()
  {
    return menuBar;
  }
  
  /****************************************************************************
   * Initialize shell
   ***************************************************************************/
  private void initializeShell()
  {
    // closing the shell raises the quit dialog
    shell.addShellListener(new ShellAdapter() { 
      public void shellClosed(ShellEvent e)
      {
        boolean confirm;
        if (MainSWT.isRestarted())
          confirm = ask("Restart Program", "Do you really want to restart?");
        else
          confirm = ask("Quit Program", "Do you really want to quit?");
        if (confirm) 
        {
          try
          {
            istream.close();
          }
          catch(IOException ex) { }
        }
        else
        { 
          e.doit = false;
          MainSWT.restart(false);
        }
      }
    });
    
    // set shell layout
    GridLayout shellLayout = new GridLayout();
    shellLayout.numColumns = 1;
    shellLayout.marginWidth = 0;
    shellLayout.marginHeight = 0;
    shell.setLayout(shellLayout);
    setLayoutData(new GridData(GridData.FILL_BOTH));
  }
  
  /****************************************************************************
   * Initialize all components.
   ***************************************************************************/
  private void initialize()
  {   
    // set own layout
    GridLayout topLayout =  new GridLayout();
    topLayout.numColumns = 1;
    topLayout.verticalSpacing = 3;
    topLayout.marginTop = 3;
    topLayout.marginHeight = 0;
    topLayout.marginWidth = 0;
    setLayout(topLayout);
    
    // create widgets
    createSashForm();
    createTreeGroup();
    createInfoGroup();
    createDeclGroup();
    createStateGroup();
    createIOGroup();
    createTree();
    createBrowser();
    createViewButton();
    
    // set relative component weights for sash forms
    sashForm.setWeights(new int[] {100, 200});
    infoGroup.setWeights(new int[] {200, 150});
    
    // create input/buttons/output fields
    output = new Text(ioGroup, SWT.BORDER | SWT.MULTI | SWT.V_SCROLL | 
        SWT.WRAP | SWT.READ_ONLY);
    input = new Text(ioGroup, SWT.BORDER | SWT.MULTI | SWT.WRAP);
    output.setLayoutData(new GridData(GridData.FILL_BOTH));
    GridData inputData = new GridData(GridData.FILL_HORIZONTAL);
    inputData.heightHint = 25;
    input.setLayoutData(inputData);
    input.setFocus();
    setFonts();
    input.addKeyListener(new KeyAdapter() {
      public void keyPressed(KeyEvent e)
      {
        if (e.character != '\t') return;
        int pos = input.getText().indexOf("[]");
        if (pos == -1) return;
        input.setSelection(pos, pos+2);
        e.doit = false;
      }
    });
        
    // add buttons to I/O group
    createButtons();
    createCommandsMenu();
    
    // set system input/output streams
    Charset charset = Charset.forName(charSetName);
    writer = new PrintWriter(new OutputStreamWriter(
               new TextOutputStream(output, charset), charset), true);
    writer.flush();
    State.setSystemOut(writer);
    istream = new TextInputStream(input, writer, charset);
    State.setSystemIn(new BufferedReader(
        new InputStreamReader(istream, charset)));
    
    // signal status
    signalRunning(false);
  }
  
  /**
   * This method initializes toolBar	
   *
   */
  private void createMenu()
  {
    menuBar = new Menu(shell, SWT.BAR);
  
    MenuItem fileTitle = new MenuItem(menuBar, SWT.CASCADE);
    fileTitle.setText("File");
    Menu fileMenu = new Menu(shell, SWT.DROP_DOWN);
    fileTitle.setMenu(fileMenu);
    MenuItem readItem = new MenuItem(fileMenu, SWT.NULL);
    readItem.setText("Read File");
    new MenuItem(fileMenu, SWT.SEPARATOR);
    // MenuItem resetItem = new MenuItem(fileMenu, SWT.NULL);
    // resetItem.setText("Restart Session");
    MenuItem quitItem = new MenuItem(fileMenu, SWT.NULL);
    quitItem.setText("Quit");

    MenuItem optionsTitle = new MenuItem(menuBar, SWT.CASCADE);
    optionsTitle.setText("Options");
    Menu optionsMenu = new Menu(shell, SWT.DROP_DOWN);
    optionsTitle.setMenu(optionsMenu);
    simplifyItem = new MenuItem(optionsMenu, SWT.NULL);
    if (State.useAutoSimplify())
      simplifyItem.setText("No Automatic Simplification");
    else
      simplifyItem.setText("Automatic Simplification");
    new MenuItem(optionsMenu, SWT.SEPARATOR);
    MenuItem fontItem = new MenuItem(optionsMenu, SWT.NULL);
    if (State.useBigFont())
      fontItem.setText("Smaller Font");
    else
      fontItem.setText("Bigger Font");
    
    MenuItem helpTitle = new MenuItem(menuBar, SWT.CASCADE);
    helpTitle.setText("Help");
    Menu helpMenu = new Menu(shell, SWT.DROP_DOWN);
    helpTitle.setMenu(helpMenu);
    MenuItem helpItem = new MenuItem(helpMenu, SWT.NULL);
    helpItem.setText("Online Manual");
    new MenuItem(helpMenu, SWT.SEPARATOR);
    MenuItem aboutItem = new MenuItem(helpMenu, SWT.NULL);
    aboutItem.setText("About RISC ProgramExplorer");
    
    // menu item "Read File": open file dialog
    readItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        FileDialog dialog = new FileDialog(getShell(), SWT.OPEN);
        dialog.setText("Read file");
        String name = dialog.open();
        if (name == null) return;
        MainSWT.process(new Read(name));
      }
    });
    
    // menu item "Reset": restart session
    // resetItem.addSelectionListener(new SelectionAdapter() {
    //   public void widgetSelected(SelectionEvent e) {
    //     MainSWT.restart(true);
    //     shell.close();
    //   }
    // });
    
    // menu item "Quit": quit program
    quitItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        shell.close();
      }
    });
   
    // menu item "(No) Automatic Simplification": turn on/off autosimplify
    simplifyItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        Proof proof = State.getProof();
        boolean auto;
        if (proof == null)
          auto = State.useAutoSimplify();
        else
          auto = proof.getCurrent().useAutoSimplify();
        if (auto)
          MainSWT.process(new Option("autosimp", "false")); 
        else
          MainSWT.process(new Option("autosimp", "true")); 
      }
    });
    
    // menu item "Bigger/Smaller Font": change font size 
    fontItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        if (topWindow != null)
          topWindow.resizeFont();
        else
          resizeFont();
        shell.setMenuBar(menuBar);
        if (topWindow == null) MainSWT.resize();
      }
    });
    
    // menu item "Help": show help page in browser
    helpItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        if (topWindow == null)
          browser.setUrl(helpURLOld);
        else
          browser.setUrl(helpURL);
        viewButton.setEnabled(true);
        viewButton.setText("View Declarations");
        declVisible = false;
      }
    });
    
    // menu item "About": show copyright message in browser
    aboutItem.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        browser.setText(aboutText);
        viewButton.setEnabled(true);
        viewButton.setText("View Declarations");
        declVisible = false;
      }
    });
    
  }
  
  public void resizeFont()
  { 
    State.setBigFont(!State.useBigFont());
    setFonts();
    State.getPresenter().resize();
    browser.refresh();
    // text is not automatically refreshed otherwise
    if (browser.getUrl().equals("about:blank"))
      browser.setText(browser.getText());
    createMenu();
  }
  
  /**
   * This method initializes sashForm 
   *
   */
  private void createSashForm()
  {
    sashForm = new SashForm(this, SWT.HORIZONTAL | SWT.SMOOTH);
    sashForm.setLayoutData(new GridData(GridData.FILL_BOTH));
  }
  
  /**
   * This method initializes the tree group.
   *
   */
  private void createTreeGroup()
  {
    treeGroup = new Group(sashForm, SWT.BORDER);
    treeGroup.setText("Proof Tree");
    treeGroup.setLayout(new FillLayout());
  }
  
  /**
   * This method initializes the info group.
   *
   */
  private void createInfoGroup()
  {
    infoGroup = new SashForm(sashForm, SWT.VERTICAL | SWT.SMOOTH);
    RowLayout infoLayout = new RowLayout();
    infoLayout.type = SWT.VERTICAL;
    infoLayout.spacing = 5;
    infoGroup.setLayout(infoLayout);;
  }
  
  /**
   * This method initializes the declaration group.
   *
   */
  private void createDeclGroup()
  {
    declGroup = new Composite(infoGroup, SWT.NULL);
    GridLayout layout = new GridLayout();
    declGroup.setLayout(layout);
    layout.numColumns = 1;
    layout.marginWidth = 0;
    layout.marginHeight = 0;
  }
  
  /**
   * This method initializes the state group.
   *
   */
  private void createStateGroup()
  {
    stateGroup = new Group(declGroup, SWT.BORDER);
    stateGroup.setText("Declarations");
    stateGroup.setLayout(new FillLayout());
    GridData data = new GridData(GridData.FILL_BOTH);
    stateGroup.setLayoutData(data);
  }
  
  /**
   * This method initializes the IO group. 
   *
   */
  private void createIOGroup()
  {
    ioGroup = new Group(infoGroup, SWT.BORDER);
    ioGroup.setText("Input/Output");
    GridLayout ioLayout = new GridLayout();
    ioLayout.numColumns = 1;
    ioLayout.verticalSpacing = 0;
    ioLayout.marginWidth = 0;
    ioLayout.marginHeight = 0;
    ioGroup.setLayout(ioLayout);
  }
  
  /**
   * This method initializes tree	
   *
   */
  private void createTree()
  {
    tree = new ProofTree(treeGroup);
  }
  
    
  /**
   * This method initializes browser	
   *
   */
  private void createBrowser()
  {
    browser = new Browser(stateGroup, SWT.MOZILLA | SWT.BORDER);
    browser.setBackground(MainSWT.getBackgroundColor());
    browser.setText(aboutText);
  }
  
  private void createViewButton()
  {
    viewButton = new Button(declGroup, SWT.PUSH);
    viewButton.setText("View Proof State");
    GridData data = new GridData(GridData.FILL_HORIZONTAL);
    viewButton.setLayoutData(data);
    viewButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        tree.setUnselected();
        if (!declVisible)
        {
          Main.displayDeclarations();
          return;
        }
        Proof proof = State.getProof();
        if (proof == null)
        {
          Main.displayDeclarations();
          return;
        }
        ProofState current = proof.getCurrent();
        MainSWT.setProofState(current);
      }
    });
    viewButton.setEnabled(false);
  }
  
  /**
   * This method creates buttons
   *
   */
  private void createButtons()
  {
    buttonGroup = new ToolBar(ioGroup, SWT.HORIZONTAL | SWT.FLAT);
    
    previousButton = new ToolItem(buttonGroup, SWT.PUSH);
    previousImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + previousImageName));
    previousButton.setImage(previousImage);
    previousButton.setToolTipText("prev: Previous Open State");
    previousButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Prev());
      }
    });
    previousButton.setEnabled(false);
    
    nextButton = new ToolItem(buttonGroup, SWT.PUSH);
    nextImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + nextImageName));
    nextButton.setImage(nextImage);
    nextButton.setToolTipText("next: Next Open State");
    nextButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Next());
      }
    });
    nextButton.setEnabled(false);
    
    undoButton = new ToolItem(buttonGroup, SWT.PUSH);
    undoImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + undoImageName));
    undoButton.setImage(undoImage);
    undoButton.setToolTipText("undo: Undo Command");
    undoButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Undo(null));
      }
    });
    undoButton.setEnabled(false);
    
    redoButton = new ToolItem(buttonGroup, SWT.PUSH);
    redoImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + redoImageName));
    redoButton.setImage(redoImage);
    redoButton.setToolTipText("redo: Redo Command");
    redoButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Redo(null));
      }
    });
    redoButton.setEnabled(false);
    
    new ToolItem(buttonGroup, SWT.SEPARATOR);
    
    scatterButton = new ToolItem(buttonGroup, SWT.PUSH);
    scatterImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + scatterImageName));
    scatterButton.setImage(scatterImage);
    scatterButton.setToolTipText("scatter: Scatter State");
    scatterButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Scatter());
      }
    });
    scatterButton.setEnabled(false);
    
    decomposeButton = new ToolItem(buttonGroup, SWT.PUSH);
    decomposeImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + decomposeImageName));
    decomposeButton.setImage(decomposeImage);
    decomposeButton.setToolTipText("decompose: Decompose State");
    decomposeButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Decompose());
      }
    });
    decomposeButton.setEnabled(false);
    
    splitButton = new ToolItem(buttonGroup, SWT.PUSH);
    splitImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + splitImageName));
    splitButton.setImage(splitImage);
    splitButton.setToolTipText("split: Split State");
    splitButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Split(null));
      }
    });
    splitButton.setEnabled(false);
    
    counterButton = new ToolItem(buttonGroup, SWT.PUSH);
    counterImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + counterImageName));
    counterButton.setImage(counterImage);
    counterButton.setToolTipText("counterexamle: Generate Counterexample");
    counterButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new CounterExample());
      }
    });
    counterButton.setEnabled(false);
       
    autostarButton = new ToolItem(buttonGroup, SWT.PUSH);
    autostarImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + autostarImageName));
    autostarButton.setImage(autostarImage);
    autostarButton.setToolTipText("autostar: Execute \"auto\" also in Sibling States");
    autostarButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new AutoStar());
      }
    });
    autostarButton.setEnabled(false);
    
    autoButton = new ToolItem(buttonGroup, SWT.PUSH);
    autoImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + autoImageName));
    autoButton.setImage(autoImage);
    autoButton.setToolTipText("auto: Close State by Automatic Formula Instantiation");
    autoButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Auto(null));
      }
    });
    autoButton.setEnabled(false);
    
    stopButton = new ToolItem(buttonGroup, SWT.PUSH);
    stopImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + stopImageName));
    stopButton.setImage(stopImage);
    stopButton.setToolTipText("Abort Prover Activity");
    stopButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        Prover prover = State.getRunningProver();
        if (prover == null) return;
        stopButton.setEnabled(false);
        boolean okay = prover.abort();
        if (!okay)
        {
          System.out.println("Prover could not be restarted");
          Main.exit(-1);
        }
        stopButton.setEnabled(true);
      }
    });
    stopButton.setEnabled(false);
 
    simplifyButton = new ToolItem(buttonGroup, SWT.PUSH);
    simplifyImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + simplifyImageName));
    simplifyButton.setImage(simplifyImage);
    simplifyButton.setToolTipText("Simplify State");
    simplifyButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Simplify(null));
      }
    });
    simplifyButton.setEnabled(false);
    
    commandsButton = new ToolItem(buttonGroup, SWT.DROP_DOWN);
    commandsImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + commandsImageName));
    commandsButton.setImage(commandsImage);
    commandsButton.setToolTipText("Command List");
    commandsButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        commandsMenu.setVisible(!commandsMenu.isVisible());
      }
    });
    commandsButton.setEnabled(false);
    
    new ToolItem(buttonGroup, SWT.SEPARATOR);
    
    quitButton = new ToolItem(buttonGroup, SWT.PUSH);
    quitImage = new Image(getDisplay(),
      Main.class.getResourceAsStream("/" + quitImageName));
    quitButton.setImage(quitImage);
    quitButton.setToolTipText("quit: Quit Proof");
    quitButton.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Quit());
      }
    });
    quitButton.setEnabled(false);
    
  }
  
  /***************************************************************************
   * Set button enabled status (except stop button).
   * @param enabled are buttons enabled?
   **************************************************************************/
  private void setButtonsEnabled(final boolean enabled)
  {
    previousButton.setEnabled(enabled);
    nextButton.setEnabled(enabled);
    scatterButton.setEnabled(enabled);
    decomposeButton.setEnabled(enabled);
    splitButton.setEnabled(enabled);
    undoButton.setEnabled(enabled);
    redoButton.setEnabled(enabled);
    counterButton.setEnabled(enabled);
    autoButton.setEnabled(enabled);
    autostarButton.setEnabled(enabled);
    // button enabled in updateMenu()
    // simplifyButton.setEnabled(enabled);
    commandsButton.setEnabled(enabled);
    quitButton.setEnabled(enabled);
  }
  
  /***************************************************************************
   * Create pop-up command menu.
   **************************************************************************/
  private void createCommandsMenu()
  {
    commandsMenu  = new Menu(shell, SWT.POP_UP);
    commandsMenu.setVisible(false);
    MenuItem item;
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("assume []: Use and Prove Assumption");
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("assume [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("case []: Performe Case Distinction");
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("case [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("expand []: Expand Definition(s)");
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("expand [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("lemma []: Import Lemma(s)"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("lemma [];");
      }
    });
    
    new MenuItem(commandsMenu, SWT.SEPARATOR);
   
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("auto []: Close State by Instantiation of Formula(s)"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("auto [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("expand [] in []: Expand Definitions in Formula"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("expand [] in [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("flatten: Flatten Formulas"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Flatten());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("flip []: Flip Formula"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("flip [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("goal []: Make Formula Goal"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("goal [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("goto []: Go to State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("goto [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("instantiate [] in []: Instantiate Variables in Formula"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("instantiate [] in [];");
      }
    });
        
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("open: Show Open States"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Open());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("skolemize: Skolemize Formulas"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Skolemize());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("split []: Split Formula"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("split [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("state []: Print State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("state [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("typeaxiom [] in []: Instantiate Type Axiom"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("typeaxiom [] in [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("undo []: Undo Back to State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("undo [];");
      }
    });
    
    new MenuItem(commandsMenu, SWT.SEPARATOR);
     
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("induction: Induction on Goal"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Induction(null, null));
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("induction []: Induction on Specific Variable of Goal"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("induction [];");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("induction [] in []: Induction on Variable of Formula"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("induction [] in [];");
      }
    });
     
    new MenuItem(commandsMenu, SWT.SEPARATOR);
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("prev: Previous Open State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Prev());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("next: Next Open State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Next());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("undo: Undo Command"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Undo(null));
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("redo: Redo Command"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Redo(null));
      }
    });
    
    new MenuItem(commandsMenu, SWT.SEPARATOR);
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("scatter: Scatter State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Scatter());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("decompose: Decompose State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Decompose());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("split: Split State"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Split(null));
      }
    });
       
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("counterexample: Generate Counterexample"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("counterexample;");
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("autostar: Execute \"auto\" also in Sibling States"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new AutoStar());
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("auto: Close State by Automatic Formula Instantiation"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Auto(null));
      }
    });
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("simplify []: Simplify Formula(s)"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        enter("simplify [];");
      }
    });

    new MenuItem(commandsMenu, SWT.SEPARATOR);
    
    item = new MenuItem(commandsMenu, SWT.PUSH);
    item.setText("quit: Quit Proof"); 
    item.addSelectionListener(new SelectionAdapter() {
      public void widgetSelected(SelectionEvent e) {
        MainSWT.process(new Quit());
      }
    });
  }
    
  /***************************************************************************
   * Open dialog to ask yes/no question.
   * @param title the dialog title
   * @param question the question asked
   * @return true iff question is answered positively.
   **************************************************************************/
  public boolean ask(String title, String question)
  {
    MessageBox box = 
      new MessageBox(shell, SWT.YES | SWT.NO | SWT.ICON_QUESTION
          | SWT.APPLICATION_MODAL);
    box.setText(title);
    box.setMessage(question);
    return box.open() == SWT.YES;
  }
  
  /***************************************************************************
   * Open dialog to ask question with a textual answer.
   * @param title the dialog title
   * @param text the question asked
   * @return the answer to the question.
   **************************************************************************/
  public String askText(String title, String text)
  {
    input.setEnabled(false);
    question = new Question(shell, text);
    question.setText(title);
    String answer = (String)question.open();
    question = null;
    // shell may be disposed
    if (!input.isDisposed()) input.setEnabled(true);
    return answer;
  }
  
  /***************************************************************************
   * Set the answer to a currently posted question.
   * @param answer the answer.
   **************************************************************************/
  public void setAnswer(String answer)
  {
    question.setAnswer(answer);
  }
  
  /****************************************************************************
   * Set the current proof.
   * @param proof the current proof.
   * @param enabled true if the proof is to be enabled.
   ***************************************************************************/
  public void setProof(Proof proof, boolean enabled)
  {
    tree.setProof(proof);
    final boolean enabled0 = enabled;
    browser.getDisplay().syncExec(new Runnable() {
      public void run() {
        Proof cproof = State.getProof();
        setButtonsEnabled(cproof != null && enabled0);
        if (cproof == null) stopButton.setEnabled(false);
      }}
    );
  }
  
  /****************************************************************************
   * Signal current proof state.
   * @param state the current proof state.
   ***************************************************************************/
  public void setProofState(ProofState state)
  {
    if (state == null)
    {
      browser.getDisplay().syncExec(new Runnable() {
        public void run() 
        { 
          browser.setText("<html><body></body></html>"); 
          input.setFocus();
        }
      });
      return;
    }
    File file = State.getPresenter().getFile(state);
    displayFile("Proof State", file, true);
  }
  
  /****************************************************************************
   * Display file.
   * @param title the title of the display.
   * @param file the file to be displayed.
   * @param state true iff file displays proof state.
   ***************************************************************************/
  public void displayFile(final String title, final File file, 
    final boolean state)
  {
    browser.getDisplay().syncExec(new Runnable() {
      public void run() {
        stateGroup.setText(title);
        if (file == null)
        {
          browser.setText("Error: " + State.getError());
          State.resetError();
          return;
        }
        browser.setUrl("file://" + file.getAbsolutePath());
        input.setFocus();
        declVisible = !state;
        if (declVisible)
        {
          if (State.getProof() == null)
            viewButton.setEnabled(false);
          else
            viewButton.setText("View Proof State");
        }
        else
        {
          viewButton.setText("View Declarations");
          viewButton.setEnabled(true);
        }
      }});
  }
  
  /***************************************************************************
   * Signal run status of background process.
   * @param running if we are waiting for background process to complete.
   **************************************************************************/
  public void signalRunning(final boolean running)
  {
    getDisplay().syncExec(new Runnable() {
      public void run() {
        if (running)
        {
          top.setCursor(waitCursor);
          browser.setCursor(waitCursor);
          input.setCursor(waitCursor);
          output.setCursor(waitCursor);
          input.setEditable(false);
          if (State.getProof() != null)
          {
            setButtonsEnabled(false);
            stopButton.setEnabled(true);
          }
        }
        else
        {
          top.setCursor(arrowCursor);
          browser.setCursor(arrowCursor);
          input.setCursor(ibeamCursor);
          output.setCursor(arrowCursor);
          input.setEditable(true);
          if (State.getProof() != null)
          {
            setButtonsEnabled(true);
            stopButton.setEnabled(false);
          }
        }
        tree.setRunning(running);
      }
    });
  }
  
  /***************************************************************************
   * Set fonts depending on user settings.
   **************************************************************************/
  private void setFonts()
  {
    int fontSize = State.useBigFont() ? 12 : 10;
    Font mono = new Font(input.getDisplay(), "Monospace", 
        fontSize, SWT.NORMAL);
    input.setFont(mono);
    output.setFont(mono);
  }
  
  /***************************************************************************
   * Update menu depending on current system state.
   **************************************************************************/
  public void updateMenu()
  {
    if (simplifyItem == null) return;
    simplifyItem.getDisplay().syncExec(new Runnable() {
      public void run() {
        Proof proof = State.getProof();
        boolean auto;
        if (proof == null)
          auto = State.useAutoSimplify();
        else
        {
          ProofState current = proof.getCurrent();
          if (current == null) return;
          auto = proof.getCurrent().useAutoSimplify();
        }
        if (auto) 
        {
          simplifyItem.setText("No Automatic Simplification");
          simplifyButton.setEnabled(false);
        }
        else 
        {
          simplifyItem.setText("Automatic Simplification");
          simplifyButton.setEnabled(true);
        }
        simplifyItem.getDisplay().update(); 
      }
    });
  }
  
  /****************************************************************************
   * Enter input denoted by string
   * @param string the input to be added.
   **************************************************************************/
  public void enter(String string)
  {
    if (question != null)
    {
      question.setAnswer(string);
      return;
    }
    input.setText(string);
    int pos = string.indexOf("[]");
    if (pos == -1) return;
    input.setSelection(pos, pos+2);
  }
  
  /****************************************************************************
   * Process command denoted by command string.
   * @param cstring the string denoting the command
   **************************************************************************/
  public void process(String cstring)
  {
    if (question != null)
    {
      question.setAnswer(cstring + ";");
      return;
    }
    cstring = cstring + ";\n";
    writer.print(cstring);
    istream.append(cstring);
    setInputFocus();
    writer.flush();
  }
  
  /***************************************************************************
   * Set focus to input field.
   **************************************************************************/
  public void setInputFocus()
  {
    getDisplay().syncExec(new Runnable() {
      public void run() {
        input.setFocus();
      }}
    );
  }
  
  /***************************************************************************
   * Make window visible or invisible.
   * @param visible the requested visibility status.
   **************************************************************************/
  public void setVisible(final boolean visible)
  {
    getDisplay().syncExec(new Runnable() {
      public void run() {
        if (topWindow == null) return;
        if (visible)
          topWindow.displayVerifyView();
        else
          { /*topWindow.displayAnalyzeView();*/ }
      }}
    );
  }
}
