// ----------------------------------------------------------------------------
// Environment.java
// The context that determines the meaning of identifiers.
// $Id: Environment.java,v 1.32 2017/07/04 08:04:35 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Semantics;

import java.util.*;

import fmrisc.ProofNavigator.*;
import fmrisc.ProofNavigator.Syntax.*;

/**
 * 
 * The context that determines the meaning of identifiers.
 *
 */
public final class Environment
{
  // the type and value environment (may be nested)
  private Stack<TypeTable> tstack = new Stack<TypeTable>(); // stack of type tables
  private Stack<ValueTable> vstack = new Stack<ValueTable>(); // stack of value tables
  private TypeTable ttable = null;    // current top of tstack
  private ValueTable vtable = null;   // current top of vstack;
  
  // the value declarations at level 0 (order preserved)
  private Vector<ValueSymbol> values = new Vector<ValueSymbol>(); 
  
  // the formula environment (not nested)
  private FormulaTable ftable = new FormulaTable(); 
  
  // the type checking conditions of the symbols in their respective order
  // nested according to the structure of tstack/vstack
  private Stack<Vector<Expression>> tccStack = new Stack<Vector<Expression>>(); // stack of tcc vectors
  private Vector<Expression> tccs = null;           // current top of tccStack
  
  // bookkeeping information for local variables
  private int depth = -1;               // the current depth
  
  /***************************************************************************
   * create empty type and value environment
   **************************************************************************/
  public Environment()
  {
    enter();
  }
  
  /***************************************************************************
   * Create clone of global part of environment; the clone is a shallow
   * copy (the symbols in the environment are not cloned).
   * @return clone of environment containing all globally declared idents 
   **************************************************************************/
  public Environment cloneGlobal()
  {
    Environment env = new Environment();
    TypeTable tnew = ((TypeTable)(tstack.elementAt(0))).cloneTypeTable();
    ValueTable vnew = ((ValueTable)(vstack.elementAt(0))).cloneValueTable();
    FormulaTable fnew = ftable.cloneFormulaTable();
    @SuppressWarnings("unchecked")
    Vector<ValueSymbol> valnew = (Vector<ValueSymbol>)values.clone();
    Vector<Expression> tccsNew = cloneTCCs();
    env.tstack = new Stack<TypeTable>();
    env.tstack.push(tnew);
    env.ttable = tnew;
    env.vstack = new Stack<ValueTable>();
    env.vstack.push(vnew);
    env.tccStack = new Stack<Vector<Expression>>();
    env.tccStack.push(tccsNew);
    env.vtable = vnew;
    env.ftable = fnew;
    env.tccs = tccsNew;
    env.values = valnew;
    return env;
  }
  
  /***************************************************************************
   * enter a new scope
   * all changes done to the environment after enter() will be undone
   * after exit()
   **************************************************************************/
  public void enter()
  {
    depth = depth+1;
    ttable = new TypeTable();
    vtable = new ValueTable();
    tccs = new Vector<Expression>();
    tstack.push(ttable);
    vstack.push(vtable);
    tccStack.push(tccs);
  }
  
  /***************************************************************************
   * exit a scope
   * all changes done after the after the last enter() are undone
   **************************************************************************/
  public void exit()
  {
    depth=depth-1;
    tstack.pop();
    vstack.pop();
    tccStack.pop();
    ttable = (TypeTable)(tstack.peek());
    vtable = (ValueTable)(vstack.peek());
    tccs = tccStack.peek();
  }
  
  /****************************************************************************
   * put type in type table
   * @param name the name of the type
   * @param type the type associated to the name
   * @return the constructed type symbol (null, if an error occured)
   ***************************************************************************/
  public TypeSymbol putType(TypeDeclIdentifier name, Type type)
  {
    Type type0 = Checking.canonicalTcc(type);
    if (type0 == null) return null;
    return ttable.put(name, type0); 
  }
  
  /****************************************************************************
   * returns type symbol associated to key
   * @param name the name of the type
   * @return the symbol associated to type
   ***************************************************************************/
  public TypeSymbol getTypeSymbol(Identifier name)
  {
    int i = tstack.size()-1;
    TypeTable table = ttable;
    while (true)
    {
      TypeSymbol symbol = table.getSymbol(name);
      if (symbol != null) return symbol;
      i = i-1;
      if (i < 0) break;
      table = (TypeTable)(tstack.elementAt(i));
    }
    return null;
  }
  
  /****************************************************************************
   * returns canonical type associated to key
   * type equality implies pointer equality of types.
   * @param name the name of the type
   * @return the canonical type associated to key
   ***************************************************************************/
  public Type getType(Identifier name)
  {
    TypeSymbol symbol = getTypeSymbol(name);
    if (symbol == null) return null;
    return symbol.getType();
  }
  
  /****************************************************************************
   * put named value in value table
   * @param name the name of the value
   * @param type the type associated to the value
   * @param value the value itself (may be null)
   * @return the constructed symbol (null, if an error occured)
   ***************************************************************************/
  public ValueSymbol putValue(ValueDeclIdentifier name, 
    Type type, Expression value)
  {
    Type type0 = null;
    if (type != null)
    {
      type0 = Checking.canonicalTcc(type);
      if (type0 == null) return null; 
      if (!Checking.validDefinitionType(type0))
      {
        State.setError("in declaration of constant " + name.getName() 
            + " type " + type0 
            + " is not first order (you must not use type " 
            + (State.booleanAsLogical() ? "LOGICAL" : "BOOLEAN")
            + " or a function type here)");
        return null;
      }
    }
    if (value != null)
    {
      Type type1 = Checking.processDefinitionTerm(value);
      if (type1 == null) return null;
      if (type0 == null)
        type0 = type1;
      else 
      {
        // cope with CVCL type matching restrictions (really needed ???)
        // Checking.setStrongerTypes(true);
        boolean okay = Checking.hasType(value, type1, type0, true);
        // Checking.setStrongerTypes(false);
        if (!okay)
        {
          State.setError("in declaration of constant " + name.getName() +
            ", value type " + type1.toString() +
            " does not match declaration type " + type.toString());
          return null;
        }
      }
    }
    ValueSymbol symbol = vtable.put(name, type0, value, depth);
    if (vstack.size() == 1) values.addElement(symbol);
    return symbol;
  }
  
  /****************************************************************************
   * returns symbol associated to name (null, if none)
   * @param name the name 
   * @return the symbol associated to the name (may be null)
   ***************************************************************************/
  public ValueSymbol getValueSymbol(Identifier name)
  {
    int i = vstack.size()-1;
    ValueTable table = vtable;
    while (true)
    {
      ValueSymbol symbol = table.getSymbol(name);
      if (symbol != null) return symbol;
      i = i-1;
      if (i < 0) break;
      table = (ValueTable)(vstack.elementAt(i));
    }
    return null;
  }
  
  /***************************************************************************
   * put type declaration into environment
   * @param tdecl type declaration
   * @return the constructed symbol (null, if an error occured)
   **************************************************************************/
  public TypeSymbol putTypeDeclaration(TypeDeclaration tdecl)
  {
    TypeDeclIdentifier ident = tdecl.getTypeDeclIdentifier();
    Type type = ttable.getType(ident);
    if (type != null)
    {
      State.setError("type " + ident.toString() +
          " already declared in current scope (as " + type.toString() +
          ")");
      return null;
    }
    ValueSymbol symbol = vtable.getSymbol(ident);
    if (symbol != null)
    {
      State.setError("value " + ident.toString() + ":" + 
        symbol.getType().toString() +
        " already declared in current scope");
      return null;
    }
    type = tdecl.getType();
    if (type == null)
      return putType(ident, Construct.atomicType(ident.getName()));
    else
      return putType(ident, type);
  }
  
  /***************************************************************************
   * put value declaration into environment
   * @param vdecl value declaration
   * @return the constructed symbol (null, if an error occured)
   **************************************************************************/
  public ValueSymbol putValueDeclaration(ValueDeclaration vdecl)
  {
    ValueDeclIdentifier ident = vdecl.getValueDeclIdentifier();
    ValueSymbol symbol = vtable.getSymbol(ident);
    if (symbol != null)
    {
      String msg = ident.toString() +
          " already declared in current scope (with type " + 
          symbol.getType().toString() + ")";
      State.setError(msg);
      return null;
    }
    Type type = ttable.getType(ident);
    if (type != null)
    {
      State.setError("type " + ident.toString() +
          " already declared in current scope (as " + type.toString() +
          ")");
      return null;
    }
    type = vdecl.getType();
    Expression value = vdecl.getValue();
    return putValue(ident, type, value);
  }
   
  /***************************************************************************
   * put variables into environment
   * @param vars the variable sequence
   * @return true iff no problem occured
   **************************************************************************/
  public boolean putVariables(TypedIdentifier[] vars)
  {
    int n = vars.length;
    for (int i=0; i<n; i++)
    {
      boolean okay = putValue(vars[i].getIdentifier(), 
        vars[i].getType(), null) != null;
      if (!okay) return false;
    }
    return true;
  }
  
  /***************************************************************************
   * get local variables (all values with environment depth gt 0)
   * @return vector of ValueSymbol objects representing the local variables
   **************************************************************************/
   public Vector<Symbol> getLocalVariables()
   {
     Vector<Symbol> v = new Vector<Symbol>();
     int n = vstack.size();
     for (int i=1; i<n; i++)
     {
       Collection<Symbol> c = vstack.elementAt(i).getAll();
       v.addAll(c);
     }
     return v;
   }

   /***************************************************************************
    * put formula in environment
    * @param name the name of the formula
    * @param formula the expression associated to the formula
    * @param axiom true iff formula is axiom
    * @return the constructed symbol (null, if an error occured)
    **************************************************************************/
   public FormulaSymbol putFormula(FormulaDeclIdentifier name, 
     Expression formula, boolean axiom)
   {
     Type type0 = Checking.processFormula(formula);
     if (type0 == null) return null;
     if (!Checking.isBoolean(type0))
     {
       if (State.booleanAsLogical())
         State.setError("formula type must be LOGICAL (is " 
             + type0.toString() + ")");
       else
         State.setError("formula type must be BOOLEAN (is " 
           + type0.toString() + ")");
       return null;
     }
     return ftable.put(name, formula, axiom, this); 
   }
   
   /***************************************************************************
    * returns symbol associated to name (null, if none)
    * @param name the name 
    * @return the symbol associated to the name (may be null)
    **************************************************************************/
   public FormulaSymbol getFormulaSymbol(Identifier name)
   {
     return ftable.getSymbol(name);
   } 
   
   /***************************************************************************
    * add formula symbol to environment 
    * @param symbol symbol to be added (name must not clash with other
    * symbol in formula table)
    **************************************************************************/
   public void putFormulaSymbol(FormulaSymbol symbol)
   {
     ftable.put(symbol.getIdentifier(), symbol);
   }
   
   /***************************************************************************
    * put formula declaration into environment
    * @param fdecl formula declaration
    * @return the constructed symbol (null, if an error occured)
    **************************************************************************/
   public FormulaSymbol putFormulaDeclaration(FormulaDeclaration fdecl)
   {
     FormulaDeclIdentifier ident = fdecl.getFormulaDeclIdentifier();
     FormulaSymbol symbol = ftable.getSymbol(ident);
     if (symbol != null)
     {
       State.setError("formula " + ident.toString() + " already declared");
       return null;
     }
     Expression formula = fdecl.getFormula();
     return putFormula(ident, formula, fdecl instanceof AxiomDeclaration);
   }
   
   /***************************************************************************
    * add the symbol associated to ident to current environment
    * @param ident identifier for an already processed value declaration
    **************************************************************************/
   public void put(ValueDeclIdentifier ident)
   {
     ValueSymbol symbol = ident.getSymbol();
     if (vstack.size() == 1) values.addElement(symbol);
     vtable.put(ident, symbol);
     symbol.setDepth(depth);
   }
   
   /***************************************************************************
    * Get all type symbols.
    * @return collection of all type symbols declared in current scope.
    **************************************************************************/
   public Collection<Symbol> getTypeSymbols()
   {
     return ttable.getSymbols();
   }

   /***************************************************************************
    * Get all value symbols.
    * @return collection of all value symbols declared in current scope.
    **************************************************************************/
   public Collection<Symbol> getValueSymbols()
   {
     return vtable.getSymbols();
   }
   
   /***************************************************************************
    * Get global value symbols in the order in which they were entered.
    * @return vector of all value symbols stored at the top level
    * of this environment in the order in which they were entered.
    **************************************************************************/
   public Vector<ValueSymbol> getGlobalValueSymbols()
   {
     return values;
   }
   
   /***************************************************************************
    * Get all formula symbols.
    * @return collection of all formula symbols declared in current scope.
    **************************************************************************/
   public Collection<Symbol> getFormulaSymbols()
   {
     return ftable.getSymbols();
   }
   
   /**************************************************************************
    * Add type checking condition to environment.
    * @param formula the condition.
    *************************************************************************/
    public void addTCC(Expression formula)
    {
      tccs.add(formula);
    }
    
    /*************************************************************************
     * Get type checking conditions of current scope (not of all scopes).
     * @return the sequence of type checking conditions in the order
     * they were added.
     ************************************************************************/
    public Expression[] getTCCs()
    {
      return ASTUtil.toExpressionArray(tccs);
    }
    
    /*************************************************************************
     * Create clone of type checking conditions.
     * @return vector of cloned type checking conditions with order preserved.
     ************************************************************************/
    @SuppressWarnings("unchecked")
    private Vector<Expression> cloneTCCs()
    {
      return (Vector<Expression>)(tccs.clone());
    }
}
