// ----------------------------------------------------------------------------
// SubType.java
// Handling of subtypes.
// $Id: SubType.java,v 1.14 2015/09/22 17:15:23 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2005, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
// ----------------------------------------------------------------------------
package fmrisc.ProofNavigator.Syntax;

import java.io.*;

/**
 * 
 * Predicated subtypes.
 *
 */
public final class SubType extends TypeBase
{
  private Expression predicate;  // the base predicate
  private Type       base;       // the base type derived from the predicate
  
  /****************************************************************************
   * construct subtype with denoted predicate.
   * @param predicate the predicate defining the subtype
   ***************************************************************************/
  public SubType(Expression predicate)
  {
    this.predicate = predicate;
  }
  
  /****************************************************************************
   * returns predicate 
   * @return the predicate associated to the subtype
   ***************************************************************************/
  public Expression getPredicate()
  {
    return predicate;
  }
  
  /****************************************************************************
   * returns base type 
   * @return the base type of the subtype
   ***************************************************************************/
  public Type getBase()
  {
    return base;
  }
  
  /****************************************************************************
   * set base type 
   * @param base the base type of the subtype
   ***************************************************************************/
  public void setBase(Type base)
  {
    this.base = base;
  }
  
  /***************************************************************************
   * return copy of AST with value references instantiated by the substitutions
   * set in the corresponding declaration identifiers
   * @return the instantiated copy
   **************************************************************************/
  public Type instantiateType()
  {
    SubType result = Construct.subType(getPredicate().instantiate());
    result.setBase(getBase().instantiateType());
    return result;
  }
  
  /***************************************************************************
   * Prints text representation of type on out (without new line termination).
   * @param out the stream on which the text is written
   **************************************************************************/
  public void printCore(PrintWriter out)
  {
    out.print("SUBTYPE(");
    predicate.printCore(out);
    out.print(")");
  }
  
  /***************************************************************************
   * Accept visitor for a visit.
   * @param visitor the visitor who is accepted by this node.
   * @return the result of the visit (may be null)
   **************************************************************************/
  public AST accept(ASTVisitor visitor) 
  { 
    AST result = visitor.visit(this);
    if (result != null) return result;
    AST predicate0 = predicate.accept(visitor);
    AST base0 = null;
    if (base != null) base0 = base.accept(visitor);
    if (!visitor.isCloned()) return null;
    SubType stype0 = Construct.subType((Expression)predicate0);
    stype0.setBase((Type)base0);
    return stype0;
  }
}
